
{*******************************************************}
{                                                       }
{       Delphi Runtime Library                          }
{       System Utilities Unit                           }
{                                                       }
{       Copyright (C) 1995,96 Borland International     }
{                                                       }
{       Virtual Pascal v2.1                             }
{       Copyright (C) 1996-2000 vpascal.com             }
{                                                       }
{*******************************************************}

unit SysUtils;

{$H+,J+,P+,S-,T-,W-,R-,Z-,X+}
{&Delphi+,AlignRec-,AlignData+,CDecl-,Use32-,Open32-}

{$IFDEF LINUX}
  {$UNDEF Win32}
{$ENDIF}

interface

uses
{$IFDEF Win32}
  Windows,
{$ENDIF}
  VpSysLow;

type
  Integer = Longint;

const
  MaxInt = MaxLongint;

{ File open modes }

  fmOpenRead       = open_access_readOnly;
  fmOpenWrite      = open_access_writeOnly;
  fmOpenReadWrite  = open_access_ReadWrite;
  fmShareCompat    = open_share_denyNone;
  fmShareExclusive = open_share_denyReadWrite;
  fmShareDenyWrite = open_share_denyWrite;
  fmShareDenyRead  = open_share_denyRead;
  fmShareDenyNone  = open_share_denyNone;

{ File attribute constants }

  faReadOnly  = $00000001;
  faHidden    = $00000002;
  faSysFile   = $00000004;
  faVolumeID  = $00000008;      { Not used by OS/2 }
  faDirectory = $00000010;
  faArchive   = $00000020;
  faAnyFile   = $00000037;

{ File mode magic numbers }

  fmClosed = $A55AD7B0;
  fmInput  = $A55AD7B1;
  fmOutput = $A55AD7B2;
  fmInOut  = $A55AD7B3;

{ Seconds and milliseconds per day }

  SecsPerDay = 24 * 60 * 60;
  MSecsPerDay = SecsPerDay * 1000;

{ Days between 1/1/0001 and 12/31/1899 }

  DateDelta = 693594;

type

{ Type conversion records }

  WordRec = packed record
    Lo, Hi: Byte;
  end;

  LongRec = packed record
    Lo, Hi: Word;
  end;

  PtrRec = record
    Ofs: Longint;
  end;

  TMethod = record
    Code, Data: Pointer;
  end;

{ General arrays }

  PByteArray = ^TByteArray;
  TByteArray = array[0..512*1024*1024-1] of Byte;

  PWordArray = ^TWordArray;
  TWordArray = array[0..256*1024*1024-1] of Word;

{ Generic procedure pointer }

  TProcedure = procedure;

{ Generic filename type }

  TFileName = ShortString;

  TFixedPartOfSearchRec = packed record
    FindHandle: Longint;
    NameLStr: Pointer;
    Attr: Byte;
    Time: Longint;
    Size: Longint;
    Name: TFileName;
    Filler: array[0..3] of Char;
  end;

const
  TFixedSRecSize = SizeOf(TFixedPartOfSearchRec);

type
{ Search record used by FindFirst, FindNext, and FindClose }

  TSearchRec = packed record
    FindHandle: Longint;
    NameLStr: Pointer;
    Attr: Byte;
    Time: Longint;
    Size: Longint;
    Name: TFileName;
    Filler: array[0..3] of Char;
{$IFDEF LINUX}
    // Placed here to make patch work with full version as
    // well as with pure interface version of SysUtils.
    Private_data: array[1..sizeof(TOSSearchRec)-TFixedSRecSize] of Byte;
{$ENDIF}
{$IFDEF WIN32}
    ExcludeAttr: Longint;
    FindData: TWin32FindData;
{$ENDIF}
{$IFDEF DPMI32}
    Private_data: array[1..sizeof(TOSSearchRec)-4-4-1-4-4-256-4] of Byte;
{$ENDIF}
  end;

{ Typed-file and untyped-file record }

  TFileRec = record
    Handle: Integer;
    Mode: Integer;
    RecSize: Cardinal;
    Private: array[1..28] of Byte;
    UserData: array[1..32] of Byte;
    Name: array[0..259] of Char;
  end;

{ Text file record structure used for Text files }

  PTextBuf = ^TTextBuf;
  TTextBuf = array[0..127] of Char;
  TTextRec = record
    Handle: Integer;
    Mode: Integer;
    BufSize: Cardinal;
    BufPos: Cardinal;
    BufEnd: Cardinal;
    BufPtr: PChar;
    OpenFunc: Pointer;
    InOutFunc: Pointer;
    FlushFunc: Pointer;
    CloseFunc: Pointer;
    UserData: array[1..32] of Byte;
    Name: array[0..259] of Char;
    Buffer: TTextBuf;
  end;

{ FloatToText, FloatToTextFmt, TextToFloat, and FloatToDecimal type codes }

  TFloatValue = (fvExtended, fvCurrency);

{ FloatToText format codes }

  TFloatFormat = (ffGeneral, ffExponent, ffFixed, ffNumber, ffCurrency);

{ FloatToDecimal result record }

  TFloatRec = packed record
    Exponent: Smallint;
    Negative: Boolean;
    Digits: array[0..20] of Char;
  end;

{ Date and time record }

  TTimeStamp = record
    Time: Integer;      { Number of milliseconds since midnight }
    Date: Integer;      { One plus number of days since 1/1/0001 }
  end;

{ Exceptions }

  Exception = class(TObject)
  private
    FMessage: String;
    FHelpContext: Integer;
  public
    constructor Create(const Msg: string);
    constructor CreateFmt(const Msg: string; const Args: array of const);
    constructor CreateRes(Ident: Integer);
    constructor CreateResFmt(Ident: Integer; const Args: array of const);
    constructor CreateHelp(const Msg: string; AHelpContext: Integer);
    constructor CreateFmtHelp(const Msg: string; const Args: array of const;
      AHelpContext: Integer);
    constructor CreateResHelp(Ident: Integer; AHelpContext: Integer);
    constructor CreateResFmtHelp(Ident: Integer; const Args: array of const;
      AHelpContext: Integer);
    property HelpContext: Integer read FHelpContext write FHelpContext;
    property Message: string read FMessage write FMessage;
  end;

  ExceptClass = class of Exception;

  EAbort = class(Exception);

  EOutOfMemory = class(Exception)
  private
    AllowFree: Boolean;
  public
    destructor Destroy; override;
    procedure FreeInstance; override;
  end;

  EInOutError = class(Exception)
  public
    ErrorCode: Integer;
  end;

  EIntError = class(Exception);
  EDivByZero = class(EIntError);
  ERangeError = class(EIntError);
  EIntOverflow = class(EIntError);

  EMathError = class(Exception);
  EInvalidOp = class(EMathError);
  EZeroDivide = class(EMathError);
  EOverflow = class(EMathError);
  EUnderflow = class(EMathError);

  EInvalidPointer = class(Exception);

  EInvalidCast = class(Exception);

  EConvertError = class(Exception);

  EAccessViolation = class(Exception);
  EPrivilege = class(Exception);
  EStackOverflow = class(Exception);
  EControlC = class(Exception);

  EPropReadOnly = class(Exception);
  EPropWriteOnly = class(Exception);

  EExternalException = class(Exception)
  public
    ExceptionRecord: POSExceptionRecord;
  end;

const

{ Empty string and null string pointer. These constants are provided for
  backwards compatibility only. }

  EmptyStr: string = '';
  NullStr: PString = @EmptyStr;

{ Win32 platform identifier.  This will be one of the following values:

    0 : Win32s      VER_PLATFORM_WIN32s
    1 : Windows 95  VER_PLATFORM_WIN32_WINDOWS
    2 : Windows NT  VER_PLATFORM_WIN32_NT
    -1: OS/2 }

  Win32Platform: Integer = -1;

{ Currency and date/time formatting options }

{ CurrencyString - Defines the currency symbol used in floating-point to
  decimal conversions. Fetched using SysGetCurrencyFormat.

  CurrencyFormat - Defines the currency symbol placement and separation
  used in floating-point to decimal conversions. Possible values are:

    0 = '$1'
    1 = '1$'
    2 = '$ 1'
    3 = '1 $'

  The initial value is fetched using SysGetCurrencyFormat.

  NegCurrFormat - Defines the currency format for used in floating-point to
  decimal conversions of negative numbers. Possible values are:

    0 = '($1)'      4 = '(1$)'      8 = '-1 $'      12 = '$ -1'
    1 = '-$1'       5 = '-1$'       9 = '-$ 1'      13 = '1- $'
    2 = '$-1'       6 = '1-$'      10 = '1 $-'      14 = '($ 1)'
    3 = '$1-'       7 = '1$-'      11 = '$ 1-'      15 = '(1 $)'

  The initial value is fetched using SysGetCurrencyFormat.

  ThousandSeparator - The character used to separate thousands in numbers
  with more than three digits to the left of the decimal separator.
  The initial value is fetched using SysGetCurrencyFormat.

  DecimalSeparator - The character used to separate the integer part from
  the fractional part of a number. The initial value is fetched using
  SysGetCurrencyFormat.

  CurrencyDecimals - The number of digits to the right of the decimal point
  in a currency amount. The initial value is fetched using
  SysGetCurrencyFormat.

  DateSeparator - The character used to separate the year, month, and day
  parts of a date value. The initial value is fetched using SysGetDateFormat.

  ShortDateFormat - The format string used to convert a date value to a
  short string suitable for editing. For a complete description of date and
  time format strings, refer to the documentation for the FormatDate
  function. The short date format should only use the date separator
  character and the  m, mm, d, dd, yy, and yyyy format specifiers. The
  initial value is fetched using SysGetDateFormat.

  LongDateFormat - The format string used to convert a date value to a long
  string suitable for display but not for editing. For a complete description
  of date and time format strings, refer to the documentation for the
  FormatDate function. The initial value is fetched using SysGetDateFormat.

  TimeSeparator - The character used to separate the hour, minute, and
  second parts of a time value. The initial value is fetched using
  SysGetTimeFormat.

  TimeAMString - The suffix string used for time values between 00:00 and
  11:59 in 12-hour clock format. The initial value is fetched using
  SysGetTimeFormat.

  TimePMString - The suffix string used for time values between 12:00 and
  23:59 in 12-hour clock format. The initial value is fetched using
  SysGetTimeFormat.

  ShortTimeFormat - The format string used to convert a time value to a
  short string with only hours and minutes. The default value is computed
  by SysGetTimeFormat.

  LongTimeFormat - The format string used to convert a time value to a long
  string with hours, minutes, and seconds. The default value is computed
  by SysGetTimeFormat.

  ShortMonthNames - Array of strings containing short month names. The mmm
  format specifier in a format string passed to FormatDate causes a short
  month name to be substituted.

  LongMonthNames - Array of strings containing long month names. The mmmm
  format specifier in a format string passed to FormatDate causes a long
  month name to be substituted.

  ShortDayNames - Array of strings containing short day names. The ddd
  format specifier in a format string passed to FormatDate causes a short
  day name to be substituted.

  LongDayNames - Array of strings containing long day names. The dddd
  format specifier in a format string passed to FormatDate causes a long
  day name to be substituted. }

var
  CurrencyString: string;
  CurrencyFormat: Byte;
  NegCurrFormat: Byte;
  ThousandSeparator: Char;
  DecimalSeparator: Char;
  CurrencyDecimals: Byte;
  DateSeparator: Char;
  ShortDateFormat: string;
  LongDateFormat: string;
  TimeSeparator: Char;
  TimeAMString: string;
  TimePMString: string;
  ShortTimeFormat: string;
  LongTimeFormat: string;
  ShortMonthNames: array[1..12] of string;
  LongMonthNames: array[1..12] of string;
  ShortDayNames: array[1..7] of string;
  LongDayNames: array[1..7] of string;

{ Memory management routines }

{ AllocMem allocates a block of the given size on the heap. Each byte in
  the allocated buffer is set to zero. To dispose the buffer, use the
  FreeMem standard procedure. }

function AllocMem(Size: Cardinal): Pointer;

{ String handling routines }

{ NewStr allocates a copy of the given string on the heap. The size of the
  allocated heap block is Length(S) + 1. If the string is empty, NewStr
  returns NullStr and doesn't allocate any heap space. To dispose the
  string, use DisposeStr. }

function NewStr(const S: String): PString;

{ DisposeStr disposes a string pointer that was previously allocated using
  NewStr. If the given pointer is NIL or points to an empty string,
  StrDispose does nothing. }

procedure DisposeStr(P: PString);

{ AssignStr assigns a new dynamically allocated string to the given string
  pointer. AssignStr corresponds to the statement "DisposeStr(P)" followed
  by the statement "P := NewStr(S)". Note that P must be NIL or contain a
  valid string pointer before calling AssignStr. In other words, AssignStr
  cannot be used to initialize a string pointer variable. }

procedure AssignStr(var P: PString; const S: string);

{ AppendStr appends S to the end of Dest. AppendStr corresponds to the
  statement "Dest := Dest + S", which is more efficient. }

procedure AppendStr(var Dest: string; const S: string);

{ UpperCase converts all ASCII characters in the given string to upper case.
  The conversion affects only 7-bit ASCII characters between 'a' and 'z'. To
  convert 8-bit international characters, use AnsiUpperCase. }

function UpperCase(const S: string): string;

{ LowerCase converts all ASCII characters in the given string to lower case.
  The conversion affects only 7-bit ASCII characters between 'A' and 'Z'. To
  convert 8-bit international characters, use AnsiLowerCase. }

function LowerCase(const S: string): string;

{ CompareStr compares S1 to S2, with case-sensitivity. The return value is
  less than 0 if S1 < S2, 0 if S1 = S2, or greater than 0 if S1 > S2. The
  compare operation is based on the 8-bit ordinal value of each character
  and is not affected by the current code page. }

function CompareStr(const S1, S2: string): Integer;

{ CompareMem performs a binary compare of Length bytes of memory referenced
  by P1 to that of P2.  CompareMem returns True if the memory referenced by
  P1 is identical to that of P2. }

function CompareMem(P1, P2: Pointer; Length: Integer): Boolean;

{ CompareText compares S1 to S2, without case-sensitivity. The return value
  is the same as for CompareStr. The compare operation is based on the 8-bit
  ordinal value of each character, after converting 'a'..'z' to 'A'..'Z',
  and is not affected by the current code page. }

function CompareText(const S1, S2: string): Integer;

{ SameText compares S1 to S2, without case-sensitivity. Returns true if
  S1 and S2 are the equal, that is, if CompareText would return 0. SameText
  has the same 8-bit limitations as CompareText }

function SameText(const S1, S2: string): Boolean;

{ AnsiUpperCase converts all characters in the given string to upper case.
  The conversion uses the current Windows locale. }

function AnsiUpperCase(const S: string): string;

{ AnsiLowerCase converts all characters in the given string to lower case.
  The conversion uses the current code page. }

function AnsiLowerCase(const S: string): string;

{ AnsiCompareStr compares S1 to S2, with case-sensitivity. The compare
  operation is controlled by the current code page. The return value is the
  same as for CompareStr. }

function AnsiCompareStr(const S1, S2: string): Integer;

{ AnsiSameStr compares S1 to S2, with case-sensitivity. The compare
  operation is controlled by the current Windows locale. The return value
  is True if AnsiCompareStr would have returned 0. }

function AnsiSameStr(const S1, S2: string): Boolean;

{ AnsiCompareText compares S1 to S2, without case-sensitivity. The compare
  operation is controlled by the current Windows locale. The return value
  is the same as for CompareStr. }

function AnsiCompareText(const S1, S2: string): Integer;

{ AnsiSameText compares S1 to S2, without case-sensitivity. The compare
  operation is controlled by the current Windows locale. The return value
  is True if AnsiCompareText would have returned 0. }

function AnsiSameText(const S1, S2: string): Boolean;

{ AnsiStrComp compares S1 to S2, with case-sensitivity. The compare
  operation is controlled by the current Windows locale. The return value
  is the same as for CompareStr. }

function AnsiStrComp(S1, S2: PChar): Integer;

{ AnsiStrIComp compares S1 to S2, without case-sensitivity. The compare
  operation is controlled by the current code page. The return value
  is the same as for CompareStr. }

function AnsiStrIComp(S1, S2: PChar): Integer;

{ AnsiStrLComp compares S1 to S2, with case-sensitivity, up to a maximum
  length of MaxLen bytes. The compare operation is controlled by the
  current code page. The return value is the same as for CompareStr. }

function AnsiStrLComp(S1, S2: PChar; MaxLen: Cardinal): Integer;

{ AnsiStrLIComp compares S1 to S2, without case-sensitivity, up to a maximum
  length of MaxLen bytes. The compare operation is controlled by the
  current code page. The return value is the same as for CompareStr. }

function AnsiStrLIComp(S1, S2: PChar; MaxLen: Cardinal): Integer;

{ AnsiStrLower converts all characters in the given string to lower case.
  The conversion uses the current code page. }

function AnsiStrLower(Str: PChar): PChar;

{ AnsiStrUpper converts all characters in the given string to upper case.
  The conversion uses the current code page. }

function AnsiStrUpper(Str: PChar): PChar;

{ Trim trims leading and trailing spaces and control characters from the
  given string. }

function Trim(const S: string): string;

{ TrimLeft trims leading spaces and control characters from the given
  string. }

function TrimLeft(const S: string): string;

{ TrimRight trims trailing spaces and control characters from the given
  string. }

function TrimRight(const S: string): string;

{ QuotedStr returns the given string as a quoted string. A single quote
  character is inserted at the beginning and the end of the string, and
  for each single quote character in the string, another one is added. }

function QuotedStr(const S: string): string;

{ AdjustLineBreaks adjusts all line breaks in the given string to be true
  CR/LF sequences. The function changes any CR characters not followed by
  a LF and any LF characters not preceded by a CR into CR/LF pairs. }

function AdjustLineBreaks(const S: string): string;

{ IsValidIdent returns true if the given string is a valid identifier. An
  identifier is defined as a character from the set ['A'..'Z', 'a'..'z', '_']
  followed by zero or more characters from the set ['A'..'Z', 'a'..'z',
  '0..'9', '_']. }

function IsValidIdent(const Ident: string): Boolean;

{ IntToStr converts the given value to its decimal string representation. }

function IntToStr(Value: Integer): string;

{ IntToHex converts the given value to a hexadecimal string representation
  with the minimum number of digits specified. }

function IntToHex(Value: Integer; Digits: Integer): string;

{ StrToInt converts the given string to an integer value. If the string
  doesn't contain a valid value, an EConvertError exception is raised. }

function StrToInt(const S: string): Integer;

{ StrToIntDef converts the given string to an integer value. If the string
  doesn't contain a valid value, the value given by Default is returned. }

function StrToIntDef(const S: string; Default: Integer): Integer;

{ LoadStr loads the string resource given by Ident from the application's
  executable file. If the string resource does not exist, an empty string
  is returned. }

function LoadStr(Ident: Integer): string;

{ LoadStr loads the string resource given by Ident from the application's
  executable file, and uses it as the format string in a call to the
  Format function with the given arguments. }

function FmtLoadStr(Ident: Integer; const Args: array of const): string;

{ File management routines }

{ FileOpen opens the specified file using the specified access mode. The
  access mode value is constructed by OR-ing one of the fmOpenXXXX constants
  with one of the fmShareXXXX constants. If the return value is positive,
  the function was successful and the value is the file handle of the opened
  file. If the return value is negative, an error occurred and the value is
  a negative OS/2 error code. }

function FileOpen(const FileName: string; Mode: Integer): Integer;

{ FileCreate creates a new file by the specified name. If the return value
  is positive, the function was successful and the value is the file handle
  of the new file. If the return value is negative, an error occurred and
  the value is a negative OS/2 error code. }

function FileCreate(const FileName: string): Integer;

{ FileRead reads Count bytes from the file given by Handle into the buffer
  specified by Buffer. The return value is the number of bytes actually
  read; it is less than Count if the end of the file was reached. The return
  value is -1 if an error occurred. }

function FileRead(Handle: Integer; var Buffer; Count: Integer): Integer;

{ FileWrite writes Count bytes to the file given by Handle from the buffer
  specified by Buffer. The return value is the number of bytes actually
  written, or -1 if an error occurred. }

function FileWrite(Handle: Integer; const Buffer; Count: Integer): Integer;

{ FileSeek changes the current position of the file given by Handle to be
  Offset bytes relative to the point given by Origin. Origin = 0 means that
  Offset is relative to the beginning of the file, Origin = 1 means that
  Offset is relative to the current position, and Origin = 2 means that
  Offset is relative to the end of the file. The return value is the new
  current position, relative to the beginning of the file, or -1 if an error
  occurred. }

function FileSeek(Handle, Offset, Origin: Integer): Integer;

{ FileClose closes the specified file. }

procedure FileClose(Handle: Integer);

{ FileAge returns the date-and-time stamp of the specified file. The return
  value can be converted to a TDateTime value using the FileDateToDateTime
  function. The return value is -1 if the file does not exist. }

function FileAge(const FileName: string): Integer;

{ FileExists returns a boolean value that indicates whether the specified
  file exists. }

function FileExists(const FileName: string): Boolean;

{ FindFirst searches the directory given by Path for the first entry that
  matches the filename given by Path and the attributes given by Attr. The
  result is returned in the search record given by SearchRec. The return
  value is zero if the function was successful. Otherwise the return value
  is an OS/2 error code; a value of 18 indicates that no files were
  found. FindFirst is typically used in conjunction with FindNext and
  FindClose as follows:

    Result := FindFirst(Path, Attr, SearchRec);
    while Result = 0 do
    begin
      ProcessSearchRec(SearchRec);
      Result := FindNext(SearchRec);
    end;
    FindClose(SearchRec);

  where ProcessSearchRec represents user-defined code that processes the
  information in a search record. }

function FindFirst(const Path: string; Attr: Integer;
  var F: TSearchRec): Integer;

{ FindNext returs the next entry that matches the name and attributes
  specified in a previous call to FindFirst. The search record must be one
  that was passed to FindFirst. The return value is zero if the function was
  successful. Otherwise the return value is an OS/2 error code; a
  value of 18 indicates that there are no more files matching the search
  criteria. }

function FindNext(var F: TSearchRec): Integer;

{ FindClose terminates a FindFirst/FindNext sequence. FindClose does nothing
  in the 16-bit version of Windows, but is required in the 32-bit version,
  so for maximum portability every FindFirst/FindNext sequence should end
  with a call to FindClose. }

procedure FindClose(var F: TSearchRec);

{ FileGetDate returns the DOS date-and-time stamp of the file given by
  Handle. The return value is -1 if the handle is invalid. The
  FileDateToDateTime function can be used to convert the returned value to
  a TDateTime value. }

function FileGetDate(Handle: Integer): Integer;

{ FileSetDate sets the DOS date-and-time stamp of the file given by Handle
  to the value given by Age. The DateTimeToFileDate function can be used to
  convert a TDateTime value to a DOS date-and-time stamp. The return value
  is zero if the function was successful. Otherwise the return value is an
  OS/2 error code. }

function FileSetDate(Handle, Age: Integer): Integer;

{ FileGetAttr returns the file attributes of the file given by FileName. The
  attributes can be examined by AND-ing with the faXXXX constants defined
  above. A return value of -1 indicates that an error occurred. }

function FileGetAttr(const FileName: string): Integer;

{ FileSetAttr sets the file attributes of the file given by FileName to the
  value given by Attr. The attribute value is formed by OR-ing the
  appropriate faXXXX constants. The return value is zero if the function was
  successful. Otherwise the return value is an OS/2 error code. }

function FileSetAttr(const FileName: string; Attr: Integer): Integer;

{ DeleteFile deletes the file given by FileName. The return value is True if
  the file was successfully deleted, or False if an error occurred. }

function DeleteFile(const FileName: string): Boolean;

{ RenameFile renames the file given by OldName to the name given by NewName.
  The return value is True if the file was successfully renamed, or False if
  an error occurred. }

function RenameFile(const OldName, NewName: string): Boolean;

{ ChangeFileExt changes the extension of a filename. FileName specifies a
  filename with or without an extension, and Extension specifies the new
  extension for the filename. The new extension can be a an empty string or
  a period followed by up to three characters. }

function ChangeFileExt(const FileName, Extension: string): string;

{ ExtractFilePath extracts the drive and directory parts of the given
  filename. The resulting string is the leftmost characters of FileName,
  up to and including the colon or backslash that separates the path
  information from the name and extension. The resulting string is empty
  if FileName contains no drive and directory parts. }

function ExtractFilePath(const FileName: string): string;

{ ExtractFileDir extracts the drive and directory parts of the given
  filename. The resulting string is a directory name suitable for passing
  to SetCurrentDir, CreateDir, etc. The resulting string is empty if
  FileName contains no drive and directory parts. }

function ExtractFileDir(const FileName: string): string;

{ ExtractFileDrive extracts the drive part of the given filename.  For
  filenames with drive letters, the resulting string is '<drive>:'.
  For filenames with a UNC path, the resulting string is in the form
  '\\<servername>\<sharename>'.  If the given path contains neither
  style of filename, the result is an empty string. }

function ExtractFileDrive(const FileName: string): string;

{ ExtractFileName extracts the name and extension parts of the given
  filename. The resulting string is the leftmost characters of FileName,
  starting with the first character after the colon or backslash that
  separates the path information from the name and extension. The resulting
  string is equal to FileName if FileName contains no drive and directory
  parts. }

function ExtractFileName(const FileName: string): string;

{ ExtractFileExt extracts the extension part of the given filename. The
  resulting string includes the period character that separates the name
  and extension parts. The resulting string is empty if the given filename
  has no extension. }

function ExtractFileExt(const FileName: string): string;

{ ExpandFileName expands the given filename to a fully qualified filename.
  The resulting string consists of a drive letter, a colon, a root relative
  directory path, and a filename. Embedded '.' and '..' directory references
  are removed. }

function ExpandFileName(const FileName: string): string;

{ ExpandUNCFileName expands the given filename to a fully qualified filename.
  This function is the same as ExpandFileName except that it will return the
  drive portion of the filename in the format '\\<servername>\<sharename> if
  that drive is actually a network resource instead of a local resource.
  Like ExpandFileName, embedded '.' and '..' directory references are
  removed. }

function ExpandUNCFileName(const FileName: string): string;

{  ExtractRelativePath will return a file path name relative to the given
   BaseName.  It strips the common path dirs and adds '..\' for each level
   up from the BaseName path. }

function ExtractRelativePath(const BaseName, DestName: string): string;

{ FileSearch searches for the file given by Name in the list of directories
  given by DirList. The directory paths in DirList must be separated by
  semicolons. The search always starts with the current directory of the
  current drive. The returned value is a concatenation of one of the
  directory paths and the filename, or an empty string if the file could not
  be located. }

function FileSearch(const Name, DirList: string): string;

{ DiskFree returns the number of free bytes on the specified drive number,
  where 0 = Current, 1 = A, 2 = B, etc. DiskFree returns -1 if the drive
  number is invalid. }

function DiskFree(Drive: Byte): Integer;

{ DiskSize returns the size in bytes of the specified drive number, where
  0 = Current, 1 = A, 2 = B, etc. DiskSize returns -1 if the drive number
  is invalid. }

function DiskSize(Drive: Byte): Integer;

{ FileDateToDateTime converts a DOS date-and-time value to a TDateTime
  value. The FileAge, FileGetDate, and FileSetDate routines operate on DOS
  date-and-time values, and the Time field of a TSearchRec used by the
  FindFirst and FindNext functions contains a DOS date-and-time value. }

function FileDateToDateTime(FileDate: Integer): TDateTime;

{ DateTimeToFileDate converts a TDateTime value to a DOS date-and-time
  value. The FileAge, FileGetDate, and FileSetDate routines operate on DOS
  date-and-time values, and the Time field of a TSearchRec used by the
  FindFirst and FindNext functions contains a DOS date-and-time value. }

function DateTimeToFileDate(DateTime: TDateTime): Integer;

{ GetCurrentDir returns the current directory. }

function GetCurrentDir: string;

{ SetCurrentDir sets the current directory. The return value is True if
  the current directory was successfully changed, or False if an error
  occurred. }

function SetCurrentDir(const Dir: string): Boolean;

{ CreateDir creates a new directory. The return value is True if a new
  directory was successfully created, or False if an error occurred. }

function CreateDir(const Dir: string): Boolean;

{ RemoveDir deletes an existing empty directory. The return value is
  True if the directory was successfully deleted, or False if an error
  occurred. }

function RemoveDir(const Dir: string): Boolean;

{ PChar routines }

{ StrLen returns the number of characters in Str, not counting the null
  terminator. }

function StrLen(Str: PChar): Cardinal;

{ StrEnd returns a pointer to the null character that terminates Str. }

function StrEnd(Str: PChar): PChar;

{ StrMove copies exactly Count characters from Source to Dest and returns
  Dest. Source and Dest may overlap. }

function StrMove(Dest, Source: PChar; Count: Cardinal): PChar;

{ StrCopy copies Source to Dest and returns Dest. }

function StrCopy(Dest, Source: PChar): PChar;

{ StrECopy copies Source to Dest and returns StrEnd(Dest). }

function StrECopy(Dest, Source: PChar): PChar;

{ StrLCopy copies at most MaxLen characters from Source to Dest and
  returns Dest. }

function StrLCopy(Dest, Source: PChar; MaxLen: Cardinal): PChar;

{ StrPCopy copies the Pascal style string Source into Dest and
  returns Dest. }

function StrPCopy(Dest: PChar; const Source: String): PChar;

{ StrPLCopy copies at most MaxLen characters from the Pascal style string
  Source into Dest and returns Dest. }

function StrPLCopy(Dest: PChar; const Source: string; MaxLen: Cardinal): PChar;

{ StrCat appends a copy of Source to the end of Dest and returns Dest. }

function StrCat(Dest, Source: PChar): PChar;

{ StrLCat appends at most MaxLen - StrLen(Dest) characters from Source to
  the end of Dest, and returns Dest. }

function StrLCat(Dest, Source: PChar; MaxLen: Cardinal): PChar;

{ StrComp compares Str1 to Str2. The return value is less than 0 if
  Str1 < Str2, 0 if Str1 = Str2, or greater than 0 if Str1 > Str2. }

function StrComp(Str1, Str2: PChar): Integer;

{ StrIComp compares Str1 to Str2, without case sensitivity. The return
  value is the same as StrComp. }

function StrIComp(Str1, Str2: PChar): Integer;

{ StrLComp compares Str1 to Str2, for a maximum length of MaxLen
  characters. The return value is the same as StrComp. }

function StrLComp(Str1, Str2: PChar; MaxLen: Cardinal): Integer;

{ StrLIComp compares Str1 to Str2, for a maximum length of MaxLen
  characters, without case sensitivity. The return value is the same
  as StrComp. }

function StrLIComp(Str1, Str2: PChar; MaxLen: Cardinal): Integer;

{ StrScan returns a pointer to the first occurrence of Chr in Str. If Chr
  does not occur in Str, StrScan returns NIL. The null terminator is
  considered to be part of the string. }

function StrScan(Str: PChar; Chr: Char): PChar;

{ StrRScan returns a pointer to the last occurrence of Chr in Str. If Chr
  does not occur in Str, StrRScan returns NIL. The null terminator is
  considered to be part of the string. }

function StrRScan(Str: PChar; Chr: Char): PChar;

{ StrPos returns a pointer to the first occurrence of Str2 in Str1. If
  Str2 does not occur in Str1, StrPos returns NIL. }

function StrPos(Str1, Str2: PChar): PChar;

{ StrUpper converts Str to upper case and returns Str. }

function StrUpper(Str: PChar): PChar;

{ StrLower converts Str to lower case and returns Str. }

function StrLower(Str: PChar): PChar;

{ StrPas converts Str to a Pascal style string. }

function StrPas(Str: PChar): String;

{ StrAlloc allocates a buffer of the given size on the heap. The size of
  the allocated buffer is encoded in a four byte header that immediately
  preceeds the buffer. To dispose the buffer, use StrDispose. }

function StrAlloc(Size: Cardinal): PChar;

{ StrBufSize returns the allocated size of the given buffer, not including
  the two byte header. }

function StrBufSize(Str: PChar): Cardinal;

{ StrNew allocates a copy of Str on the heap. If Str is NIL, StrNew returns
  NIL and doesn't allocate any heap space. Otherwise, StrNew makes a
  duplicate of Str, obtaining space with a call to the StrAlloc function,
  and returns a pointer to the duplicated string. To dispose the string,
  use StrDispose. }

function StrNew(Str: PChar): PChar;

{ StrPNew allocates a copy of Str on the heap. If Str is empty, StrNew returns
  NIL and doesn't allocate any heap space. Otherwise, StrNew makes a
  duplicate of Str, obtaining space with a call to the StrAlloc function,
  and returns a pointer to the duplicated string. To dispose the string,
  use StrDispose. }

function StrPNew(Str: String): PChar;

{ StrDispose disposes a string that was previously allocated with StrAlloc
  or StrNew. If Str is NIL, StrDispose does nothing. }

procedure StrDispose(Str: PChar);

{ String formatting routines }

{ The Format routine formats the argument list given by the Args parameter
  using the format string given by the Format parameter.

  Format strings contain two types of objects--plain characters and format
  specifiers. Plain characters are copied verbatim to the resulting string.
  Format specifiers fetch arguments from the argument list and apply
  formatting to them.

  Format specifiers have the following form:

    "%" [index ":"] ["-"] [width] ["." prec] type

  A format specifier begins with a % character. After the % come the
  following, in this order:

  -  an optional argument index specifier, [index ":"]
  -  an optional left-justification indicator, ["-"]
  -  an optional width specifier, [width]
  -  an optional precision specifier, ["." prec]
  -  the conversion type character, type

  The following conversion characters are supported:

  d  Decimal. The argument must be an integer value. The value is converted
     to a string of decimal digits. If the format string contains a precision
     specifier, it indicates that the resulting string must contain at least
     the specified number of digits; if the value has less digits, the
     resulting string is left-padded with zeros.

  e  Scientific. The argument must be a floating-point value. The value is
     converted to a string of the form "-d.ddd...E+ddd". The resulting
     string starts with a minus sign if the number is negative, and one digit
     always precedes the decimal point. The total number of digits in the
     resulting string (including the one before the decimal point) is given
     by the precision specifer in the format string--a default precision of
     15 is assumed if no precision specifer is present. The "E" exponent
     character in the resulting string is always followed by a plus or minus
     sign and at least three digits.

  f  Fixed. The argument must be a floating-point value. The value is
     converted to a string of the form "-ddd.ddd...". The resulting string
     starts with a minus sign if the number is negative. The number of digits
     after the decimal point is given by the precision specifier in the
     format string--a default of 2 decimal digits is assumed if no precision
     specifier is present.

  g  General. The argument must be a floating-point value. The value is
     converted to the shortest possible decimal string using fixed or
     scientific format. The number of significant digits in the resulting
     string is given by the precision specifier in the format string--a
     default precision of 15 is assumed if no precision specifier is present.
     Trailing zeros are removed from the resulting string, and a decimal
     point appears only if necessary. The resulting string uses fixed point
     format if the number of digits to the left of the decimal point in the
     value is less than or equal to the specified precision, and if the
     value is greater than or equal to 0.00001. Otherwise the resulting
     string uses scientific format.

  n  Number. The argument must be a floating-point value. The value is
     converted to a string of the form "-d,ddd,ddd.ddd...". The "n" format
     corresponds to the "f" format, except that the resulting string
     contains thousand separators.

  m  Money. The argument must be a floating-point value. The value is
     converted to a string that represents a currency amount. The conversion
     is controlled by the CurrencyString, CurrencyFormat, NegCurrFormat,
     ThousandSeparator, DecimalSeparator, and CurrencyDecimals global
     variables, all of which are initialized from the Currency Format in
     the International section of the Windows Control Panel. If the format
     string contains a precision specifier, it overrides the value given
     by the CurrencyDecimals global variable.

  p  Pointer. The argument must be a pointer value. The value is converted
     to a string of the form "XXXX:YYYY" where XXXX and YYYY are the
     segment and offset parts of the pointer expressed as four hexadecimal
     digits.

  s  String. The argument must be a character, a string, or a PChar value.
     The string or character is inserted in place of the format specifier.
     The precision specifier, if present in the format string, specifies the
     maximum length of the resulting string. If the argument is a string
     that is longer than this maximum, the string is truncated.

  x  Hexadecimal. The argument must be an integer value. The value is
     converted to a string of hexadecimal digits. If the format string
     contains a precision specifier, it indicates that the resulting string
     must contain at least the specified number of digits; if the value has
     less digits, the resulting string is left-padded with zeros.

  Conversion characters may be specified in upper case as well as in lower
  case--both produce the same results.

  For all floating-point formats, the actual characters used as decimal and
  thousand separators are obtained from the DecimalSeparator and
  ThousandSeparator global variables.

  Index, width, and precision specifiers can be specified directly using
  decimal digit string (for example "%10d"), or indirectly using an asterisk
  charcater (for example "%*.*f"). When using an asterisk, the next argument
  in the argument list (which must be an integer value) becomes the value
  that is actually used. For example "Format('%*.*f', [8, 2, 123.456])" is
  the same as "Format('%8.2f', [123.456])".

  A width specifier sets the minimum field width for a conversion. If the
  resulting string is shorter than the minimum field width, it is padded
  with blanks to increase the field width. The default is to right-justify
  the result by adding blanks in front of the value, but if the format
  specifier contains a left-justification indicator (a "-" character
  preceding the width specifier), the result is left-justified by adding
  blanks after the value.

  An index specifier sets the current argument list index to the specified
  value. The index of the first argument in the argument list is 0. Using
  index specifiers, it is possible to format the same argument multiple
  times. For example "Format('%d %d %0:d %d', [10, 20])" produces the string
  '10 20 10 20'.

  The Format function can be combined with other formatting functions. For
  example

    S := Format('Your total was %s on %s', [
      FormatFloat('$#,##0.00;;zero', Total),
      FormatDateTime('mm/dd/yy', Date)]);

  which uses the FormatFloat and FormatDateTime functions to customize the
  format beyond what is possible with Format. }

function Format(const Format: string; const Args: array of const): string;

{ FmtStr formats the argument list given by Args using the format string
  given by Format into the string variable given by Result. For further
  details, see the description of the Format function. }

procedure FmtStr(var Result: string; const Format: string;
  const Args: array of const);

{ StrFmt formats the argument list given by Args using the format string
  given by Format into the buffer given by Buffer. It is up to the caller to
  ensure that Buffer is large enough for the resulting string. The returned
  value is Buffer. For further details, see the description of the Format
  function. }

function StrFmt(Buffer, Format: PChar; const Args: array of const): PChar;

{ StrLFmt formats the argument list given by Args using the format string
  given by Format into the buffer given by Buffer. The resulting string will
  contain no more than MaxLen characters, not including the null terminator.
  The returned value is Buffer. For further details, see the description of
  the Format function. }

function StrLFmt(Buffer: PChar; MaxLen: Cardinal; Format: PChar;
  const Args: array of const): PChar;

{ FormatBuf formats the argument list given by Args using the format string
  given by Format and FmtLen into the buffer given by Buffer and BufLen.
  The Format parameter is a reference to a buffer containing FmtLen
  characters, and the Buffer parameter is a reference to a buffer of BufLen
  characters. The returned value is the number of characters actually stored
  in Buffer. The returned value is always less than or equal to BufLen. For
  further details, see the description of the Format function. }

function FormatBuf(var Buffer; BufLen: Cardinal; const Format;
  FmtLen: Cardinal; const Args: array of const): Cardinal;

{ Floating point conversion routines }

{ FloatToStr converts the floating-point value given by Value to its string
  representation. The conversion uses general number format with 15
  significant digits. For further details, see the description of the
  FloatToStrF function. }

function FloatToStr(Value: Extended): string;

{ CurrToStr converts the currency value given by Value to its string
  representation. The conversion uses general number format. For further
  details, see the description of the CurrToStrF function. }

function CurrToStr(Value: Currency): string;

{ FloatToStrF converts the floating-point value given by Value to its string
  representation. The Format parameter controls the format of the resulting
  string. The Precision parameter specifies the precision of the given value.
  It should be 7 or less for values of type Single, 15 or less for values of
  type Double, and 18 or less for values of type Extended. The meaning of the
  Digits parameter depends on the particular format selected.

  The possible values of the Format parameter, and the meaning of each, are
  described below.

  ffGeneral - General number format. The value is converted to the shortest
  possible decimal string using fixed or scientific format. Trailing zeros
  are removed from the resulting string, and a decimal point appears only
  if necessary. The resulting string uses fixed point format if the number
  of digits to the left of the decimal point in the value is less than or
  equal to the specified precision, and if the value is greater than or
  equal to 0.00001. Otherwise the resulting string uses scientific format,
  and the Digits parameter specifies the minimum number of digits in the
  exponent (between 0 and 4).

  ffExponent - Scientific format. The value is converted to a string of the
  form "-d.ddd...E+dddd". The resulting string starts with a minus sign if
  the number is negative, and one digit always precedes the decimal point.
  The total number of digits in the resulting string (including the one
  before the decimal point) is given by the Precision parameter. The "E"
  exponent character in the resulting string is always followed by a plus
  or minus sign and up to four digits. The Digits parameter specifies the
  minimum number of digits in the exponent (between 0 and 4).

  ffFixed - Fixed point format. The value is converted to a string of the
  form "-ddd.ddd...". The resulting string starts with a minus sign if the
  number is negative, and at least one digit always precedes the decimal
  point. The number of digits after the decimal point is given by the Digits
  parameter--it must be between 0 and 18. If the number of digits to the
  left of the decimal point is greater than the specified precision, the
  resulting value will use scientific format.

  ffNumber - Number format. The value is converted to a string of the form
  "-d,ddd,ddd.ddd...". The ffNumber format corresponds to the ffFixed format,
  except that the resulting string contains thousand separators.

  ffCurrency - Currency format. The value is converted to a string that
  represents a currency amount. The conversion is controlled by the
  CurrencyString, CurrencyFormat, NegCurrFormat, ThousandSeparator, and
  DecimalSeparator global variables, all of which are initialized from the
  Currency Format in the International section of the Windows Control Panel.
  The number of digits after the decimal point is given by the Digits
  parameter--it must be between 0 and 18.

  For all formats, the actual characters used as decimal and thousand
  separators are obtained from the DecimalSeparator and ThousandSeparator
  global variables.

  If the given value is a NAN (not-a-number), the resulting string is 'NAN'.
  If the given value is positive infinity, the resulting string is 'INF'. If
  the given value is negative infinity, the resulting string is '-INF'. }

function FloatToStrF(Value: Extended; Format: TFloatFormat;
  Precision, Digits: Integer): string;

{ CurrToStrF converts the currency value given by Value to its string
  representation. A call to CurrToStrF corresponds to a call to
  FloatToStrF with an implied precision of 19 digits. }

function CurrToStrF(Value: Currency; Format: TFloatFormat;
  Digits: Integer): string;

{ FloatToText converts the given floating-point value to its decimal
  representation using the specified format, precision, and digits. The
  Value parameter must be a variable of type Extended or Currency, as
  indicated by the ValueType parameter. The resulting string of characters
  is stored in the given buffer, and the returned value is the number of
  characters stored. The resulting string is not null-terminated. For
  further details, see the description of the FloatToStrF function. }

function FloatToText(Buffer: PChar; const Value; ValueType: TFloatValue;
  Format: TFloatFormat; Precision, Digits: Integer): Integer;

{ FormatFloat formats the floating-point value given by Value using the
  format string given by Format. The following format specifiers are
  supported in the format string:

  0     Digit placeholder. If the value being formatted has a digit in the
        position where the '0' appears in the format string, then that digit
        is copied to the output string. Otherwise, a '0' is stored in that
        position in the output string.

  #     Digit placeholder. If the value being formatted has a digit in the
        position where the '#' appears in the format string, then that digit
        is copied to the output string. Otherwise, nothing is stored in that
        position in the output string.

  .     Decimal point. The first '.' character in the format string
        determines the location of the decimal separator in the formatted
        value; any additional '.' characters are ignored. The actual
        character used as a the decimal separator in the output string is
        determined by the DecimalSeparator global variable. The default value
        of DecimalSeparator is specified in the Number Format of the
        International section in the Windows Control Panel.

  ,     Thousand separator. If the format string contains one or more ','
        characters, the output will have thousand separators inserted between
        each group of three digits to the left of the decimal point. The
        placement and number of ',' characters in the format string does not
        affect the output, except to indicate that thousand separators are
        wanted. The actual character used as a the thousand separator in the
        output is determined by the ThousandSeparator global variable. The
        default value of ThousandSeparator is specified in the Number Format
        of the International section in the Windows Control Panel.

  E+    Scientific notation. If any of the strings 'E+', 'E-', 'e+', or 'e-'
  E-    are contained in the format string, the number is formatted using
  e+    scientific notation. A group of up to four '0' characters can
  e-    immediately follow the 'E+', 'E-', 'e+', or 'e-' to determine the
        minimum number of digits in the exponent. The 'E+' and 'e+' formats
        cause a plus sign to be output for positive exponents and a minus
        sign to be output for negative exponents. The 'E-' and 'e-' formats
        output a sign character only for negative exponents.

  'xx'  Characters enclosed in single or double quotes are output as-is, and
  "xx"  do not affect formatting.

  ;     Separates sections for positive, negative, and zero numbers in the
        format string.

  The locations of the leftmost '0' before the decimal point in the format
  string and the rightmost '0' after the decimal point in the format string
  determine the range of digits that are always present in the output string.

  The number being formatted is always rounded to as many decimal places as
  there are digit placeholders ('0' or '#') to the right of the decimal
  point. If the format string contains no decimal point, the value being
  formatted is rounded to the nearest whole number.

  If the number being formatted has more digits to the left of the decimal
  separator than there are digit placeholders to the left of the '.'
  character in the format string, the extra digits are output before the
  first digit placeholder.

  To allow different formats for positive, negative, and zero values, the
  format string can contain between one and three sections separated by
  semicolons.

  One section - The format string applies to all values.

  Two sections - The first section applies to positive values and zeros, and
  the second section applies to negative values.

  Three sections - The first section applies to positive values, the second
  applies to negative values, and the third applies to zeros.

  If the section for negative values or the section for zero values is empty,
  that is if there is nothing between the semicolons that delimit the
  section, the section for positive values is used instead.

  If the section for positive values is empty, or if the entire format string
  is empty, the value is formatted using general floating-point formatting
  with 15 significant digits, corresponding to a call to FloatToStrF with
  the ffGeneral format. General floating-point formatting is also used if
  the value has more than 18 digits to the left of the decimal point and
  the format string does not specify scientific notation.

  The table below shows some sample formats and the results produced when
  the formats are applied to different values:

  Format string          1234        -1234       0.5         0
  -----------------------------------------------------------------------
                         1234        -1234       0.5         0
  0                      1234        -1234       1           0
  0.00                   1234.00     -1234.00    0.50        0.00
  #.##                   1234        -1234       .5
  #,##0.00               1,234.00    -1,234.00   0.50        0.00
  #,##0.00;(#,##0.00)    1,234.00    (1,234.00)  0.50        0.00
  #,##0.00;;Zero         1,234.00    -1,234.00   0.50        Zero
  0.000E+00              1.234E+03   -1.234E+03  5.000E-01   0.000E+00
  #.###E-0               1.234E3     -1.234E3    5E-1        0E0
  ----------------------------------------------------------------------- }

function FormatFloat(const Format: string; Value: Extended): string;

{ FormatCurr formats the currency value given by Value using the format
  string given by Format. For further details, see the description of the
  FormatFloat function. }

function FormatCurr(const Format: string; Value: Currency): string;

{ FloatToTextFmt converts the given floating-point value to its decimal
  representation using the specified format. The Value parameter must be a
  variable of type Extended or Currency, as indicated by the ValueType
  parameter. The resulting string of characters is stored in the given
  buffer, and the returned value is the number of characters stored. The
  resulting string is not null-terminated. For further details, see the
  description of the FormatFloat function. }

function FloatToTextFmt(Buffer: PChar; const Value; ValueType: TFloatValue;
  Format: PChar): Integer;

{ StrToFloat converts the given string to a floating-point value. The string
  must consist of an optional sign (+ or -), a string of digits with an
  optional decimal point, and an optional 'E' or 'e' followed by a signed
  integer. Leading and trailing blanks in the string are ignored. The
  DecimalSeparator global variable defines the character that must be used
  as a decimal point. Thousand separators and currency symbols are not
  allowed in the string. If the string doesn't contain a valid value, an
  EConvertError exception is raised. }

function StrToFloat(const S: string): Extended;

{ StrToCurr converts the given string to a currency value. For further
  details, see the description of the StrToFloat function. }

function StrToCurr(const S: string): Currency;

{ TextToFloat converts the null-terminated string given by Buffer to a
  floating-point value which is returned in the variable given by Value.
  The Value parameter must be a variable of type Extended or Currency, as
  indicated by the ValueType parameter. The return value is True if the
  conversion was successful, or False if the string is not a valid
  floating-point value. For further details, see the description of the
  StrToFloat function. }

function TextToFloat(Buffer: PChar; var Value; ValueType: TFloatValue): Boolean;

{ FloatToDecimal converts a floating-point value to a decimal representation
  that is suited for further formatting. The Value parameter must be a
  variable of type Extended or Currency, as indicated by the ValueType
  parameter. For values of type Extended, the Precision parameter specifies
  the requested number of significant digits in the result--the allowed range
  is 1..18. For values of type Currency, the Precision parameter is ignored,
  and the implied precision of the conversion is 19 digits. The Decimals
  parameter specifies the requested maximum number of digits to the left of
  the decimal point in the result. Precision and Decimals together control
  how the result is rounded. To produce a result that always has a given
  number of significant digits regardless of the magnitude of the number,
  specify 9999 for the Decimals parameter. The result of the conversion is
  stored in the specified TFloatRec record as follows:

  Exponent - Contains the magnitude of the number, i.e. the number of
  significant digits to the right of the decimal point. The Exponent field
  is negative if the absolute value of the number is less than one. If the
  number is a NAN (not-a-number), Exponent is set to -32768. If the number
  is INF or -INF (positive or negative infinity), Exponent is set to 32767.

  Negative - True if the number is negative, False if the number is zero
  or positive.

  Digits - Contains up to 18 (for type Extended) or 19 (for type Currency)
  significant digits followed by a null terminator. The implied decimal
  point (if any) is not stored in Digits. Trailing zeros are removed, and
  if the resulting number is zero, NAN, or INF, Digits contains nothing but
  the null terminator. }

procedure FloatToDecimal(var Result: TFloatRec; const Value;
  ValueType: TFloatValue; Precision, Decimals: Integer);

{ Date/time support routines }

function DateTimeToTimeStamp(DateTime: TDateTime): TTimeStamp;

function TimeStampToDateTime(const TimeStamp: TTimeStamp): TDateTime;
function MSecsToTimeStamp(MSecs: Comp): TTimeStamp;
function TimeStampToMSecs(const TimeStamp: TTimeStamp): Comp;

{ EncodeDate encodes the given year, month, and day into a TDateTime value.
  The year must be between 1 and 9999, the month must be between 1 and 12,
  and the day must be between 1 and N, where N is the number of days in the
  specified month. If the specified values are not within range, an
  EConvertError exception is raised. The resulting value is the number of
  days between 12/30/1899 and the given date. }

function EncodeDate(Year, Month, Day: Word): TDateTime;

{ EncodeTime encodes the given hour, minute, second, and millisecond into a
  TDateTime value. The hour must be between 0 and 23, the minute must be
  between 0 and 59, the second must be between 0 and 59, and the millisecond
  must be between 0 and 999. If the specified values are not within range, an
  EConvertError exception is raised. The resulting value is a number between
  0 (inclusive) and 1 (not inclusive) that indicates the fractional part of
  a day given by the specified time. The value 0 corresponds to midnight,
  0.5 corresponds to noon, 0.75 corresponds to 6:00 pm, etc. }

function EncodeTime(Hour, Min, Sec, MSec: Word): TDateTime;

{ DecodeDate decodes the integral (date) part of the given TDateTime value
  into its corresponding year, month, and day. If the given TDateTime value
  is less than or equal to zero, the year, month, and day return parameters
  are all set to zero. }

procedure DecodeDate(Date: TDateTime; var Year, Month, Day: Word);

{ DecodeTime decodes the fractional (time) part of the given TDateTime value
  into its corresponding hour, minute, second, and millisecond. }

procedure DecodeTime(Time: TDateTime; var Hour, Min, Sec, MSec: Word);

{ DayOfWeek returns the day of the week of the given date. The result is an
  integer between 1 and 7, corresponding to Sunday through Saturday. }

function DayOfWeek(Date: TDateTime): Integer;

{ Date returns the current date. }

function Date: TDateTime;

{ Time returns the current time. }

function Time: TDateTime;

{ Now returns the current date and time, corresponding to Date + Time. }

function Now: TDateTime;

{ DateToStr converts the date part of the given TDateTime value to a string.
  The conversion uses the format specified by the ShortDateFormat global
  variable. }

{ IncMonth returns Date shifted by the specified number of months.
  NumberOfMonths parameter can be negative, to return a date N months ago.
  If the input day of month is greater than the last day of the resulting
  month, the day is set to the last day of the resulting month.
  Input time of day is copied to the DateTime result.  }

function IncMonth(const Date: TDateTime; NumberOfMonths: Integer): TDateTime;

{ IsLeapYear determines whether the given year is a leap year. }

function IsLeapYear(Year: Word): Boolean;

{ DateToStr converts the date part of the given TDateTime value to a string.
  The conversion uses the format specified by the ShortDateFormat global
  variable. }

function DateToStr(Date: TDateTime): string;

{ TimeToStr converts the time part of the given TDateTime value to a string.
  The conversion uses the format specified by the LongTimeFormat global
  variable. }

function TimeToStr(Time: TDateTime): string;

{ DateTimeToStr converts the given date and time to a string. The resulting
  string consists of a date and time formatted using the ShortDateFormat and
  LongTimeFormat global variables. Time information is included in the
  resulting string only if the fractional part of the given date and time
  value is non-zero. }

function DateTimeToStr(DateTime: TDateTime): string;

{ StrToDate converts the given string to a date value. The string must
  consist of two or three numbers, separated by the character defined by
  the DateSeparator global variable. The order for month, day, and year is
  determined by the ShortDateFormat global variable--possible combinations
  are m/d/y, d/m/y, and y/m/d. If the string contains only two numbers, it
  is interpreted as a date (m/d or d/m) in the current year. Year values
  between 0 and 99 are assumed to be in the current century. If the given
  string does not contain a valid date, an EConvertError exception is
  raised. }

function StrToDate(const S: string): TDateTime;

{ StrToTime converts the given string to a time value. The string must
  consist of two or three numbers, separated by the character defined by
  the TimeSeparator global variable, optionally followed by an AM or PM
  indicator. The numbers represent hour, minute, and (optionally) second,
  in that order. If the time is followed by AM or PM, it is assumed to be
  in 12-hour clock format. If no AM or PM indicator is included, the time
  is assumed to be in 24-hour clock format. If the given string does not
  contain a valid time, an EConvertError exception is raised. }

function StrToTime(const S: string): TDateTime;

{ StrToDateTime converts the given string to a date and time value. The
  string must contain a date optionally followed by a time. The date and
  time parts of the string must follow the formats described for the
  StrToDate and StrToTime functions. }

function StrToDateTime(const S: string): TDateTime;

{ FormatDateTime formats the date-and-time value given by DateTime using the
  format given by Format. The following format specifiers are supported:

  c       Displays the date using the format given by the ShortDateFormat
          global variable, followed by the time using the format given by
          the LongTimeFormat global variable. The time is not displayed if
          the fractional part of the DateTime value is zero.

  d       Displays the day as a number without a leading zero (1-31).

  dd      Displays the day as a number with a leading zero (01-31).

  ddd     Displays the day as an abbreviation (Sun-Sat) using the strings
          given by the ShortDayNames global variable.

  dddd    Displays the day as a full name (Sunday-Saturday) using the strings
          given by the LongDayNames global variable.

  ddddd   Displays the date using the format given by the ShortDateFormat
          global variable.

  dddddd  Displays the date using the format given by the LongDateFormat
          global variable.

  m       Displays the month as a number without a leading zero (1-12). If
          the m specifier immediately follows an h or hh specifier, the
          minute rather than the month is displayed.

  mm      Displays the month as a number with a leading zero (01-12). If
          the mm specifier immediately follows an h or hh specifier, the
          minute rather than the month is displayed.

  mmm     Displays the month as an abbreviation (Jan-Dec) using the strings
          given by the ShortMonthNames global variable.

  mmmm    Displays the month as a full name (January-December) using the
          strings given by the LongMonthNames global variable.

  yy      Displays the year as a two-digit number (00-99).

  yyyy    Displays the year as a four-digit number (0000-9999).

  h       Displays the hour without a leading zero (0-23).

  hh      Displays the hour with a leading zero (00-23).

  n       Displays the minute without a leading zero (0-59).

  nn      Displays the minute with a leading zero (00-59).

  s       Displays the second without a leading zero (0-59).

  ss      Displays the second with a leading zero (00-59).

  t       Displays the time using the format given by the ShortTimeFormat
          global variable.

  tt      Displays the time using the format given by the LongTimeFormat
          global variable.

  am/pm   Uses the 12-hour clock for the preceding h or hh specifier, and
          displays 'am' for any hour before noon, and 'pm' for any hour
          after noon. The am/pm specifier can use lower, upper, or mixed
          case, and the result is displayed accordingly.

  a/p     Uses the 12-hour clock for the preceding h or hh specifier, and
          displays 'a' for any hour before noon, and 'p' for any hour after
          noon. The a/p specifier can use lower, upper, or mixed case, and
          the result is displayed accordingly.

  ampm    Uses the 12-hour clock for the preceding h or hh specifier, and
          displays the contents of the TimeAMString global variable for any
          hour before noon, and the contents of the TimePMString global
          variable for any hour after noon.

  /       Displays the date separator character given by the DateSeparator
          global variable.

  :       Displays the time separator character given by the TimeSeparator
          global variable.

  'xx'    Characters enclosed in single or double quotes are displayed as-is,
  "xx"    and do not affect formatting.

  Format specifiers may be written in upper case as well as in lower case
  letters--both produce the same result.

  If the string given by the Format parameter is empty, the date and time
  value is formatted as if a 'c' format specifier had been given.

  The following example:

    S := FormatDateTime('"The meeting is on" dddd, mmmm d, yyyy, ' +
      '"at" hh:mm AM/PM', StrToDateTime('2/15/95 10:30am'));

  assigns 'The meeting is on Wednesday, February 15, 1995 at 10:30 AM' to
  the string variable S. }

function FormatDateTime(const Format: string; DateTime: TDateTime): string;

{ DateTimeToString converts the date and time value given by DateTime using
  the format string given by Format into the string variable given by Result.
  For further details, see the description of the FormatDateTime function. }

procedure DateTimeToString(var Result: string; const Format: string;
  DateTime: TDateTime);

{ System error messages }

function SysErrorMessage(ErrorCode: Integer): string;

{ GetFormatSettings resets all date and number format variables to their
  default values. }

procedure GetFormatSettings;

{ FreeAndNil frees the given TObject instance and sets the variable reference
  to nil.  Be careful to only pass TObjects to this routine. }

procedure FreeAndNil(var Obj);

{ Exception handling routines }

function ExceptObject: TObject;
function ExceptAddr: Pointer;

procedure ShowException(ExceptObject: TObject; ExceptAddr: Pointer);

procedure Abort;

procedure OutOfMemoryError;

procedure Beep;

{ For console applications only. Specifies whether error messages
  should be displayed in a popup window or not }

const
  PopupErrors: Boolean = False;
  IsLibrary: Boolean = False;   // Delphi 32: Declared in the System unit
  InitProc: Pointer = nil;      // Delphi 32: Declared in the System unit

{$R SYSUTILS.RES}

implementation

{$I SYSUTILS.INC}

(* !!!
{$IFNDEF PM_VERSION}
  {$I SYSUTILS.PA1}
{$ENDIF}
*)

{ Utility routines }

procedure DivMod(Dividend: Longint; Divisor: Word; var Result, Remainder: Word); {&USES None} {&FRAME-}
asm
                mov     ecx,Result
                mov     eax,Dividend
                mov     edx,eax
                shr     edx,16
                div     Divisor
                mov     [ecx],ax
                mov     ecx,Remainder
                mov     [ecx],dx
end;

procedure ConvertError(Ident: Integer);
begin
  raise EConvertError.CreateRes(Ident);
end;

procedure ConvertErrorFmt(Ident: Integer; const Args: array of const);
begin
  raise EConvertError.CreateResFmt(Ident, Args);
end;

{ Memory management routines }

function AllocMem(Size: Cardinal): Pointer;
begin
  GetMem(Result, Size);
  FillChar(Result^, Size, 0);
end;

{ String handling routines }

function NewStr(const S: String): PString;
begin
  if S = '' then Result := NullStr else
  begin
    New(Result);
    Result^ := S;
  end;
end;

procedure DisposeStr(P: PString);
begin
  if (P <> nil) and (P^ <> '') then Dispose(P);
end;

procedure AssignStr(var P: PString; const S: string);
var
  Temp: PString;
begin
  Temp := P;
  P := NewStr(S);
  DisposeStr(Temp);
end;

procedure AppendStr(var Dest: string; const S: string);
begin
  Dest := Dest + S;
end;

function UpperCase(const S: string): string;
var
  L: Longint;
begin
  L := Length(S);
  SetLength(Result, L);
  SysChangeCase(Pchar(S), PChar(Result), L, ccUpper);
end;

function LowerCase(const S: string): string;
var
  L: Longint;
begin
  L := Length(S);
  SetLength(Result, L);
  SysChangeCase(Pchar(S), PChar(Result), L, ccLower);
end;

function AnsiUpperCase(const S: string): string;
var
  L: Longint;
begin
  L := Length(S);
  SetLength(Result, L);
  SysChangeCase(Pchar(S), PChar(Result), L, ccAnsiUpper);
end;

function AnsiLowerCase(const S: string): string;
var
  L: Longint;
begin
  L := Length(S);
  SetLength(Result, L);
  SysChangeCase(Pchar(S), PChar(Result), L, ccAnsiLower);
end;

function CompareStr(const S1, S2: string): Integer; {&USES esi,edi} {&FRAME-}
asm
                mov     esi,S1
                mov     edi,S2
                mov     eax,esi
                mov     edx,edi
                test    esi,esi
                jz      @@1
                mov     eax,[esi-4]
              @@1:
                test    edi,edi
                jz      @@2
                mov     edx,[edi-4]
              @@2:
                mov     ecx,eax
                cmp     ecx,edx
                jbe     @@3
                mov     ecx,edx
              @@3:
                cmp     eax,eax
                repe    cmpsb
                je      @@4
                movzx   eax,Byte Ptr [esi-1]
                movzx   edx,Byte Ptr [edi-1]
              @@4:
                sub     eax,edx
end;

function CompareText(const S1, S2: string): Integer; {&USES ebx,esi,edi} {&FRAME-}
asm
                mov     esi,S1
                mov     edi,S2
                mov     eax,esi
                mov     edx,edi
                test    esi,esi
                jz      @@0
                mov     eax,[esi-4]
              @@0:
                test    edi,edi
                jz      @@1
                mov     edx,[edi-4]
              @@1:
                mov     ecx,eax
                cmp     ecx,edx
                jbe     @@2
                mov     ecx,edx
              @@2:
                cmp     eax,eax
              @@3:
                repe    cmpsb
                je      @@6
                mov     bl,[esi-1]
                cmp     bl,'a'
                jb      @@4
                cmp     bl,'z'
                ja      @@4
                sub     bl,'a'-'A'
              @@4:
                mov     bh,[edi-1]
                cmp     bh,'a'
                jb      @@5
                cmp     bh,'z'
                ja      @@5
                sub     bh,'a'-'A'
              @@5:
                cmp     bl,bh
                je      @@3
                movzx   eax,bl
                movzx   edx,bh
              @@6:
                sub     eax,edx
end;

function SameText(const S1, S2: string): Boolean; {&uses none} {&frame-}
asm
                mov     eax,s1
                mov     edx,s2
                cmp     eax,edx
                jz      @@equal
                test    eax,eax
                jz      @@different
                or      edx,edx
                jz      @@different
                mov     ecx,[eax-4]
                cmp     ecx,[edx-4]
                jne     @@different
                push    eax
                push    edx
                call    CompareText
                test    eax,eax
                jnz     @@different
              @@equal:
                mov     al,True
                jmp     @@ret
              @@different:
                mov     al,False
              @@ret:
end;

function CompareMem(p1, p2: Pointer; Length: Integer): Boolean; {&uses esi,edi} {&Frame-}
asm
                mov     esi,p1
                mov     edi,p2
                mov     edx,Length
                mov     ecx,edx
                xor     eax,eax
                and     edx,3
                shr     ecx,1
                shr     ecx,1
                repe    cmpsd
                jne     @@2
                mov     ecx,edx
                repe    cmpsb
                jne     @@2
              @@1:
                inc     eax
              @@2:
end;

function AnsiCompareStr(const S1, S2: string): Integer;
begin
  Result := SysCompareStrings(PChar(S1), PChar(s2), length(s1), length(s2), False);
end;

function AnsiSameStr(const S1, S2: string): Boolean;
begin
  Result := SysCompareStrings(PChar(S1), PChar(s2), length(s1), length(s2), False) = 0;
end;

// Case insensitive compare
function AnsiCompareText(const S1, S2: string): Integer;
begin
  Result := SysCompareStrings(PChar(S1), PChar(s2), length(s1), length(s2), True);
end;

function AnsiSameText(const S1, S2: string): Boolean;
begin
  Result := SysCompareStrings(PChar(S1), PChar(s2), length(s1), length(s2), True) = 0;
end;

// Case sensitive compare
function AnsiStrComp(S1, S2: PChar): Integer;
begin
  Result := SysCompareStrings(S1, S2, strlen(S1), strlen(S2), False);
end;

// Case insensitive compare
function AnsiStrIComp(S1, S2: PChar): Integer;
begin
  Result := SysCompareStrings(S1, S2, strlen(S1), strlen(S2), True);
end;

// Case sensitive compare
function AnsiStrLComp(S1, S2: PChar; MaxLen: Cardinal): Integer;
begin
  if strlen(S1) < MaxLen then MaxLen := strlen(S1);
  if strlen(S2) < MaxLen then MaxLen := strlen(S2);
  Result := SysCompareStrings(S1, S2, MaxLen, MaxLen, False);
end;

function AnsiStrLIComp(S1, S2: PChar; MaxLen: Cardinal): Integer;
begin
  if strlen(S1) < MaxLen then MaxLen := strlen(S1);
  if strlen(S2) < MaxLen then MaxLen := strlen(S2);
  Result := SysCompareStrings(S1, S2, MaxLen, MaxLen, True);
end;

function AnsiStrLower(Str: PChar): PChar;
begin
  Result := SysLowerCase(Str);
end;

function AnsiStrUpper(Str: PChar): PChar;
begin
  Result := SysUpperCase(Str);
end;

function Trim(const S: string): string;
var
  I, L: Integer;
begin
  L := Length(S);
  I := 1;
  while (I <= L) and (S[I] <= ' ') do Inc(I);
  if I > L then Result := '' else
  begin
    while S[L] <= ' ' do Dec(L);
    Result := Copy(S, I, L - I + 1);
  end;
end;

function TrimLeft(const S: string): string;
var
  I, L: Integer;
begin
  L := Length(S);
  I := 1;
  while (I <= L) and (S[I] <= ' ') do Inc(I);
  Result := Copy(S, I, Maxint);
end;

function TrimRight(const S: string): string;
var
  I: Integer;
begin
  I := Length(S);
  while (I > 0) and (S[I] <= ' ') do Dec(I);
  Result := Copy(S, 1, I);
end;

function QuotedStr(const S: string): string;
var
  I: Integer;
begin
  Result := S;
  for I := Length(Result) downto 1 do
    if Result[I] = '''' then Insert('''', Result, I);
  Result := '''' + Result + '''';
end;

function AdjustLineBreaks(const S: string): string;
var
  Source, SourceEnd, Dest: PChar;
  Extra: Integer;
begin
  Source := Pointer(S);
  SourceEnd := Source + Length(S);
  Extra := 0;
  while Source < SourceEnd do
  begin
    case Source^ of
      #10:
        Inc(Extra);
      #13:
        if Source[1] = #10 then Inc(Source) else Inc(Extra);
    end;
    Inc(Source);
  end;
  if Extra = 0 then Result := S else
  begin
    Source := Pointer(S);
    SetString(Result, nil, SourceEnd - Source + Extra);
    Dest := Pointer(Result);
    while Source < SourceEnd do
      case Source^ of
        #10:
          begin
            Dest^ := #13;
            Inc(Dest);
            Dest^ := #10;
            Inc(Dest);
            Inc(Source);
          end;
        #13:
          begin
            Dest^ := #13;
            Inc(Dest);
            Dest^ := #10;
            Inc(Dest);
            Inc(Source);
            if Source^ = #10 then Inc(Source);
          end;
      else
        Dest^ := Source^;
        Inc(Dest);
        Inc(Source);
      end;
  end;
end;

function IsValidIdent(const Ident: string): Boolean;
const
  Alpha = ['A'..'Z', 'a'..'z', '_'];
  AlphaNumeric = Alpha + ['0'..'9'];
var
  I: Integer;
begin
  Result := False;
  if (Length(Ident) = 0) or not (Ident[1] in Alpha) then Exit;
  for I := 2 to Length(Ident) do if not (Ident[I] in AlphaNumeric) then Exit;
  Result := True;
end;

function IntToStr(Value: Integer): string;
begin
  FmtStr(Result, '%d', [Value]);
end;

function IntToHex(Value: Integer; Digits: Integer): string;
begin
  FmtStr(Result, '%.*x', [Digits, Value]);
end;

function StrToInt(const S: string): Integer;
var
  E: Integer;
begin
  Val(S, Result, E);
  if E <> 0 then ConvertErrorFmt(SInvalidInteger, [S]);
end;

function StrToIntDef(const S: string; Default: Integer): Integer;
var
  E: Integer;
begin
  Val(S, Result, E);
  if E <> 0 then Result := Default;
end;

function LoadStr(Ident: Integer): string;
var
  Buffer: array[0..1023] of Char;
begin
  Result := SysLoadResourceString(Ident, Buffer, SizeOf(Buffer));
end;

function FmtLoadStr(Ident: Integer; const Args: array of const): string;
begin
  FmtStr(Result, LoadStr(Ident), Args);
end;

{ File management routines }

function FileOpen(const FileName: string; Mode: Integer): Integer;
var
  RC: Longint;
begin
  RC := SysFileOpen(PChar(FileName), Mode, Result);
  if RC <> 0 then Result := - RC;
end;

function FileCreate(const FileName: string): Integer;
var
  RC: Longint;
begin
  RC := SysFileCreate(PChar(FileName), open_access_ReadWrite, 0, Result );
  if RC <> 0 then Result := - RC;
end;

function FileRead(Handle: Integer; var Buffer; Count: Integer): Integer;
begin
  if SysFileRead(Handle, Buffer, Count, Result) <> 0 then Result := -1;
end;

function FileWrite(Handle: Integer; const Buffer; Count: Integer): Integer;
begin
  if SysFileWrite(Handle, Buffer, Count, Result) <> 0 then Result := -1;
end;

function FileSeek(Handle, Offset, Origin: Integer): Integer;
var
  RC: Longint;
begin
  RC := SysFileSeek(Handle, Offset, Origin, Result);
  if RC <> 0 then Result := -1;
end;

procedure FileClose(Handle: Integer);
begin
  SysFileClose(Handle);
end;

function FileAge(const FileName: string): Integer;
var
  SR: TSearchRec;
begin
  Result := -1;
  if FindFirst(FileName, $3f, SR) = 0 then
    begin
      if SR.Attr and faDirectory = 0 then
        // Ignore directories
        Result := SR.Time;
      FindClose(SR);
    end;
end;

function FileExists(const FileName: string): Boolean;
begin
  Result := FileAge(FileName) <> -1;
end;

function FileGetDate(Handle: Integer): Integer;
begin
  if SysGetFileTime(Handle, Result) <> 0 then
    Result := -1;
end;

function FileSetDate(Handle, Age: Integer): Integer;
begin
  Result := SysSetFileTime(Handle, Age);
end;

function FileGetAttr(const FileName: string): Integer;
begin
  if SysGetFileAttr(PChar(FileName), Result) <> 0 then
    Result := -1;
end;

function FileSetAttr(const FileName: string; Attr: Integer): Integer;
begin
  Result := SysSetFileAttr(PChar(FileName), Attr);
end;

function FindFirst(const Path: String; Attr: Integer; var F: TSearchRec): Integer;
begin
  Result := SysFindFirst(PChar(Path), Attr, TOSSearchRec(F), False );
end;

function FindNext(var F: TSearchRec): Integer;
begin
  Result := SysFindNext(TOSSearchRec(F), False);
end;

procedure FindClose(var F: TSearchRec);
begin
  SysFindClose(TOSSearchRec(F));
end;

function DeleteFile(const FileName: string): Boolean;
begin
  Result := SysFileDelete(PChar(FileName)) = 0;
end;

function RenameFile(const OldName, NewName: string): Boolean;
begin
  Result := SysFileMove(PChar(OldName), PChar(NewName)) = 0;
end;

function ChangeFileExt(const FileName, Extension: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['.', '\', ':']) do Dec(I);
  if (I = 0) or (FileName[I] <> '.') then I := MaxInt;
  Result := Copy(FileName, 1, I - 1) + Extension;
end;

function ExtractFilePath(const FileName: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['\', ':']) do Dec(I);
  Result := Copy(FileName, 1, I);
end;

function ExtractFileDir(const FileName: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['\', ':']) do Dec(I);
  if (I > 1) and (FileName[I] = '\') and
    not (FileName[I - 1] in ['\', ':']) then Dec(I);
  Result := Copy(FileName, 1, I);
end;

function ExtractFileDrive(const FileName: string): string;
var
  I, J: Integer;
begin
  if (Length(FileName) >= 3) and (FileName[2] = ':') then
    Result := Copy(FileName, 1, 2)
  else if (Length(FileName) >= 2) and (FileName[1] = '\') and
    (FileName[2] = '\') then
  begin
    J := 0;
    I := 3;
    While (I < Length(FileName)) and (J < 2) do
    begin
      if FileName[I] = '\' then Inc(J);
      if J < 2 then Inc(I);
    end;
    if FileName[I] = '\' then Dec(I);
    Result := Copy(FileName, 1, I);
  end else Result := '';
end;

function ExtractFileName(const FileName: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['\', ':'] + [SysPathSep]) do
    Dec(I);
  Result := Copy(FileName, I + 1, 255);
end;

function ExtractFileExt(const FileName: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['.', '\', ':']) do Dec(I);
  if (I > 0) and (FileName[I] = '.') then
    Result := Copy(FileName, I, MaxInt) else
    Result := '';
end;

function ExpandFileName(const FileName: string): string;
var
  Buffer: array[0..511] of Char;
begin
  Result := SysFileExpand(Buffer, PChar(FileName));
end;

function ExpandUNCFileName(const FileName: String): String;
var
  Buffer: array[0..511] of Char;
begin
  Result := SysFileUNCExpand(Buffer, PChar(FileName));
end;

function ExtractRelativePath(const BaseName, DestName: string): string;
var
  BasePath, DestPath: string;
  BaseDirs, DestDirs: array[0..129] of PChar;
  BaseDirCount, DestDirCount: Integer;
  I, J: Integer;

  function ExtractFilePathNoDrive(const FileName: string): string;
  begin
    Result := ExtractFilePath(FileName);
    Result := Copy(Result, Length(ExtractFileDrive(FileName)) + 1, MaxInt);
  end;

  procedure SplitDirs(var Path: string; var Dirs: array of PChar;
    var DirCount: Integer);
  var
    I, J: Integer;
  begin
    I := 1;
    J := 0;
    while I <= Length(Path) do
      begin
        if Path[I] = '\' then             { Do not localize }
          begin
            Path[I] := #0;
            Dirs[J] := @Path[I + 1];
            Inc(J);
          end;
        Inc(I);
      end;
    DirCount := J - 1;
  end;

begin
  if AnsiCompareText(ExtractFileDrive(BaseName), ExtractFileDrive(DestName)) = 0 then
  begin
    BasePath := ExtractFilePathNoDrive(BaseName);
    DestPath := ExtractFilePathNoDrive(DestName);
    SplitDirs(BasePath, BaseDirs, BaseDirCount);
    SplitDirs(DestPath, DestDirs, DestDirCount);
    I := 0;
    while (I < BaseDirCount) and (I < DestDirCount) do
    begin
      if AnsiStrIComp(BaseDirs[I], DestDirs[I]) = 0 then
        Inc(I)
      else Break;
    end;
    Result := '';
    for J := I to BaseDirCount - 1 do
      Result := Result + '..\';              { Do not localize }
    for J := I to DestDirCount - 1 do
      Result := Result + DestDirs[J] + '\';  { Do not localize }
    Result := Result + ExtractFileName(DestName);
  end else Result := DestName;
end;

function FileSearch(const Name, DirList: string): string;
var
  I, P, L: Integer;
begin
  Result := Name;
  P := 1;
  L := Length(DirList);
  while True do
  begin
    if FileExists(Result) then Exit;
    while (P <= L) and (DirList[P] = ';') do Inc(P);
    if P > L then Break;
    I := P;
    while (P <= L) and (DirList[P] <> ';') do Inc(P);
    Result := Copy(DirList, I, P - I);
    if not (DirList[P - 1] in [':', '\']) then Result := Result + '\';
    Result := Result + Name;
  end;
  Result := '';
end;

function DiskFree(Drive: Byte): Integer;
begin
  Result := SysDiskFree(Drive);
end;

function DiskSize(Drive: Byte): Integer;
begin
  Result := SysDiskSize(Drive);
end;

function FileDateToDateTime(FileDate: Integer): TDateTime;
begin
  Result :=
    EncodeDate(
      LongRec(FileDate).Hi shr 9 + 1980,
      LongRec(FileDate).Hi shr 5 and 15,
      LongRec(FileDate).Hi and 31) +
    EncodeTime(
      LongRec(FileDate).Lo shr 11,
      LongRec(FileDate).Lo shr 5 and 63,
      LongRec(FileDate).Lo and 31 shl 1, 0);
end;

function DateTimeToFileDate(DateTime: TDateTime): Integer;
var
  Year, Month, Day, Hour, Min, Sec, MSec: Word;
begin
  DecodeDate(DateTime, Year, Month, Day);
  if (Year < 1980) or (Year > 2099) then Result := 0 else
  begin
    DecodeTime(DateTime, Hour, Min, Sec, MSec);
    LongRec(Result).Lo := (Sec shr 1) or (Min shl 5) or (Hour shl 11);
    LongRec(Result).Hi := Day or (Month shl 5) or ((Year - 1980) shl 9);
  end;
end;

function GetCurrentDir: string;
var
  Buffer: array[0..259] of Char;
begin
  if SysDirGetCurrent(0, Buffer) = 0 then
    Result := Buffer
  else
    Result := '';
end;

function SetCurrentDir(const Dir: string): Boolean;
begin
  Result := SysDirSetCurrent(PChar(Dir)) = 0;
end;

function CreateDir(const Dir: string): Boolean;
begin
  Result := SysDirCreate(PChar(Dir)) = 0;
end;

function RemoveDir(const Dir: string): Boolean;
begin
  Result := SysDirDelete(PChar(Dir)) = 0;
end;

{ PChar routines }

function StrLen(Str: PChar): Cardinal; {&USES edi} {&FRAME-}
asm
                cld
                mov     edi,Str
                or      ecx,-1
                xor     eax,eax
                repne   scasb
                sub     eax,ecx
                sub     eax,2
end;

function StrEnd(Str: PChar): PChar; {&USES edi} {&FRAME-}
asm
                cld
                mov     edi,Str
                or      ecx,-1
                mov     al,0
                repne   scasb
                lea     eax,[edi-1]
end;

function StrMove(Dest, Source: PChar; Count: Cardinal): PChar; {&USES esi,edi} {&FRAME-}
asm
                mov     esi,Source
                mov     edi,Dest
                mov     ecx,Count
                mov     edx,edi
                mov     eax,ecx
                cmp     esi,edi
                ja      @@1
                je      @@2
                std
                lea     esi,[esi+ecx-1]
                lea     edi,[edi+ecx-1]
                and     ecx,11b
                shr     eax,2
                rep     movsb
                mov     ecx,eax
                sub     esi,3
                sub     edi,3
                rep     movsd
                cld
                jmp     @@2
              @@1:
                cld
                shr     ecx,2
                and     al,11b
                rep     movsd
                mov     cl,al
                rep     movsb
              @@2:
                mov     eax,edx
end;

function StrCopy(Dest, Source: PChar): PChar; {&USES esi,edi} {&FRAME-}
asm
                cld
                mov     edi,Source
                mov     al,0
                mov     esi,edi
                or      ecx,-1
                repne   scasb
                not     ecx
                mov     dl,cl
                mov     edi,Dest
                mov     eax,edi
                shr     ecx,2
                and     dl,11b
                rep     movsd
                mov     cl,dl
                rep     movsb
end;

function StrECopy(Dest, Source: PChar): PChar; {&USES esi,edi} {&FRAME-}
asm
                cld
                mov     edi,Source
                mov     al,0
                mov     esi,edi
                or      ecx,-1
                repne   scasb
                not     ecx
                mov     al,cl
                mov     edi,Dest
                shr     ecx,2
                and     al,11b
                rep     movsd
                mov     cl,al
                rep     movsb
                lea     eax,[edi-1]
end;

function StrLCopy(Dest, Source: PChar; MaxLen: Cardinal): PChar; {&USES esi,edi} {&FRAME-}
asm
                cld
                mov     edi,Source
                mov     ecx,MaxLen
                mov     eax,Dest
                mov     Byte Ptr [eax],0
                jecxz   @@RET
                mov     al,0
                mov     esi,edi
                mov     edx,ecx
                repne   scasb
                sub     edx,ecx
                mov     edi,Dest
                mov     ecx,edx
                mov     eax,edi
                shr     ecx,2
                and     dl,11b
                rep     movsd
                mov     cl,dl
                rep     movsb
                mov     [edi],cl
              @@RET:
end;

function StrPCopy(Dest: PChar; const Source: String): PChar;
begin
  Result := StrLCopy(Dest, PChar(Source), MaxLongInt);
end;

function StrPLCopy(Dest: PChar; const Source: string;
  MaxLen: Cardinal): PChar;
begin
  Result := StrLCopy(Dest, PChar(Source), MaxLen);
end;

function StrCat(Dest, Source: PChar): PChar;
begin
  StrCopy(StrEnd(Dest), Source);
  Result := Dest;
end;

function StrLCat(Dest, Source: PChar; MaxLen: Cardinal): PChar; {&USES None} {&FRAME+}
asm
                push    Dest
                Call    StrEnd
                mov     ecx,Dest
                add     ecx,MaxLen
                sub     ecx,eax
                jbe     @@1
                push    eax
                push    Source
                push    ecx
                Call    StrLCopy
              @@1:
                mov     eax,Dest
end;

function StrComp(Str1, Str2: PChar): Integer; {&USES esi,edi} {&FRAME-}
asm
                cld
                mov     edi,Str2
                or      ecx,-1
                mov     esi,edi
                xor     eax,eax
                xor     edx,edx
                repne   scasb
                mov     edi,esi
                not     ecx
                mov     esi,Str1
                repe    cmpsb
                mov     al,[esi-1]
                mov     dl,[edi-1]
                sub     eax,edx
end;

function StrIComp(Str1, Str2: PChar): Integer; {&USES esi,edi} {&FRAME-}
asm
                cld
                mov     edi,Str2
                or      ecx,-1
                mov     esi,edi
                xor     eax,eax
                xor     edx,edx
                repne   scasb
                mov     edi,esi
                not     ecx
                mov     esi,Str1
              @@1:
                repe    cmpsb
                je      @@4
                mov     al,[esi-1]
                cmp     al,'a'
                jb      @@2
                cmp     al,'z'
                ja      @@2
                sub     al,'a'-'A'
              @@2:
                mov     dl,[edi-1]
                cmp     dl,'a'
                jb      @@3
                cmp     dl,'z'
                ja      @@3
                sub     dl,'a'-'A'
              @@3:
                sub     eax,edx
                je      @@1
              @@4:
end;

function StrLComp(Str1, Str2: PChar; MaxLen: Cardinal): Integer; {&USES esi,edi} {&FRAME-}
asm
                cld
                mov     edi,Str2
                mov     eax,MaxLen
                mov     esi,edi
                mov     ecx,eax
                jecxz   @@1
                mov     edx,eax
                xor     eax,eax
                repne   scasb
                sub     edx,ecx
                mov     edi,esi
                mov     ecx,edx
                mov     esi,Str1
                repe    cmpsb
                xor     edx,edx
                mov     al,[esi-1]
                mov     dl,[edi-1]
                sub     eax,edx
              @@1:
end;

function StrLIComp(Str1, Str2: PChar; MaxLen: Cardinal): Integer; {&USES esi,edi} {&FRAME-}
asm
                mov     edi,Str2
                mov     eax,MaxLen
                mov     esi,edi
                mov     ecx,eax
                jecxz   @@4
                cld
                mov     edx,eax
                xor     eax,eax
                repne   scasb
                sub     edx,ecx
                mov     ecx,edx
                mov     edi,esi
                mov     esi,Str1
                xor     edx,edx
              @@1:
                repe    cmpsb
                je      @@4
                mov     al,[esi-1]
                cmp     al,'a'
                jb      @@2
                cmp     al,'z'
                ja      @@2
                sub     al,'a'-'A'
              @@2:
                mov     dl,[edi-1]
                cmp     dl,'a'
                jb      @@3
                cmp     dl,'z'
                ja      @@3
                sub     dl,'a'-'A'
              @@3:
                sub     eax,edx
                je      @@1
              @@4:
end;

function StrScan(Str: PChar; Chr: Char): PChar; {&USES edi} {&FRAME-}
asm
                cld
                mov     edi,Str
                mov     edx,edi
                or      ecx,-1
                xor     eax,eax
                repne   scasb
                not     ecx
                mov     al,Chr
                mov     edi,edx
                repne   scasb
                mov     al,0
                jne     @@1
                lea     eax,[edi-1]
              @@1:
end;

function StrRScan(Str: PChar; Chr: Char): PChar; {&USES edi} {&FRAME-}
asm
                cld
                mov     edi,Str
                or      ecx,-1
                xor     eax,eax
                repne   scasb
                not     ecx
                std
                dec     edi
                mov     al,Chr
                repne   scasb
                mov     al,0
                jne     @@1
                lea     eax,[edi+1]
              @@1:
                cld
end;

function StrPos(Str1, Str2: PChar): PChar; {&USES ebx,esi,edi} {&FRAME-}
asm
                cld
                mov     al,0
                mov     edi,Str2
                or      ecx,-1
                repne   scasb
                not     ecx
                dec     ecx
                je      @@2
                mov     edx,ecx
                mov     edi,Str1
                mov     ebx,edi
                or      ecx,-1
                repne   scasb
                not     ecx
                sub     ecx,edx
                jbe     @@2
                mov     edi,ebx
              @@1:
                mov     esi,Str2
                lodsb
                repne   scasb
                jne     @@2
                mov     eax,ecx
                mov     ebx,edi
                mov     ecx,edx
                dec     ecx
                repe    cmpsb
                mov     ecx,eax
                mov     edi,ebx
                jne     @@1
                lea     eax,[edi-1]
                jmp     @@3
              @@2:
                xor     eax,eax
              @@3:
end;

function StrUpper(Str: PChar): PChar; {&USES esi} {&FRAME-}
asm
                cld
                mov     esi,Str
                mov     eax,esi
              @@1:
                mov     dl,[esi]
                test    dl,dl
                jz      @@2
                inc     esi
                cmp     dl,'a'
                jb      @@1
                cmp     dl,'z'
                ja      @@1
                sub     dl,'a'-'A'
                mov     [esi-1],dl
                jmp     @@1
              @@2:
end;

function StrLower(Str: PChar): PChar; {&USES esi} {&FRAME-}
asm
                cld
                mov     esi,Str
                mov     eax,esi
              @@1:
                mov     dl,[esi]
                test    dl,dl
                jz      @@2
                inc     esi
                cmp     dl,'A'
                jb      @@1
                cmp     dl,'Z'
                ja      @@1
                add     dl,'a'-'A'
                mov     [esi-1],dl
                jmp     @@1
              @@2:
end;

function StrPas(Str: PChar): string;
begin
  Result := Str;
end;

function StrAlloc(Size: Cardinal): PChar;
begin
  Inc(Size, SizeOf(Cardinal));
  GetMem(Result, Size);
  Cardinal(Pointer(Result)^) := Size;
  Inc(Result, SizeOf(Cardinal));
end;

function StrBufSize(Str: PChar): Cardinal;
begin
  Dec(Str, SizeOf(Cardinal));
  Result := Cardinal(Pointer(Str)^) - SizeOf(Cardinal);
end;

function StrNew(Str: PChar): PChar;
var
  Size: Cardinal;
begin
  if Str = nil then Result := nil else
  begin
    Size := StrLen(Str) + 1;
    Result := StrMove(StrAlloc(Size), Str, Size);
  end;
end;

function StrPNew(Str: String): PChar;
var
  Size: Cardinal;
begin
  if Str = '' then Result := nil else
  begin
    Size := Length(Str) + 1;
    Result := StrAlloc(Size);
    Move(Str[1], Result^, Size-1);
    Result[Size-1] := #0;
  end;
end;

procedure StrDispose(Str: PChar);
begin
  if Str <> nil then
  begin
    Dec(Str, SizeOf(Cardinal));
    FreeMem(Str, Cardinal(Pointer(Str)^));
  end;
end;

{ String formatting routines }

const
  LongP10Table: array[0..3] of Longint = (10, 100, 1000, 10000);

procedure FormatError(ErrorCode: Integer; Format: PChar; FmtLen: Cardinal);
var
  Buffer: array[0..31] of Char;
begin
  if FmtLen > 31 then FmtLen := 31;
  StrMove(Buffer, Format, FmtLen);
  Buffer[FmtLen] := #0;
  ConvertErrorFmt(SInvalidFormat + ErrorCode, [PChar(@Buffer)]);
end;

procedure FormatClearStr(var S: string);
begin
  S := '';
end;

function FormatBuf(var Buffer; BufLen: Cardinal; const Format; FmtLen: Cardinal;
  const Args: array of const): Cardinal; assembler; {&USES ebx,esi,edi} {&FRAME+}
var
  ArgIndex, Width, Prec: Longint;
  FormatOrg, FormatPtr: PChar;
  JustFlag: Byte;
  StrBuf: array[0..39] of Byte;
asm
                mov     edi,Buffer
                mov     edx,BufLen
                mov     esi,Format
                mov     ecx,FmtLen
                mov     ArgIndex,0
                cld
              @@0:
                test    edx,edx
                jz      @@Done
              @@1:
                dec     ecx
                js      @@Done
                lodsb
                cmp     al,'%'
                je      @@Format
              @@2:
                stosb
                dec     edx
                jne     @@1
              @@Done:
                mov     eax,edi
                sub     eax,Buffer
                jmp     @@RET

              @@Format:
                dec     ecx
                js      @@Done
                lodsb
                cmp     al,'%'
                je      @@2
                lea     ebx,[esi-2]
                mov     FormatOrg,ebx
              @@10:
                mov     JustFlag,al
                cmp     al,'-'
                jne     @@11
                dec     ecx
                js      @@Done
                lodsb
              @@11:
                Call    @@Specifier
                cmp     al,':'
                jne     @@12
                mov     ArgIndex,ebx
                dec     ecx
                js      @@Done
                lodsb
                jmp     @@10
              @@12:
                mov     Width,ebx
                or      ebx,-1
                cmp     al,'.'
                jne     @@20
                dec     ecx
                js      @@Done
                lodsb
                Call    @@Specifier
              @@20:
                mov     Prec,ebx
                mov     FormatPtr,esi
                push    ecx
                push    edx
                Call    @@Convert
                pop     edx
                mov     ebx,Width
                sub     ebx,ecx
                jae     @@21
                xor     ebx,ebx
              @@21:
                cmp     JustFlag,'-'
                jne     @@23
                sub     edx,ecx
                jae     @@22
                add     ecx,edx
                xor     edx,edx
              @@22:
                rep     movsb
              @@23:
                xchg    ebx,ecx
                sub     edx,ecx
                jae     @@24
                add     ecx,edx
                xor     edx,edx
              @@24:
                mov     al,' '
                rep     stosb
                xchg    ebx,ecx
                sub     edx,ecx
                jae     @@25
                add     ecx,edx
                xor     edx,edx
              @@25:
                rep     movsb
                pop     ecx
                mov     esi,FormatPtr
                jmp     @@0
{ Return converted Integer in ebx }
              @@Specifier:
                xor     ebx,ebx
                cmp     al,'*'
                je      @@32
              @@31:
                cmp     al,'0'
                jb      @@34
                cmp     al,'9'
                ja      @@34
                sub     al,'0'
                movzx   eax,al
                lea     ebx,[ebx*4+ebx]
                lea     ebx,[ebx*2+eax] { ebx * 10 + eax }
                dec     ecx
                js      @@35
                lodsb
                jmp     @@31
              @@32:
                mov     eax,ArgIndex
                cmp     eax,Args.Longint[-4]
                ja      @@33
                inc     ArgIndex
                mov     ebx,Args
                cmp     [ebx+eax*8].Byte[4],vtInteger
                mov     ebx,[ebx+eax*8]
                je      @@33
                xor     ebx,ebx
              @@33:
                dec     ecx
                js      @@35
                lodsb
              @@34:
                ret
              @@35:
                pop     eax
                jmp     @@Done
{ Convert one format item }
@@Convert:
                and     al,0DFh
                mov     cl,al
                mov     eax,1
                mov     ebx,ArgIndex
                cmp     ebx,Args.Longint[-4]
                ja      @@ErrorExit
                inc     ArgIndex
                mov     esi,Args
                lea     edx,[esi+ebx*8]
                mov     eax,[edx]
                movzx   ebx,[esi+ebx*8].Byte[4]
                jmp     @@CvtVector[ebx*4].Pointer

@@CvtVector:

                dd      OFFSET @@CvtInteger
                dd      OFFSET @@CvtBoolean
                dd      OFFSET @@CvtChar
                dd      OFFSET @@CvtExtended
                dd      OFFSET @@CvtString
                dd      OFFSET @@CvtPointer
                dd      OFFSET @@CvtPChar
                dd      OFFSET @@CvtObject
                dd      OFFSET @@CvtClass
                dd      OFFSET @@CvtCurrency
                dd      OFFSET @@LongString
@@CvtInteger:

                cmp     cl,'D'
                je      @@41
                cmp     cl,'U'
                je      @@42
                cmp     cl,'X'
                jne     @@CvtError
                mov     ecx,16
                jmp     @@CvtLong
              @@41:
                test    eax,eax
                jns     @@42
                neg     eax
                Call    @@42
                mov     al,'-'
                dec     esi
                inc     ecx
                mov     [esi],al
                ret
              @@42:
                mov     ecx,10
{               jmp     @@CvtLong}

@@CvtLong:

                lea     esi,StrBuf[16]
              @@51:
                xor     edx,edx
                div     ecx
                add     dl,'0'
                cmp     dl,'0'+10
                jb      @@52
                add     dl,'A'-'0'-10
              @@52:
                dec     esi
                mov     [esi],dl
                test    eax,eax
                jne     @@51
                lea     ecx,StrBuf[16]
                sub     ecx,esi
                mov     edx,Prec
                cmp     edx,16
                jb      @@53
                ret
              @@53:
                sub     edx,ecx
                jbe     @@55
                add     ecx,edx
                mov     al,'0'
              @@54:
                dec     esi
                mov     [esi],al
                dec     edx
                jne     @@54
              @@55:
                ret

@@CvtChar:

                cmp     cl,'S'
                jne     @@CvtError
                mov     esi,edx
                mov     ecx,1
                ret

@@CvtString:

                cmp     cl,'S'
                jne     @@CvtError
                mov     esi,eax
                lodsb
                movzx   ecx,al
              @@StrLen:
                cmp     ecx,Prec
                ja      @@61
                ret
              @@61:
                mov     ecx,Prec
              @@62:
                ret

@@LongString:

                cmp     cl,'S'
                jne     @@CvtError
                mov     esi,eax
                mov     ecx,eax
                test    esi,esi
                jz      @@62
                mov     ecx,[esi-4]
                jmp     @@StrLen
@@CvtBoolean:
@@CvtObject:
@@CvtClass:
@@CvtError:
                xor     eax,eax
@@ErrorExit:
                push    eax                     { [1]:Longint = ErrorCode }
                mov     eax,FormatOrg
                push    eax                     { [2]:Longint = Format }
                mov     ecx,FormatPtr
                sub     ecx,eax
                push    ecx                     { [3]:Longint = Format length }
                Call    FormatError

@@CvtPChar:

                cmp     cl,'S'
                jne     @@CvtError
                mov     esi,eax
                push    edi
                mov     edi,eax
                mov     al,0
                mov     ecx,Prec
                jecxz   @@71
                repne   scasb
                jne     @@71
                dec     edi
              @@71:
                mov     ecx,edi
                sub     ecx,esi
                pop     edi
                ret

@@CvtPointer:

                cmp     cl,'P'
                jne     @@CvtError
                mov     Prec,8
                mov     ecx,16
                jmp     @@CvtLong

@@CvtCurrency:
                mov     bh,fvCurrency
                jmp     @@CvtFloat

@@CvtExtended:
                mov     bh,fvExtended

@@CvtFloat:
                mov     esi,eax
                mov     bl,ffGeneral
                cmp     cl,'G'
                je      @@81
                mov     bl,ffExponent
                cmp     cl,'E'
                je      @@81
                mov     bl,ffFixed
                cmp     cl,'F'
                je      @@80
                mov     bl,ffNumber
                cmp     cl,'N'
                je      @@80
                cmp     cl,'M'
                jne     @@CvtError
                mov     bl,ffCurrency
              @@80:
                mov     eax,18
                mov     edx,Prec
                cmp     edx,eax
                jbe     @@82
                mov     edx,2
                cmp     cl,'M'
                jne     @@82
                movzx   edx,CurrencyDecimals
                jmp     @@82
              @@81:
                mov     eax,Prec
                mov     edx,3
                cmp     eax,18
                jbe     @@82
                mov     eax,15
              @@82:
                lea     ecx,StrBuf
                push    ecx                     { [1]:PChar        = Buffer   }
                push    esi                     { [2]:Extended     = @Value   }
                mov     cl,bh
                push    ecx                     { [3]:TFloatValue  = Value    }
                push    ebx                     { [4]:TFloatFormat = Format   }
                push    eax                     { [5]:Integer      = Precision}
                push    edx                     { [6]:Integer      = Digits   }
                Call    FloatToText
                mov     ecx,eax
                lea     esi,StrBuf
                ret
              @@RET:
end;

function StrFmt(Buffer, Format: PChar; const Args: array of const): PChar;
begin
  Buffer[FormatBuf(Buffer^, MaxInt, Format^, StrLen(Format), Args)] := #0;
  Result := Buffer;
end;

function StrLFmt(Buffer: PChar; MaxLen: Cardinal; Format: PChar;
  const Args: array of const): PChar;
begin
  Buffer[FormatBuf(Buffer^, MaxLen, Format^, StrLen(Format), Args)] := #0;
  Result := Buffer;
end;

function Format(const Format: string; const Args: array of const): string;
begin
  FmtStr(Result, Format, Args);
end;

procedure FmtStr(var Result: string; const Format: string;
  const Args: array of const);
var
  Len: Integer;
  Buffer: array[0..4097] of Char;
begin
  Len := FormatBuf(Buffer, SizeOf(Buffer) - 1, Pointer(Format)^,
    Length(Format), Args);
  if Len = SizeOf(Buffer) - 1 then ConvertError(SResultTooLong);
  SetString(Result, Buffer, Len);
end;

{ Floating point conversion routines }

{ Store exponent
  EXPECTS:      al      = Exponent character ('E' or 'e')
                ah      = Positive sign character ('+' or 0)
                bl      = Zero indicator
                ecx     = Minimum number of digits (0..4)
                edx     = Exponent
                edi     = Output buffer }

procedure PutExponent; {&USES None} {&FRAME-}
asm
                stosb
                test    bl,bl
                jne     @@1
                xor     edx,edx
                jmp     @@2
              @@1:
                test    edx,edx
                jge     @@2
                mov     al,'-'
                neg     edx
                jmp     @@3
              @@2:
                test    ah,ah
                jz      @@4
                mov     al,ah
              @@3:
                stosb
              @@4:
                xchg    eax,edx
                push    eax
                mov     ebx,esp
              @@5:
                xor     edx,edx
                div     LongP10Table.Longint    { 10 }
                add     dl,'0'
                mov     [ebx],dl
                inc     ebx
                dec     ecx
                test    eax,eax
                jne     @@5
                test    ecx,ecx
                jg      @@5
              @@6:
                dec     ebx
                mov     al,[ebx]
                stosb
                cmp     ebx,esp
                jne     @@6
                pop     eax
                ret
end;

procedure FloatToDecimal(var Result: TFloatRec; const Value;
  ValueType: TFloatValue; Precision, Decimals: Integer); assembler; {&USES ebx,esi,edi} {&FRAME+}
var
  Exponent: Longint;
  BCDValue: Extended;
  CurValue: Comp;
const
  C1e18: Extended = 1e18;

{ EXPECTS:      esi     = Floating point value  }
{               ebx     = Result TFloatRec      }
{ RETURNS:      ax      = Exponent              }
{               dl      = Sign                  }

procedure ExtendedToDecimal; {&USES None} {&FRAME-}
asm
                mov     ax,[esi].SmallWord[8]
                mov     edx,eax
                and     eax,7FFFh
                jz      @@1
                cmp     eax,7FFFh
                jne     @@10
                cmp     [esi].SmallWord[6],8000h
                je      @@2
                inc     eax
              @@1:
                xor     edx,edx
              @@2:
                mov     [ebx].TFloatRec.Digits.Byte,0
                jmp     @@31
              @@10:
                fld     TByte Ptr [esi]
                sub     eax,3FFFh
                imul    eax,19728
                sar     eax,16
                inc     eax
                mov     Exponent,eax
                sub     eax,18
                neg     eax
                fabs
                Call    FPower10
                frndint
                fld     C1e18
                fcomp
                fnstsw  ax
                test    ah,41h                  { C0 & C3 }
                jz      @@11
                fidiv   LongP10Table.Longint    { 10 }
                inc     Exponent
              @@11:
                lea     edi,[ebx].TFloatRec.Digits.Byte
                fbstp   BCDValue
                mov     edx,9
                fwait
              @@12:
                mov     al,BCDValue[edx-1].Byte
                mov     ah,al
                shr     al,4
                and     ah,0Fh
                add     ax,'00'
                stosw
                dec     edx
                jne     @@12
                mov     al,0
                stosb
              @@20:
                mov     edi,Decimals
                add     edi,Exponent
                jns     @@21
              @@29:
                xor     eax,eax
                jmp     @@1
              @@21:
                cmp     edi,Precision
                jb      @@22
                mov     edi,Precision
              @@22:
                cmp     edi,18
                jae     @@26
                cmp     [ebx].TFloatRec.Digits.Byte[edi],'5'
                jb      @@25
              @@23:
                mov     [ebx].TFloatRec.Digits.Byte[edi],0
                dec     edi
                js      @@24
                inc     [ebx].TFloatRec.Digits.Byte[edi]
                cmp     [ebx].TFloatRec.Digits.Byte[edi],'9'
                ja      @@23
                jmp     @@30
              @@24:
                mov     [ebx].TFloatRec.Digits.SmallWord,'1'
                inc     Exponent
                jmp     @@30
              @@26:
                mov     edi,18
              @@25:
                mov     [ebx].TFloatRec.Digits.Byte[edi],0
                dec     edi
                js      @@29
                cmp     [ebx].TFloatRec.Digits.Byte[edi],'0'
                je      @@25
              @@30:
                test    [esi].Byte[9],80h
                mov     eax,Exponent
                setnz   dl
              @@31:
end;

procedure CurrencyToDecimal; {&USES None} {&FRAME-}
asm
                mov     eax,[esi]
                mov     edx,[esi+4]
                mov     ecx,eax
                or      ecx,edx
                jz      @@Zero
                test    edx,edx
                jns     @@1
                neg     edx
                neg     eax
                sbb     edx,0
              @@1:
                xor     ecx,ecx
                mov     edi,Decimals
                or      edi,edi
                jge     @@2
                xor     edi,edi
              @@2:
                cmp     edi,4
                jl      @@4
                mov     edi,4
              @@3:
                inc     ecx
                sub     eax,0A7640000h  { 0DE0B6B3A7640000h = Comp(1.0e+18) }
                sbb     edx,0DE0B6B3h
                jnc     @@3
                dec     ecx
                add     eax,0A7640000h
                adc     edx,0DE0B6B3h
              @@4:
                mov     CurValue.Longint[0],eax
                mov     CurValue.Longint[4],edx
                fild    CurValue
                mov     edx,edi
                mov     eax,4
                sub     eax,edx
                je      @@5
                fidiv   LongP10Table[eax*4-4].Longint   { 10, 100, 1000 }
              @@5:
                fbstp   BCDValue
                lea     edi,[ebx].TFloatRec.Digits
                fwait
                test    ecx,ecx
                jne     @@7
                mov     ecx,9
              @@6:
                mov     al,BCDValue[ecx-1].Byte
                mov     ah,al
                shr     al,4
                jne     @@9
                mov     al,ah
                and     al,0Fh
                jne     @@10
                dec     ecx
                jne     @@6
                jmp     @@Zero
              @@7:
                mov     al,cl
                add     al,'0'
                stosb
                mov     ecx,9
              @@8:
                mov     al,BCDValue[ecx-1].Byte
                mov     ah,al
                shr     al,4
              @@9:
                add     al,'0'
                stosb
                mov     al,ah
                and     al,0Fh
              @@10:
                add     al,'0'
                stosb
                dec     ecx
                jne     @@8
                mov     eax,edi
                lea     ecx,[ebx+edx].TFloatRec.Digits
                sub     eax,ecx
              @@11:
                mov    [edi].Byte,0
                dec    edi
                cmp    [edi].Byte,'0'
                je     @@11
                test   [esi].Byte[7],80h
                setne  dl
                jmp    @@12
              @@Zero:
                xor    eax,eax
                xor    edx,edx
                mov    [ebx].TFloatRec.Digits.Byte,al
              @@12:
end;

asm
                mov     ebx,Result
                mov     esi,Value
                cmp     ValueType,fvExtended
                je      @@1
                Call    CurrencyToDecimal
                jmp     @@2
              @@1:
                Call    ExtendedToDecimal
              @@2:
                mov     [ebx].TFloatRec.Exponent,ax
                mov     [ebx].TFloatRec.Negative,dl
end;

function FloatToText(Buffer: PChar; const Value; ValueType: TFloatValue;
  Format: TFloatFormat; Precision, Digits: Integer): Integer; assembler; {&USES ebx,esi,edi} {&FRAME+}
var
  FloatRec: TFloatRec;
const
  MoneyFormats: array [0..19, 0..4] of Char =
    ( '$*@@@', '*$@@@', '$ *@@', '* $@@', '($*)@', '-$*@@', '$-*@@', '$*-@@',
      '(*$)@', '-*$@@', '*-$@@', '*$-@@', '-* $@', '-$ *@', '* $-@', '$ *-@',
      '$ -*@', '*- $@', '($ *)', '(* $)');

  NANINF: array [0..5] of Char = 'NANINF';
asm
                mov     eax,19
                cmp     ValueType,fvCurrency
                je      @@2
                mov     eax,Precision
                cmp     eax,2
                jge     @@1
                mov     eax,2
              @@1:
                cmp     eax,18
                jle     @@2
                mov     eax,18
              @@2:
                mov     Precision,eax
                lea     eax,FloatRec
                push    eax                     { [1]:TFloatRec = Result       }
                push    Value                   { [2]:const    = Value         }
                push    ValueType.Longint       { [3]:TFloatValue = Value type }
                push    Precision               { [4]:Integer  = Precision     }
                mov     eax,9999
                cmp     Format,ffFixed          { [5]:Integer  = Decimals      }
                jb      @@3
                mov     eax,Digits
              @@3:
                push    eax
                Call    FloatToDecimal
                mov     edi,Buffer
                cld
                movzx   eax,FloatRec.Exponent
                sub     eax,7FFFh
                cmp     eax,2
                jae     @@4
                lea     esi,NANINF[eax+eax*2]
                movsw
                movsb
                jmp     @@7
              @@4:
                lea     esi,FloatRec.Digits
                movzx   ebx,Format
                cmp     bl,ffExponent
                je      @@6
                cmp     bl,ffCurrency
                ja      @@5
                movsx   eax,FloatRec.Exponent
                cmp     eax,Precision
                jle     @@6
              @@5:
                mov     bl,ffGeneral
              @@6:
                Call    DWord Ptr @@FormatVector[ebx*4]
              @@7:
                mov     eax,edi
                sub     eax,Buffer
                jmp     @@RET

@@FormatVector:

                dd      OFFSET @@PutFGeneral
                dd      OFFSET @@PutFExponent
                dd      OFFSET @@PutFFixed
                dd      OFFSET @@PutFNumber
                dd      OFFSET @@PutFCurrency

{ Get digit or '0' if at end of digit string }

@@GetDigit:

                lodsb
                test    al,al
                jne     @@11
                mov     al,'0'
                dec     esi
              @@11:
                ret

{ Store '-' if number is negative }

@@PutSign:

                cmp     FloatRec.Negative,0
                je      @@21
                mov     al,'-'
                stosb
              @@21:
                ret

{ Convert number using ffGeneral format }

@@PutFGeneral:

                Call    @@PutSign
                movsx   ecx,FloatRec.Exponent
                xor     edx,edx
                cmp     ecx,Precision
                jg      @@31
                cmp     ecx,-3
                jl      @@31
                test    ecx,ecx
                jg      @@32
                mov     al,'0'
                stosb
                cmp     Byte Ptr [esi],0
                je      @@36
                mov     al,DecimalSeparator
                stosb
                neg     ecx
                mov     al,'0'
                rep     stosb
                jmp     @@33
              @@31:
                mov     ecx,1
                inc     edx
              @@32:
                lodsb
                test    al,al
                jz      @@34
                stosb
                loop    @@32
                lodsb
                test    al,al
                jz      @@35
                mov     ah,al
                mov     al,DecimalSeparator
                stosw
              @@33:
                lodsb
                test    al,al
                jz      @@35
                stosb
                jmp     @@33
              @@34:
                mov     al,'0'
                rep     stosb
              @@35:
                test    edx,edx
                je      @@36
                xor     eax,eax
                xor     ecx,ecx
                jmp     @@PutFloatExp
              @@36:
                ret

{ Convert number using ffExponent format }

@@PutFExponent:

                Call    @@PutSign
                Call    @@GetDigit
                mov     ah,DecimalSeparator
                stosw
                mov     ecx,Precision
                dec     ecx
              @@41:
                Call    @@GetDigit
                stosb
                loop    @@41
                mov     ah,'+'
                mov     ecx,Digits
                cmp     ecx,4
                jb      @@PutFloatExp
                xor     ecx,ecx
{               jmp     @@PutFloatExp }

{ Store exponent
  EXPECTS:      ah      = Positive sign character ('+' or 0)
                ecx     = Minimum number of digits (0..4) }

@@PutFloatExp:

                mov     al,'E'
                mov     bl,FloatRec.Digits.Byte
                movsx   edx,FloatRec.Exponent
                dec     edx
                Call    PutExponent
                ret

{ Convert number using ffFixed or ffNumber format }

@@PutFFixed:
@@PutFNumber:

                Call    @@PutSign

{ Store number in fixed point format }

@@PutNumber:

                mov     edx,Digits
                cmp     edx,18
                jb      @@61
                mov     edx,18
              @@61:
                movsx   ecx,FloatRec.Exponent
                test    ecx,ecx
                jg      @@62
                mov     al,'0'
                stosb
                jmp     @@64
              @@62:
                xor     ebx,ebx
                cmp     Format,ffFixed
                je      @@63
                lea     eax,[ecx-1]
                mov     bl,3
                div     bl
                mov     bl,ah
                inc     ebx
              @@63:
                Call    @@GetDigit
                stosb
                dec     ecx
                je      @@64
                dec     ebx
                jne     @@63
                mov     al,ThousandSeparator
                stosb
                mov     bl,3
                jmp     @@63
              @@64:
                test    edx,edx
                je      @@67
                mov     al,DecimalSeparator
                stosb
                jecxz   @@66
                mov     al,'0'
              @@65:
                stosb
                dec     edx
                je      @@67
                inc     ecx
                jne     @@65
              @@66:
                Call    @@GetDigit
                stosb
                dec     edx
                jne     @@66
              @@67:
                ret

{ Convert number using ffCurrency format }

@@PutFCurrency:

                movzx   ebx,CurrencyFormat
                mov     ecx,3
                cmp     FloatRec.Negative,0
                je      @@71
                mov     bl,NegCurrFormat
                mov     ecx,040Fh
              @@71:
                cmp     bl,cl
                jbe     @@72
                mov     bl,cl
              @@72:
                add     bl,ch
                lea     ebx,MoneyFormats[ebx*4+ebx]
                mov     ecx,5
              @@80:
                mov     al,[ebx]
                cmp     al,'@'
                je      @@84
                push    ecx
                push    ebx
                cmp     al,'$'
                je      @@81
                cmp     al,'*'
                je      @@82
                stosb
                jmp     @@83
              @@81:
                Call    @@PutCurSym
                jmp     @@83
              @@82:
                Call    @@PutNumber
              @@83:
                pop     ebx
                pop     ecx
                inc     ebx
                loop    @@80
              @@84:
                ret

{ Store currency symbol string }

@@PutCurSym:

                push    esi
                mov     esi,CurrencyString
                test    esi,esi
                jz      @@85
                mov     ecx,[esi-4]
                rep     movsb
              @@85:
                pop     esi
                ret
              @@RET:
end;

function FloatToTextFmt(Buffer: PChar; const Value; ValueType: TFloatValue;
  Format: PChar): Integer; assembler; {&USES ebx,esi,edi} {&FRAME+}
var
  Section:      Longint;
  DigitCount:   Longint;
  DecimalIndex: Longint;
  FirstDigit:   Longint;
  LastDigit:    Longint;
  DigitPlace:   Longint;
  DigitDelta:   Longint;
  ThousandSep:  Byte;
  Scientific:   Byte;
  FloatRec:     TFloatRec;
asm
                cld
                mov     edi,Value
                mov     ecx,2
                cmp     ValueType,fvExtended
                je      @@Extended
                mov     eax,[edi+4]
                or      eax,[edi+0]
                je      @@1
                mov     ecx,[edi+4]
                shr     ecx,31
                jmp     @@1
              @@Extended:
                movzx   ecx,[edi+8].SmallWord
                mov     eax,[edi+4]
                or      eax,[edi+0]
                or      eax,ecx
                shr     ecx,15
              @@1:
                Call    @@FindSection
                je      @@4
                Call    @@ScanSection
                lea     eax,FloatRec
                push    eax                     { [1]:TFloatRec   = Result    }
                push    Value                   { [2]:const       = Value     }
                push    ValueType.Longint       { [3]:TFloatValue = ValueType }
                mov     eax,DigitCount
                mov     edx,9999
                cmp     Scientific,0
                jne     @@2
                sub     eax,DecimalIndex
                mov     edx,eax
                mov     eax,18
              @@2:
                push    eax                     { [4]:Integer  = Precision    }
                push    edx                     { [5]:Integer  = Decimals     }
                Call    FloatToDecimal
                mov     ax,FloatRec.Exponent
                cmp     ax,8000h
                je      @@4
                cmp     ax,7FFFh
                je      @@4
                cmp     ValueType,fvCurrency
                je      @@5
                cmp     eax,18
                jle     @@5
                cmp     Scientific,0
                jne     @@5
              @@4:
                push    Buffer                  { [1]:PChar     = Buffer      }
                push    Value                   { [2]:const     = Value       }
                push    ValueType.Longint       { [3]:TFloatValue = ValueType }
                push    ffGeneral               { [4]:TFloatFormat = Format   }
                push    15                      { [5]:Integer   = Precision   }
                push    0                       { [6]:Integer   = Digits      }
                Call    FloatToText
                jmp     @@7
              @@5:
                cmp     FloatRec.Digits.Byte,0
                jne     @@6
                mov     ecx,2
                Call    @@FindSection
                je      @@4
                cmp     esi,Section
                je      @@6
                Call    @@ScanSection
              @@6:
                Call    @@ApplyFormat
              @@7:
                jmp     @@RET

{ Find format section
  EXPECTS:      ecx     = Section index
  RETURNS:      esi     = Section offset
                zf      = 1 if section is empty  }

@@FindSection:

                mov     esi,Format
                jecxz   @@12
              @@11:
                lodsb
                cmp     al,"'"
                je      @@14
                cmp     al,'"'
                je      @@14
                test    al,al
                je      @@12
                cmp     al,';'
                jne     @@11
                loop    @@11
                mov     al,[esi]
                test    al,al
                je      @@12
                cmp     al,';'
                jne     @@13
              @@12:
                mov     esi,Format
                mov     al,[esi]
                test    al,al
                je      @@13
                cmp     al,';'
              @@13:
                ret
              @@14:
                mov     ah,al
              @@15:
                lodsb
                cmp     al,ah
                je      @@11
                test    al,al
                jne     @@15
                jmp     @@12

{ Scan format section }

@@ScanSection:

                mov     Section,esi
                mov     ebx,32767
                xor     ecx,ecx
                xor     edx,edx
                or      DecimalIndex,-1
                mov     Scientific,dl
                mov     ThousandSep,dl
              @@21:
                lodsb
              @@22:
                cmp     al,'#'
                je      @@30
                cmp     al,'0'
                je      @@31
                cmp     al,'.'
                je      @@33
                cmp     al,','
                je      @@34
                cmp     al,"'"
                je      @@35
                cmp     al,'"'
                je      @@35
                cmp     al,'E'
                je      @@40
                cmp     al,'e'
                je      @@40
                cmp     al,';'
                je      @@50
                test    al,al
                jne     @@21
                jmp     @@50
              @@30:
                inc     edx
                jmp     @@21
              @@31:
                cmp     edx,ebx
                jge     @@32
                mov     ebx,edx
              @@32:
                inc     edx
                mov     ecx,edx
                jmp     @@21
              @@33:
                cmp     DecimalIndex,-1
                jne     @@21
                mov     DecimalIndex,edx
                jmp     @@21
              @@34:
                mov     ThousandSep,1
                jmp     @@21
              @@35:
                mov     ah,al
              @@36:
                lodsb
                cmp     al,ah
                je      @@21
                test    al,al
                jne     @@36
                jmp     @@50
              @@40:
                lodsb
                cmp     al,'-'
                je      @@41
                cmp     al,'+'
                jne     @@22
              @@41:
                mov     Scientific,1
              @@42:
                lodsb
                cmp     al,'0'
                je      @@42
                jmp     @@22
              @@50:
                mov     DigitCount,edx
                cmp     DecimalIndex,-1
                jne     @@51
                mov     DecimalIndex,edx
              @@51:
                mov     eax,DecimalIndex
                sub     eax,ecx
                jle     @@52
                xor     eax,eax
              @@52:
                mov     LastDigit,eax
                mov     eax,DecimalIndex
                sub     eax,ebx
                jge     @@53
                xor     eax,eax
              @@53:
                mov     FirstDigit,eax
                ret

{ Apply format string }

@@ApplyFormat:

                cmp     Scientific,0
                je      @@61
                mov     eax,DecimalIndex
                xor     edx,edx
                jmp     @@63
              @@61:
                movsx   eax,FloatRec.Exponent
                cmp     eax,DecimalIndex
                jg      @@62
                mov     eax,DecimalIndex
              @@62:
                movsx   edx,FloatRec.Exponent
                sub     edx,DecimalIndex
              @@63:
                mov     DigitPlace,eax
                mov     DigitDelta,edx
                mov     esi,Section
                mov     edi,Buffer
                lea     ebx,FloatRec.Digits
                cmp     FloatRec.Negative,0
                je      @@70
                cmp     esi,Format
                jne     @@70
                mov     al,'-'
                stosb
              @@70:
                lodsb
                cmp     al,'#'
                je      @@80
                cmp     al,'0'
                je      @@80
                cmp     al,'.'
                je      @@70
                cmp     al,','
                je      @@70
                cmp     al,"'"
                je      @@85
                cmp     al,'"'
                je      @@85
                cmp     al,'E'
                je      @@90
                cmp     al,'e'
                je      @@90
                cmp     al,';'
                je      @@100
                test    al,al
                je      @@100
              @@71:
                stosb
                jmp     @@70
              @@80:
                Call    @@PutFmtDigit
                jmp     @@70
              @@85:
                mov     ah,al
              @@86:
                lodsb
                cmp     al,ah
                je      @@70
                test    al,al
                je      @@100
                stosb
                jmp     @@86
              @@90:
                mov     ah,[esi]
                cmp     ah,'+'
                je      @@91
                cmp     ah,'-'
                jne     @@71
                xor     ah,ah
              @@91:
                or      ecx,-1
              @@92:
                inc     ecx
                inc     esi
                cmp     [esi].Byte,'0'
                je      @@92
                cmp     ecx,4
                jb      @@93
                mov     ecx,4
              @@93:
                push    ebx
                mov     bl,FloatRec.Digits.Byte
                movsx   edx,FloatRec.Exponent
                sub     edx,DecimalIndex
                Call    PutExponent
                pop     ebx
                jmp     @@70
              @@100:
                mov     eax,edi
                sub     eax,Buffer
                ret

{ Store formatted digit }

@@PutFmtDigit:

                cmp     DigitDelta,0
                je      @@03
                jl      @@02
              @@01:
                Call    @@03
                dec     DigitDelta
                jne     @@01
                jmp     @@03
              @@02:
                inc     DigitDelta
                mov     eax,DigitPlace
                cmp     eax,FirstDigit
                jle     @@04
                jmp     @@07
              @@03:
                mov     al,[ebx]
                inc     ebx
                test    al,al
                jne     @@05
                dec     ebx
                mov     eax,DigitPlace
                cmp     eax,LastDigit
                jle     @@07
              @@04:
                mov     al,'0'
              @@05:
                cmp     DigitPlace,0
                jne     @@06
                mov     ah,al
                mov     al,DecimalSeparator
                stosw
                jmp     @@07
              @@06:
                stosb
                cmp     ThousandSep,0
                je      @@07
                mov     eax,DigitPlace
                cmp     eax,1
                jle     @@07
                mov     dl,3
                div     dl
                cmp     ah,1
                jne     @@07
                mov     al,ThousandSeparator
                stosb
              @@07:
                dec     DigitPlace
                ret
              @@RET:
end;

function TextToFloat(Buffer: PChar; var Value; ValueType: TFloatValue): Boolean; assembler;
  {&USES ebx,esi,edi} {&FRAME+}
var
  CtrlWord: SmallWord;
  TempLong: Longint;
const
  CWNear: SmallWord = $133F;
asm

                mov     edi,Value
                fstcw   CtrlWord
                fclex
                fldcw   CWNear
                fldz
                mov     esi,Buffer
                Call    @@SkipBlanks
                mov     bl,[esi]
                cmp     bl,'+'
                je      @@1
                cmp     bl,'-'
                jne     @@2
              @@1:
                inc     esi
              @@2:
                mov     ecx,esi
                Call    @@GetDigitStr
                xor     edx,edx
                mov     al,[esi]
                cmp     al,DecimalSeparator
                jne     @@3
                inc     esi
                Call    @@GetDigitStr
                neg     edx
              @@3:
                cmp     ecx,esi
                je      @@Error
                mov     al,[esi]
                and     al,0DFh
                cmp     al,'E'
                jne     @@4
                inc     esi
                push    edx
                Call    @@GetExponent
                pop     eax
                add     edx,eax
              @@4:
                Call    @@SkipBlanks
                cmp     [esi].Byte,0
                jnz     @@Error
                mov     eax,edx
                cmp     ValueType,fvCurrency
                jne     @@5
                add     eax,4
              @@5:
                Call    FPower10
                cmp     bl,'-'
                jne     @@6
                fchs
              @@6:
                cmp     ValueType,fvExtended
                je      @@7
                fistp   QWord Ptr [edi]
                jmp     @@8
              @@7:
                fstp    TByte Ptr [edi]
              @@8:
                fnstsw  ax
                test    al,1001B                { IE & OE }
                jz      @@9
              @@Error:
                fstp    st(0)
                xor     eax,eax
                jmp     @@10
              @@9:
                mov     al,1
              @@10:
                fclex
                fldcw   CtrlWord
                fwait
                jmp     @@RET

{ Skip blanks }

@@SkipBlanks:
                lodsb
                test    al,al
                jz      @@11
                cmp     al,' '
                je      @@SkipBlanks
              @@11:
                dec     esi
                ret

{ Process string of digits
  RETURNS:      edx     = Digit count }

@@GetDigitStr:

                xor     eax,eax
                xor     edx,edx
              @@21:
                mov     al,[esi]
                sub     al,'0'+10
                add     al,10
                jnc     @@22
                fimul   LongP10Table.Longint    { 10 }
                mov     TempLong,eax
                fiadd   TempLong
                inc     esi
                inc     edx
                jmp     @@21
              @@22:
                ret

{ Get exponent
  RETURNS:      edx = Exponent (-4999..4999) }

@@GetExponent:

                xor     eax,eax
                xor     edx,edx
                mov     cl,[esi]
                cmp     cl,'+'
                je      @@31
                cmp     cl,'-'
                jne     @@32
              @@31:
                inc     esi
                jz      @@33
              @@32:
                mov     al,[esi]
                sub     al,'0'+10
                add     al,10
                jnc     @@33
                inc     esi
                imul    edx,10
                add     edx,eax
                cmp     edx,500
                jb      @@32
              @@33:
                cmp     cl,'-'
                jne     @@34
                neg     edx
              @@34:
                ret
              @@RET:
end;

function FloatToStr(Value: Extended): string;
var
  Buffer: array[0..63] of Char;
begin
  SetString(Result, Buffer, FloatToText(Buffer, Value, fvExtended,
    ffGeneral, 15, 0));
end;

function CurrToStr(Value: Currency): string;
var
  Buffer: array[0..63] of Char;
begin
  SetString(Result, Buffer, FloatToText(Buffer, Value, fvCurrency,
    ffGeneral, 0, 0));
end;

function FloatToStrF(Value: Extended; Format: TFloatFormat;
  Precision, Digits: Integer): string;
var
  Buffer: array[0..63] of Char;
begin
  SetString(Result, Buffer, FloatToText(Buffer, Value, fvExtended,
    Format, Precision, Digits));
end;

function CurrToStrF(Value: Currency; Format: TFloatFormat;
  Digits: Integer): string;
var
  Buffer: array[0..63] of Char;
begin
  SetString(Result, Buffer, FloatToText(Buffer, Value, fvCurrency,
    Format, 0, Digits));
end;

function FormatFloat(const Format: string; Value: Extended): string;
var
  Buffer: array[0..255] of Char;
begin
  if Length(Format) > SizeOf(Buffer) - 32 then ConvertError(SFormatTooLong);
  SetString(Result, Buffer, FloatToTextFmt(Buffer, Value, fvExtended,
    PChar(Format)));
end;

function FormatCurr(const Format: string; Value: Currency): string;
var
  Buffer: array[0..255] of Char;
begin
  if Length(Format) > SizeOf(Buffer) - 32 then ConvertError(SFormatTooLong);
  SetString(Result, Buffer, FloatToTextFmt(Buffer, Value, fvCurrency,
    PChar(Format)));
end;

function StrToFloat(const S: string): Extended;
begin
  if not TextToFloat(PChar(S), Result, fvExtended) then
    ConvertErrorFmt(SInvalidFloat, [S]);
end;

function StrToCurr(const S: string): Currency;
begin
  if not TextToFloat(PChar(S), Result, fvCurrency) then
    ConvertErrorFmt(SInvalidFloat, [S]);
end;

{ Date/time support routines }

type
  PDayTable = ^TDayTable;
  TDayTable = array[1..12] of Word;

const
  FMSecsPerDay: Single = MSecsPerDay;
  IMSecsPerDay: Integer = MSecsPerDay;

function DateTimeToTimeStamp(DateTime: TDateTime): TTimeStamp; {&USES None} {&FRAME-}
asm
                mov     ecx,@Result
                fld     DateTime
                fmul    FMSecsPerDay
                sub     esp,8
                fistp   QWord Ptr [esp]
                fwait
                pop     eax
                pop     edx
                test    edx,edx
                jns     @@1
                neg     edx
                neg     eax
                sbb     edx,0
                div     IMSecsPerDay
                neg     eax
                jmp     @@2
              @@1:
                div     IMSecsPerDay
              @@2:
                add     eax,DateDelta
                mov     [ecx].TTimeStamp.Time,edx
                mov     [ecx].TTimeStamp.Date,eax
end;

function TimeStampToDateTime(const TimeStamp: TTimeStamp): TDateTime; {&USES None} {&FRAME-}
asm
                mov     eax,TimeStamp
                mov     ecx,[eax].TTimeStamp.Time
                mov     eax,[eax].TTimeStamp.Date
                sub     eax,DateDelta
                imul    IMSecsPerDay
                test    edx,edx
                jns     @@1
                sub     eax,ecx
                sbb     edx,0
                jmp     @@2
              @@1:
                add     eax,ecx
                adc     edx,0
              @@2:
                push    edx
                push    eax
                fild    QWord Ptr [esp]
                fdiv    FMSecsPerDay
                add     esp,8
end;

function MSecsToTimeStamp(MSecs: Comp): TTimeStamp; {&USES None} {&FRAME-}
asm
                mov     ecx,@Result
                mov     eax,MSecs.Integer[0]
                mov     edx,MSecs.Integer[4]
                div     IMSecsPerDay
                mov     [ecx].TTimeStamp.Time,edx
                mov     [ecx].TTimeStamp.Date,eax
end;

function TimeStampToMSecs(const TimeStamp: TTimeStamp): Comp; {&USES None} {&FRAME-}
asm
                mov     eax,TimeStamp
                fild    [eax].TTimeStamp.Date
                fmul    FMSecsPerDay
                fiadd   [eax].TTimeStamp.Time
end;

{ Time encoding and decoding }

function DoEncodeTime(Hour, Min, Sec, MSec: Word; var Time: TDateTime): Boolean;
begin
  Result := False;
  if (Hour < 24) and (Min < 60) and (Sec < 60) and (MSec < 1000) then
  begin
    Time := (Hour * 3600000 + Min * 60000 + Sec * 1000 + MSec) / MSecsPerDay;
    Result := True;
  end;
end;

function EncodeTime(Hour, Min, Sec, MSec: Word): TDateTime;
begin
  if not DoEncodeTime(Hour, Min, Sec, MSec, Result) then
    ConvertError(STimeEncodeError);
end;

procedure DecodeTime(Time: TDateTime; var Hour, Min, Sec, MSec: Word);
var
  MinCount, MSecCount: Word;
begin
  DivMod(DateTimeToTimeStamp(Time).Time, 60000, MinCount, MSecCount);
  DivMod(MinCount, 60, Hour, Min);
  DivMod(MSecCount, 1000, Sec, MSec);
end;

{ Date encoding and decoding }

function IsLeapYear(Year: Word): Boolean;
begin
  Result := (Year mod 4 = 0) and ((Year mod 100 <> 0) or (Year mod 400 = 0));
end;

function GetDayTable(Year: Word): PDayTable;
const
  DayTable1: TDayTable = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
  DayTable2: TDayTable = (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
  DayTables: array[Boolean] of PDayTable = (@DayTable1, @DayTable2);
begin
  Result := DayTables[IsLeapYear(Year)];
end;

function DoEncodeDate(Year, Month, Day: Word; var Date: TDateTime): Boolean;
var
  I: Integer;
  DayTable: PDayTable;
begin
  Result := False;
  DayTable := GetDayTable(Year);
  if (Year >= 1) and (Year <= 9999) and (Month >= 1) and (Month <= 12) and
    (Day >= 1) and (Day <= DayTable^[Month]) then
  begin
    for I := 1 to Month - 1 do Inc(Day, DayTable^[I]);
    I := Year - 1;
    Date := I * 365 + I div 4 - I div 100 + I div 400 + Day - DateDelta;
    Result := True;
  end;
end;

function EncodeDate(Year, Month, Day: Word): TDateTime;
begin
  if not DoEncodeDate(Year, Month, Day, Result) then
    ConvertError(SDateEncodeError);
end;

procedure DecodeDate(Date: TDateTime; var Year, Month, Day: Word);
const
  D1 = 365;
  D4 = D1 * 4 + 1;
  D100 = D4 * 25 - 1;
  D400 = D100 * 4 + 1;
var
  Y, M, D, I: Word;
  T: Integer;
  DayTable: PDayTable;
begin
  T := DateTimeToTimeStamp(Date).Date;
  if T <= 0 then
  begin
    Year := 0;
    Month := 0;
    Day := 0;
  end else
  begin
    Dec(T);
    Y := 1;
    while T >= D400 do
    begin
      Dec(T, D400);
      Inc(Y, 400);
    end;
    DivMod(T, D100, I, D);
    if I = 4 then
    begin
      Dec(I);
      Inc(D, D100);
    end;
    Inc(Y, I * 100);
    DivMod(D, D4, I, D);
    Inc(Y, I * 4);
    DivMod(D, D1, I, D);
    if I = 4 then
    begin
      Dec(I);
      Inc(D, D1);
    end;
    Inc(Y, I);
    DayTable := GetDayTable(Y);
    M := 1;
    while True do
    begin
      I := DayTable^[M];
      if D < I then Break;
      Dec(D, I);
      Inc(M);
    end;
    Year := Y;
    Month := M;
    Day := D + 1;
  end;
end;

function DayOfWeek(Date: TDateTime): Integer;
begin
  Result := DateTimeToTimeStamp(Date).Date mod 7 + 1;
end;

function Date: TDateTime;
var
  Year, Month, Day: Longint;
begin
  SysGetDateTime(@Year, @Month, @Day, nil, nil, nil, nil, nil);
  Result := EncodeDate(Year, Month, Day);
end;

function Time: TDateTime;
var
  Hour, Minute, Second, mSec: Longint;
begin
  SysGetDateTime(nil, nil, nil, nil, @Hour, @Minute, @Second, @mSec);
  Result := EncodeTime(Hour, Minute, Second, mSec);
end;

function Now: TDateTime;
begin
  Result := Date + Time;
end;

function IncMonth(const Date: TDateTime; NumberOfMonths: Integer): TDateTime;
var
  DayTable: PDayTable;
  Year, Month, Day: Word;
  Sign: Integer;
begin
  if NumberOfMonths >= 0 then Sign := 1 else Sign := -1;
  DecodeDate(Date, Year, Month, Day);
  Year := Year + (NumberOfMonths div 12);
  NumberOfMonths := NumberOfMonths mod 12;
  Inc(Month, NumberOfMonths);
  if Word(Month-1) > 11 then    // if Month <= 0, word(Month-1) > 11)
  begin
    Inc(Year, Sign);
    Inc(Month, -12 * Sign);
  end;
  DayTable := GetDayTable(Year);
  if Day > DayTable^[Month] then Day := DayTable^[Month];
  Result := EncodeDate(Year, Month, Day) + Frac(Date);
end;

function CurrentYear: Word;
var
  Year: Longint;
begin
  SysGetDateTime(@Year, nil, nil, nil, nil, nil, nil, nil);
  Result := Year;
end;

{ Date/time to string conversions }

procedure DateTimeToString(var Result: string; const Format: string;
  DateTime: TDateTime);
var
  BufPos, AppendLevel: Integer;
  Buffer: array[0..255] of Char;

  procedure AppendChars(P: PChar; Count: Integer);
  var
    N: Integer;
  begin
    N := SizeOf(Buffer) - BufPos;
    if N > Count then N := Count;
    if N <> 0 then Move(P[0], Buffer[BufPos], N);
    Inc(BufPos, N);
  end;

  procedure AppendString(const S: string);
  begin
    AppendChars(Pointer(S), Length(S));
  end;

  procedure AppendNumber(Number, Digits: Integer);
  const
    Format: array[0..3] of Char = '%.*d';
  var
    NumBuf: array[0..15] of Char;
  begin
    AppendChars(NumBuf, FormatBuf(NumBuf, SizeOf(NumBuf), Format,
      SizeOf(Format), [Digits, Number]));
  end;

  procedure AppendFormat(Format: PChar);
  var
    Starter, Token, LastToken: Char;
    DateDecoded, TimeDecoded, LastWasHour, Use12HourClock,
    BetweenQuotes: Boolean;
    P: PChar;
    Count: Integer;
    Year, Month, Day, Hour, Min, Sec, MSec, H: Word;

    procedure GetCount;
    var
      P: PChar;
    begin
      P := Format;
      while Format^ = Starter do Inc(Format);
      Count := Format - P + 1;
    end;

    procedure GetDate;
    begin
      if not DateDecoded then
      begin
        DecodeDate(DateTime, Year, Month, Day);
        DateDecoded := True;
      end;
    end;

    procedure GetTime;
    begin
      if not TimeDecoded then
      begin
        DecodeTime(DateTime, Hour, Min, Sec, MSec);
        TimeDecoded := True;
      end;
    end;

  begin
    if (Format <> nil) and (AppendLevel < 2) then
    begin
      Inc(AppendLevel);
      LastToken := ' ';
      DateDecoded := False;
      TimeDecoded := False;
      while Format^ <> #0 do
      begin
        Starter := Format^;
        Inc(Format);
        Token := Starter;
        if Token in ['a'..'z'] then Dec(Token, 32);
        if Token in ['A'..'Z'] then
        begin
          if (Token = 'M') and (LastToken = 'H') then Token := 'N';
          LastToken := Token;
        end;
        case Token of
          'Y':
            begin
              GetCount;
              GetDate;
              if Count <= 2 then
                AppendNumber(Year mod 100, 2) else
                AppendNumber(Year, 4);
            end;
          'M':
            begin
              GetCount;
              GetDate;
              case Count of
                1, 2: AppendNumber(Month, Count);
                3: AppendString(ShortMonthNames[Month]);
              else
                AppendString(LongMonthNames[Month]);
              end;
            end;
          'D':
            begin
              GetCount;
              case Count of
                1, 2:
                  begin
                    GetDate;
                    AppendNumber(Day, Count);
                  end;
                3: AppendString(ShortDayNames[DayOfWeek(DateTime)]);
                4: AppendString(LongDayNames[DayOfWeek(DateTime)]);
                5: AppendFormat(Pointer(ShortDateFormat));
              else
                AppendFormat(Pointer(LongDateFormat));
              end;
            end;
          'H':
            begin
              GetCount;
              GetTime;
              Use12HourClock := False;
              BetweenQuotes := False;
              P := Format;
              while P^ <> #0 do
              begin
                case P^ of
                  'A', 'a':
                    if not BetweenQuotes then
                    begin
                      Use12HourClock := True;
                      Break;
                    end;
                  'H', 'h':
                    Break;
                  '''', '"': BetweenQuotes := not BetweenQuotes;
                end;
                Inc(P);
              end;
              H := Hour;
              if Use12HourClock then
                if H = 0 then H := 12 else if H > 12 then Dec(H, 12);
              if Count > 2 then Count := 2;
              AppendNumber(H, Count);
            end;
          'N':
            begin
              GetCount;
              GetTime;
              if Count > 2 then Count := 2;
              AppendNumber(Min, Count);
            end;
          'S':
            begin
              GetCount;
              GetTime;
              if Count > 2 then Count := 2;
              AppendNumber(Sec, Count);
            end;
          'T':
            begin
              GetCount;
              if Count = 1 then
                AppendFormat(Pointer(ShortTimeFormat)) else
                AppendFormat(Pointer(LongTimeFormat));
            end;
          'A':
            begin
              GetTime;
              P := Format - 1;
              if StrLIComp(P, 'AM/PM', 5) = 0 then
              begin
                if Hour >= 12 then Inc(P, 3);
                AppendChars(P, 2);
                Inc(Format, 4);
              end else
              if StrLIComp(P, 'A/P', 3) = 0 then
              begin
                if Hour >= 12 then Inc(P, 2);
                AppendChars(P, 1);
                Inc(Format, 2);
              end else
              if StrLIComp(P, 'AMPM', 4) = 0 then
              begin
                if Hour < 12 then
                  AppendString(TimeAMString) else
                  AppendString(TimePMString);
                Inc(Format, 3);
              end else
              AppendChars(@Starter, 1);
            end;
          'C':
            begin
              GetCount;
              AppendFormat(Pointer(ShortDateFormat));
              GetTime;
              if (Hour <> 0) or (Min <> 0) or (Sec <> 0) then
              begin
                AppendChars(' ', 1);
                AppendFormat(Pointer(LongTimeFormat));
              end;
            end;
          '/':
            AppendChars(@DateSeparator, 1);
          ':':
            AppendChars(@TimeSeparator, 1);
          '''', '"':
            begin
              P := Format;
              while (Format^ <> #0) and (Format^ <> Starter) do Inc(Format);
              AppendChars(P, Format - P);
              if Format^ <> #0 then Inc(Format);
            end;
        else
          AppendChars(@Starter, 1);
        end;
      end;
      Dec(AppendLevel);
    end;
  end;

begin
  BufPos := 0;
  AppendLevel := 0;
  if Format <> '' then AppendFormat(Pointer(Format)) else AppendFormat('C');
  SetString(Result, Buffer, BufPos);
end;

function DateToStr(Date: TDateTime): string;
begin
  DateTimeToString(Result, ShortDateFormat, Date);
end;

function TimeToStr(Time: TDateTime): string;
begin
  DateTimeToString(Result, LongTimeFormat, Time);
end;

function DateTimeToStr(DateTime: TDateTime): string;
begin
  DateTimeToString(Result, '', DateTime);
end;

function FormatDateTime(const Format: string; DateTime: TDateTime): string;
begin
  DateTimeToString(Result, Format, DateTime);
end;

{ String to date/time conversions }

type
  TDateOrder = (doMDY, doDMY, doYMD);

procedure ScanBlanks(const S: string; var Pos: Integer);
var
  I: Integer;
begin
  I := Pos;
  while (I <= Length(S)) and (S[I] = ' ') do Inc(I);
  Pos := I;
end;

function ScanNumber(const S: string; var Pos: Integer;
  var Number: Word): Boolean;
var
  I: Integer;
  N: Word;
begin
  Result := False;
  ScanBlanks(S, Pos);
  I := Pos;
  N := 0;
  while (I <= Length(S)) and (S[I] in ['0'..'9']) and (N < 1000) do
  begin
    N := N * 10 + (Ord(S[I]) - Ord('0'));
    Inc(I);
  end;
  if I > Pos then
  begin
    Pos := I;
    Number := N;
    Result := True;
  end;
end;

function ScanString(const S: string; var Pos: Integer;
  const Symbol: string): Boolean;
begin
  Result := False;
  if Symbol <> '' then
  begin
    ScanBlanks(S, Pos);
    if AnsiCompareText(Symbol, Copy(S, Pos, Length(Symbol))) = 0 then
    begin
      Inc(Pos, Length(Symbol));
      Result := True;
    end;
  end;
end;

function ScanChar(const S: string; var Pos: Integer; Ch: Char): Boolean;
begin
  Result := False;
  ScanBlanks(S, Pos);
  if (Pos <= Length(S)) and (S[Pos] = Ch) then
  begin
    Inc(Pos);
    Result := True;
  end;
end;

function GetDateOrder(const DateFormat: string): TDateOrder;
var
  I: Integer;
begin
  I := 1;
  while I <= Length(DateFormat) do
  begin
    case Chr(Ord(DateFormat[I]) and $DF) of
      'Y': Result := doYMD;
      'M': Result := doMDY;
      'D': Result := doDMY;
    else
      Inc(I);
      Continue;
    end;
    Exit;
  end;
  Result := doMDY;
end;

function ScanDate(const S: string; var Pos: Integer;
  var Date: TDateTime): Boolean;
var
  DateOrder: TDateOrder;
  I: Integer;
  N1, N2, N3, Y, M, D: Word;
begin
  Result := False;
  DateOrder := GetDateOrder(ShortDateFormat);
  if not (ScanNumber(S, Pos, N1) and ScanChar(S, Pos, DateSeparator) and
    ScanNumber(S, Pos, N2)) then Exit;
  if ScanChar(S, Pos, DateSeparator) then
  begin
    if not ScanNumber(S, Pos, N3) then Exit;
    case DateOrder of
      doMDY: begin Y := N3; M := N1; D := N2; end;
      doDMY: begin Y := N3; M := N2; D := N1; end;
      doYMD: begin Y := N1; M := N2; D := N3; end;
    end;
    if Y <= 99 then Inc(Y, CurrentYear div 100 * 100);
  end else
  begin
    Y := CurrentYear;
    if DateOrder = doDMY then
    begin
      D := N1; M := N2;
    end else
    begin
      M := N1; D := N2;
    end;
  end;
  ScanChar(S, Pos, DateSeparator);
  ScanBlanks(S, Pos);
  Result := DoEncodeDate(Y, M, D, Date);
end;

function ScanTime(const S: string; var Pos: Integer;
  var Time: TDateTime): Boolean;
var
  BaseHour: Integer;
  Hour, Min, Sec: Word;
begin
  Result := False;
  if not ScanNumber(S, Pos, Hour) then Exit;
  Min := 0;
  if ScanChar(S, Pos, TimeSeparator) then
    if not ScanNumber(S, Pos, Min) then Exit;
  Sec := 0;
  if ScanChar(S, Pos, TimeSeparator) then
    if not ScanNumber(S, Pos, Sec) then Exit;
  BaseHour := -1;
  if ScanString(S, Pos, TimeAMString) or ScanString(S, Pos, 'AM') then
    BaseHour := 0
  else
    if ScanString(S, Pos, TimePMString) or ScanString(S, Pos, 'PM') then
      BaseHour := 12;
  if BaseHour >= 0 then
  begin
    if (Hour = 0) or (Hour > 12) then Exit;
    if Hour = 12 then Hour := 0;
    Inc(Hour, BaseHour);
  end;
  ScanBlanks(S, Pos);
  Result := DoEncodeTime(Hour, Min, Sec, 0, Time);
end;

function StrToDate(const S: string): TDateTime;
var
  Pos: Integer;
begin
  Pos := 1;
  if not ScanDate(S, Pos, Result) or (Pos <= Length(S)) then
    ConvertErrorFmt(SInvalidDate, [S]);
end;

function StrToTime(const S: string): TDateTime;
var
  Pos: Integer;
begin
  Pos := 1;
  if not ScanTime(S, Pos, Result) or (Pos <= Length(S)) then
    ConvertErrorFmt(SInvalidTime, [S]);
end;

function StrToDateTime(const S: string): TDateTime;
var
  Pos: Integer;
  Date, Time: TDateTime;
begin
  Pos := 1;
  Time := 0;
  if not ScanDate(S, Pos, Date) or not ((Pos > Length(S)) or
    ScanTime(S, Pos, Time)) then
    ConvertErrorFmt(SInvalidDateTime, [S]);
  if Date >= 0 then
    Result := Date + Time else
    Result := Date - Time;
end;

{ System error messages }

function SysErrorMessage(ErrorCode: Integer): string;
var
  Buffer: array[0..1023] of Char;
  MsgLen: Longint;
  Blank: Boolean;
  C: Char;
  I,J: Longint;
begin
  Result := SysGetSystemError(ErrorCode, Buffer, SizeOf(Buffer), MsgLen);
  { Remove all embedded linefeeds, carriage returns and extra blanks }
  I := 0;
  J := 0;
  Blank := False;
  while MsgLen > 0 do
  begin
    Dec(MsgLen);
    C := Buffer[I];
    Inc(I);
    if C in [' ', #10, #13] then Blank := True
   else
    begin
      if Blank then
      begin
        Buffer[J] := ' ';
        Inc(J);
        Blank := False;
      end;
      Buffer[J] := C;
      Inc(J);
    end;
  end;
  Buffer[J] := #0;
  Result := Buffer;
end;

procedure GetMonthDayNames;
var
  I: Integer;
begin
  for I := 1 to 12 do
  begin
    ShortMonthNames[I] := LoadStr(I + (SShortMonthNames - 1));
    LongMonthNames[I] := LoadStr(I + (SLongMonthNames - 1));
  end;
  for I := 1 to 7 do
  begin
    ShortDayNames[I] := LoadStr(I + (SShortDayNames - 1));
    LongDayNames[I] := LoadStr(I + (SLongDayNames - 1));
  end;
end;

procedure FreeAndNil(var Obj);
var
  P: TObject;
begin
  P := TObject(Obj);
  TObject(Obj) := nil;  // clear the reference before destroying the object
  P.Free;
end;

procedure GetFormatSettings;
var
  Buf1,Buf2,Buf3,Buf4: array[0..259] of Char;
begin
  SysGetCurrencyFormat(Buf1, CurrencyFormat, NegCurrFormat, CurrencyDecimals, ThousandSeparator, DecimalSeparator);
  CurrencyString := Buf1;
  SysGetDateFormat(DateSeparator, Buf1, Buf2);
  ShortDateFormat := Buf1;
  LongDateFormat := Buf2;
  SysGetTimeFormat(TimeSeparator, Buf1, Buf2, Buf3, Buf4);
  TimeAMString := Buf1;
  TimePMString := Buf2;
  ShortTimeFormat := Buf3;
  LongTimeFormat := Buf4;
end;

{ Exception handling routines }

var
  OutOfMemory: EOutOfMemory;

type
  PRaiseFrame = ^TRaiseFrame;
  TRaiseFrame = record
    NextRaise: PRaiseFrame;
    ExceptAddr: Pointer;
    ExceptObject: TObject;
    ExceptionRecord: POSExceptionRecord;
  end;

{ Return current exception object }

function ExceptObject: TObject;
begin
  if RaiseList <> nil then
    Result := PRaiseFrame(RaiseList)^.ExceptObject else
    Result := nil;
end;

{ Return current exception address }

function ExceptAddr: Pointer;
begin
  if RaiseList <> nil then
    Result := PRaiseFrame(RaiseList)^.ExceptAddr else
    Result := nil;
end;

{ Replaces all #10 by #13#10 }

function FormLine(Src,Dest: PChar): Integer;
var
  P,Q: PChar;
begin
  P := Src;
  Q := Dest;
  while P^ <> #0 do
  begin
    if P^ = #10 then
    begin
      Q^ := #13;
      Inc(Q);
    end;
    Q^ := P^;
    Inc(P);
    Inc(Q);
  end;
  Q^ := #0;
  Result := Q - Dest;
end;

{ Display exception message box }

procedure ShowException(ExceptObject: TObject; ExceptAddr: Pointer);
var
  MsgPtr,MsgEnd: PChar;
  MsgLen: Integer;
  LineNo: Longint;
  ModuleName: String;
  FileName: ShortString;
  Temp: AnsiString;
  Buffer,Buffer1: array[0..259] of Char;
const
  Showing: Boolean = False;
begin
  if Showing then
    Exit;
  Showing := True;
  Buffer1[0] := #0;
  if GetLocationInfo(ExceptAddr, FileName, LineNo) <> nil then
    StrFmt(Buffer1, 'File: %s Line#: %d'#10, [FileName, LineNo]);
  ModuleName := SysGetModuleName(ExceptAddr, Buffer, SizeOf(Buffer));
  MsgPtr := '';
  MsgEnd := '';
  if ExceptObject is Exception then
    begin
      MsgPtr := PChar(Exception(ExceptObject).Message);
      MsgLen := StrLen(MsgPtr);
      if (MsgLen <> 0) and (MsgPtr[MsgLen - 1] <> '.') then
        MsgEnd := '.';
    end;
  Temp := LoadStr(SException);
  StrFmt(Buffer, PChar(Temp), [ExceptObject.ClassName, ModuleName, ExceptAddr, Buffer1, MsgPtr, MsgEnd]);
  Temp := LoadStr(SExceptTitle);
  FormLine(Buffer, Buffer1);

  if IsConsole then
    SysDisplayConsoleError(PopupErrors, PChar(Temp), Buffer1)
  else
    SysDisplayGUIError(PChar(Temp), Buffer1);
  Showing := False;
end;

{ Get return address of caller }

function ReturnAddr: Pointer; {&USES None} {&FRAME-}
asm
                mov     eax,[ebp+4]
end;

{ Raise abort exception }

{$W+}

procedure Abort;
begin
  raise EAbort.CreateRes(SOperationAborted) at ReturnAddr;
end;

{ Raise out of memory exception }

procedure OutOfMemoryError;
begin
  raise OutOfMemory at ReturnAddr;
end;

{$W-}

{ Exception class }

constructor Exception.Create(const Msg: string);
begin
  FMessage := Msg;
end;

constructor Exception.CreateFmt(const Msg: string;
  const Args: array of const);
begin
  FMessage := Format(Msg, Args);
end;

constructor Exception.CreateRes(Ident: Integer);
begin
  FMessage := LoadStr(Ident);
end;

constructor Exception.CreateResFmt(Ident: Integer;
  const Args: array of const);
begin
  FMessage := Format(LoadStr(Ident), Args);
end;

constructor Exception.CreateHelp(const Msg: string; AHelpContext: Integer);
begin
  FMessage := Msg;
  FHelpContext := AHelpContext;
end;

constructor Exception.CreateFmtHelp(const Msg: string; const Args: array of const;
  AHelpContext: Integer);
begin
  FMessage := Format(Msg, Args);
  FHelpContext := AHelpContext;
end;

constructor Exception.CreateResHelp(Ident: Integer; AHelpContext: Integer);
begin
  FMessage := LoadStr(Ident);
  FHelpContext := AHelpContext;
end;

constructor Exception.CreateResFmtHelp(Ident: Integer;
  const Args: array of const; AHelpContext: Integer);
begin
  FMessage := Format(LoadStr(Ident), Args);
  FHelpContext := AHelpContext;
end;

{ EOutOfMemory class }

destructor EOutOfMemory.Destroy;
begin
end;

procedure EOutOfMemory.FreeInstance;
begin
  if AllowFree then
    inherited FreeInstance;
end;

{ Create I/O exception }

function CreateInOutError: EInOutError;
type
  TErrorRec = record
    Code: Byte;
    Ident: Word;
  end;
const
  ErrorMap: array[0..6] of TErrorRec = (
    (Code: 2; Ident: SFileNotFound),
    (Code: 3; Ident: SInvalidFilename),
    (Code: 4; Ident: STooManyOpenFiles),
    (Code: 5; Ident: SAccessDenied),
    (Code: 100; Ident: SEndOfFile),
    (Code: 101; Ident: SDiskFull),
    (Code: 106; Ident: SInvalidInput));
var
  I: Integer;
begin
  I := Low(ErrorMap);
  while (I <= High(ErrorMap)) and (ErrorMap[I].Code <> InOutRes) do Inc(I);
  if I <= High(ErrorMap) then
    Result := EInOutError.CreateRes(ErrorMap[I].Ident) else
    Result := EInOutError.CreateResFmt(SInOutError, [InOutRes]);
  Result.ErrorCode := InOutRes;
  InOutRes := 0;
end;

{ RTL error handler }

type
  TExceptRec = record
    EClass: ExceptClass;
    EIdent: Word;
  end;
const
  ExceptMap: array[2..15] of TExceptRec = (
    (EClass: EInvalidPointer; EIdent: SInvalidPointer),
    (EClass: EDivByZero; EIdent: SDivByZero),
    (EClass: ERangeError; EIdent: SRangeError),
    (EClass: EIntOverflow; EIdent: SIntOverflow),
    (EClass: EInvalidOp; EIdent: SInvalidOp),
    (EClass: EZeroDivide; EIdent: SZeroDivide),
    (EClass: EOverflow; EIdent: SOverflow),
    (EClass: EUnderflow; EIdent: SUnderflow),
    (EClass: EInvalidCast; EIdent: SInvalidCast),
    (EClass: EAccessViolation; EIdent: SAccessViolation),
    (EClass: EPrivilege; EIdent: SPrivilege),
    (EClass: EControlC; EIdent: SControlC),
    (EClass: EStackOverflow; EIdent: SStackOverflow),
    (EClass: EExternalException; EIdent: SExternalException));

procedure ErrorHandler(ErrorCode: Integer; ErrorAddr: Pointer);
var
  E: Exception;
begin
  case ErrorCode of
    1: E := OutOfMemory;
    2..15: with ExceptMap[ErrorCode] do E := EClass.CreateRes(EIdent);
  else
    E := CreateInOutError;
  end;
  raise E at ErrorAddr;
end;

function MapException(P: POSExceptionRecord):Byte;
begin
  case P^.fExceptionNum of
    xcpt_Integer_Divide_By_Zero:  Result := 3;
    xcpt_Array_Bounds_Exceeded:   Result := 4;
    xcpt_Integer_Overflow:        Result := 5;
    xcpt_Float_Inexact_Result,
    xcpt_Float_Invalid_Operation,
    xcpt_Float_Stack_Check:       Result := 6;
    xcpt_Float_Divide_By_Zero:    Result := 7;
    xcpt_Float_Overflow:          Result := 8;
    xcpt_Float_Underflow,
    xcpt_Float_Denormal_Operand:  Result := 9;
    {$IFDEF LINUX}
    xcpt_In_Page_Error,
    {$ENDIF}
    {$IFDEF DPMI32} // Exception E
    xcpt_In_Page_Error,
    {$ENDIF}
    xcpt_Access_Violation:        Result := 11;
    xcpt_Privileged_Instruction:  Result := 12;
    xcpt_Signal_Ctrl_C:           Result := 13;
    xcpt_Unable_To_Grow_Stack:    Result := 14;
  else                            Result := 15;
  end;
end;

function GetExceptionClass(P: POSExceptionRecord): ExceptClass;
var
  ErrorCode: Byte;
begin
  ErrorCode := MapException(P);
  Result := ExceptMap[ErrorCode].EClass;
end;

function GetExceptionObject(P: POSExceptionRecord): Exception;
var
  ErrorCode: Integer;
  AccessOp: Integer; // string ID indicating the access type READ or WRITE
  AccessAddress: Pointer;
begin
  ErrorCode := MapException(P);
  case ErrorCode of
    3..10,12..14:
      with ExceptMap[ErrorCode] do Result := EClass.CreateRes(EIdent);
    11:
      begin
        with P^ do
        begin
          {$IfDef Win32}
          if fExceptionInfo[0] = 0 {xcpt_Read_Access} then
          {$Else}
          { OS/2, DPMI32 and Linux:
            hardware 0:unknown 1:read 2:write 4:execute 8:space 16:limit }
          if (fExceptionInfo[0] and $2) = 0 then
          {$EndIf}
            AccessOp := sReadAccess
          else
            AccessOp := sWriteAccess;
          AccessAddress := Pointer(fExceptionInfo[1]);
          Result := EAccessViolation.CreateResFmt(sAccessViolation,
            [fExceptionAddress, LoadStr(AccessOp), AccessAddress]);
        end;
      end;
  else
    Result := EExternalException.CreateResFmt(SExternalException,
      [P^.fExceptionNum]);
    EExternalException(Result).ExceptionRecord := P;
  end;
end;

{ RTL exception handler }

procedure ExceptHandler(ExceptObject: TObject; ExceptAddr: Pointer);
begin
  ShowException(ExceptObject, ExceptAddr);
  Halt(1);
end;

procedure InitExceptions;
begin
  OutOfMemory := EOutOfMemory.CreateRes(SOutOfMemory);
  ErrorProc := @ErrorHandler;
  ExceptProc := @ExceptHandler;
  ExceptionClass := Exception;
  ExceptClsProc := @GetExceptionClass;
  ExceptObjProc := @GetExceptionObject;
end;

procedure Beep;
begin
  SysBeep;
end;

procedure InitPlatformId;
begin
  Win32Platform := SysPlatformID;
end;

initialization
  InitExceptions;
  GetMonthDayNames;
  GetFormatSettings;
  InitPlatformId;
end.
