unit _Action;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, SysUtils, FtnAddr, MsgAPI, _Hook, _BSO;

////////////////////////////////////////////////////////////////
//
// Action
//
////////////////////////////////////////////////////////////////

type
  PAction = ^TAction;
  TAction = object (TObject)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  end; { TAction }

  PRewriteAction = ^TRewriteAction;
  TRewriteAction = object (TAction)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  private
    pat_from: String;
    pat_to  : String;
    pat_subj: String;
    pat_orig: TAddress;
    pat_dest: TAddress;
    pat_attr: TAttrMask;
  end; { TRewriteAction }

  PDeleteAction = ^TDeleteAction;
  TDeleteAction = object (TAction)
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  end; { TDeleteAction }

  PBounceAction = ^TBounceAction;
  TBounceAction = object (TAction)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  private
    Template: String;
  end; { TBounceAction }

  PSaveAction = ^TSaveAction;
  TSaveAction = object (TAction)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  private
    FileName: String;
  end; { TSaveAction }

  PCopyAction = ^TCopyAction;
  TCopyAction = object (TAction)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  private
    target_area: String;
    target_path: String;
  end; { TCopyAction }

  PForwardAction = ^TForwardAction;
  TForwardAction = object (TAction)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  private
    target_area: String;
    pat_from: String;
    pat_to  : String;
    pat_subj: String;
    pat_orig: TAddress;
    pat_dest: TAddress;
    pat_attr: TAttrMask;
  end; { TForwardAction }

  TRouteModifier = (route_raw, route_hub, route_host, route_boss);

  PRouteAction = ^TRouteAction;
  TRouteAction = object (TAction)
    procedure Eat( const S: String ); virtual;
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  private
    route_to: TAddress;
    modifier: TRouteModifier;
    flavor  : TBsoFlavor;
  end; { TRouteAction }

  PKillSentAction = ^TKillSentAction;
  TKillSentAction = object (TAction)
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  end; { TKillSentAction }

  PIgnoreAction = ^TIgnoreAction;
  TIgnoreAction = object (TAction)
    function Apply( M: PFtnMessage ) : Boolean; virtual;
  end; { TIgnoreAction }

  EIgnoreAction = class ( Exception );

function ConstructAction( const S: String ) : PAction;

{ =================================================================== }

implementation

uses
  MyLib, _Config, _MapFile;

const
  STREAM_BUFFER_SIZE = 32 * 1024;

{ --------------------------------------------------------- }
{ CopyMsgBody                                               }
{ --------------------------------------------------------- }

procedure CopyMsgBody( I, O: PFtnMessage; Printable: Boolean );

  procedure DoCopy( V: VString ); far;
  begin
    if Printable then
      O^.Insert( AllocStr( MakePrintable(Invalidate(V^)) ))
    else
      O^.Insert( AllocStr(V^) );
  end; { DoCopy }

begin
  I^.ForEach( @DoCopy );
end; { CopyMsgBody }

////////////////////////////////////////////////////////////////
//
// Action
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TAction.Eat( const S: String );
begin
end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TAction.Apply( M: PFtnMessage ) : Boolean;
begin
  Abstract;
end; { Apply }

////////////////////////////////////////////////////////////////
//
// RewriteAction
//
// Rewrite <from>[, <orig>[, <to>[, <dest>[, <subj> ]]]]
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TRewriteAction.Eat( const S: String );
const
  any = '*';
var
  j: Integer;
  item: String;
begin
  pat_from := any;
  pat_to   := any;
  pat_subj := any;
  pat_orig.Assign( any );
  pat_dest.Assign( any );

  for j := 1 to 6 do
  begin
    item := Trim( ExtractWord( j, S, _COMMA ) );
    if item = '' then break;
    case j of
      1: pat_from := item;
      2: pat_orig.Assign( item );
      3: pat_to   := item;
      4: pat_dest.Assign( item );
      5: pat_subj := item;
      6: pat_attr.Assign( item );
    end;
  end;

end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TRewriteAction.Apply( M: PFtnMessage ) : Boolean;
var
  A: TAddress;
begin
  A := M^.GetOrig;
  pat_orig.Apply( A );
  M^.SetOrig( A );

  A := M^.GetDest;
  pat_dest.Apply( A );
  M^.SetDest( A );

  M^.SetFrom( ReplaceSubstr(pat_from, '*', M^.GetFrom) );
  M^.SetTo  ( ReplaceSubstr(pat_to,   '*', M^.GetTo) );
  M^.SetSubj( ReplaceSubstr(pat_subj, '*', M^.GetSubj) );

  M^.SetAttr( pat_attr.Apply(M^.GetAttr) );

  Log( ll_Action, 'Message header changed' );
  Result := True;
end; { Apply }

////////////////////////////////////////////////////////////////
//
// DeleteAction
//
// Delete
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TDeleteAction.Apply( M: PFtnMessage ) : Boolean;
begin
  Log( ll_Action, 'Killing message' );
  M^.Kill;
  Result := True;
end; { Apply }

////////////////////////////////////////////////////////////////
//
// BounceAction
//
// Bounce <template_name>
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TBounceAction.Eat( const S: String );
begin
  CheckFile( S );
  Template := S;
end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TBounceAction.Apply( M: PFtnMessage ) : Boolean;
var
  S: String;
  Map: TMappedFile;
  Out: PNetmailMessage;
begin
  Result := False;
  Log( ll_Action, 'Bouncing message using template ' + Template );

  Map.Init( Template );

  Out := New( PNetmailMessage, Init );

  with Out^ do
  begin
    SetArea( _NETMAIL );
    SetOrig( Cfg.PrimaryAddr );
    SetDest( M^.GetOrig );
    SetFrom( Cfg.RobotName );
    SetTo  ( M^.GetFrom );
    SetDate( SysUtils.Now );
    SetAttr( ATTR_LOCAL );
    SetSubj( 'Bounced message' );
  end;

  while Map.GetLine( S ) do
  begin
    if Pos( '%BODY', S ) <> 0 then
      CopyMsgBody( M, Out, True )
    else
    begin
      ReplaceSubstr( S, '%PID',   PID );
      ReplaceSubstr( S, '%NOW',   MsgDateStr(Out^.GetDate) );
      ReplaceSubstr( S, '%ATTR',  AttrStr(M^.GetAttr) );
      ReplaceSubstr( S, '%DATE',  MsgDateStr(M^.GetDate) );
      ReplaceSubstr( S, '%AREA',  M^.GetArea );
      ReplaceSubstr( S, '%SUBJ',  M^.GetSubj );
      ReplaceSubstr( S, '%ORIG',  M^.GetOrig.MakeStr );
      ReplaceSubstr( S, '%DEST',  M^.GetDest.MakeStr );
      ReplaceSubstr( S, '%FROM',  M^.GetFrom );
      ReplaceSubstr( S, '%TO',    M^.GetTo );
      ReplaceSubstr( S, '%FFROM', ExtractWord( 1, M^.GetFrom, _BLANK ));
      ReplaceSubstr( S, '%FTO',   ExtractWord( 1, M^.GetTo, _BLANK ));
      Out^.Insert( AllocStr(S) );
    end;
  end;

  Map.Done;

  Out^.SaveAs( NewMsgFileName(Cfg.Netmail) );
  Destroy( Out );
end; { Apply }

////////////////////////////////////////////////////////////////
//
// SaveAction
//
// Save <file_name>
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TSaveAction.Eat( const S: String );
begin
  FileName := S;
end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TSaveAction.Apply( M: PFtnMessage ) : Boolean;
var
  F: Text;
begin
  Result := False;
  Log( ll_Action, 'Saving message in ' + FileName );
  Assign( F, FileName );
  try
    Append( F );
  except
    try
      Rewrite( F );
    except
      Log( ll_Error, 'Could not write the file "' + FileName + '"' );
      Exit;
    end;
  end;

  Writeln( F, ^M^J' * Message saved at ', MsgDateStr(SysUtils.Now), ' by ', PID, ^M^J );
  M^.TxtSave( F );
  Writeln( F, CharStr( #249, 75 ) );
  Close( F );
end; { Apply }

////////////////////////////////////////////////////////////////
//
// CopyAction
//
// Copy {<echotag> | *} [<alternate_path>]
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TCopyAction.Eat( const S: String );
begin
  target_area := ExtractWord( 1, S, _BLANK );
  target_path := ExtractWord( 2, S, _BLANK );
  if target_area = '' then
    raise Exception.Create( 'CopyAction: target area missed' );
end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TCopyAction.Apply( M: PFtnMessage ) : Boolean;
var
  Name: String;
  Area: String;
  out_msg: PFtnMessage;
  out_pkt: PPacket;
begin
  Result := False;
  Log( ll_Action, 'Copying message into ' + target_area );
  if target_path <> '' then
    Log( ll_Action, 'Using alternate path ' + target_path );

  out_msg := nil;
  out_pkt := nil;

  if target_area = '*' then
    Area := M^.GetArea
  else
    Area := target_area;

  if is_netmail( Area ) then
    out_msg := New( PNetmailMessage, Init )
  else
  begin
    out_pkt := New( PPacket, Init );
    out_msg := New( PPackedMessage, Init );
    with out_pkt^ do
    begin
      SetOrig( Cfg.RobotAddr );
      SetDest( Cfg.PrimaryAddr );
      SetDate( SysUtils.Now );
      Insert( out_msg );
    end;
  end;

  out_msg^.Clone( M );

  out_msg^.SetArea( Area );

  if not out_msg^.HasTearline then
    out_msg^.SetTearline( PID );

  if out_msg^.Echomail and not out_msg.HasOrigin then
    out_msg^.SetOrigin( Cfg.GetOrigin );

  if out_pkt <> nil then
  begin
    if target_path <> '' then
      Name := NewPktFileName( target_path )
    else
      Name := NewPktFileName( Cfg.Inbound );

    out_pkt^.SaveAs( Name );
    Destroy( out_pkt );
  end
  else
  begin
    if target_path <> '' then
      Name := NewMsgFileName( target_path )
    else
      Name := NewMsgFileName( Cfg.Netmail );
    PNetmailMessage( out_msg )^.SaveAs( Name );
    Destroy( out_msg );
  end;
end; { Apply }

////////////////////////////////////////////////////////////////
//
// ForwardAction
//
// Forward <echotag> [, <from>, <orig>, <to>, <dest>, <subj>, <attr> ]
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TForwardAction.Eat( const S: String );
const
  any = '*';
var
  j: Integer;
  item: String;
begin
  target_area := ExtractWord( 1, S, _COMMA );
  pat_from := any;
  pat_to   := any;
  pat_subj := any;
  pat_orig.Assign( any );
  pat_dest.Assign( any );

  if target_area = '' then
    raise Exception.Create( 'ForwardAction: target area missed' );

  for j := 2 to 7 do
  begin
    item := Trim( ExtractWord( j, S, _COMMA ) );
    if item = '' then break;
    case j of
      2: pat_from := item;
      3: pat_orig.Assign( item );
      4: pat_to   := item;
      5: pat_dest.Assign( item );
      6: pat_subj := item;
      7: pat_attr.Assign( item );
    end;
  end;

end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TForwardAction.Apply( M: PFtnMessage ) : Boolean;

  function has_fwd( V: VString ) : Boolean; far;
  begin
    Result := Pos( ^A'FWD', V^ ) <> 0;
  end; { has_fwd }

var
  S: String;
  A: TAddress;
  Area   : String;
  out_msg: PFtnMessage;
  out_pkt: PPacket;
begin
  Result := False;
  Log( ll_Action, 'Forwarding message into ' + target_area );

  out_msg := nil;
  out_pkt := nil;

  if target_area = '*' then
    Area := M^.GetArea
  else
    Area := target_area;

  if is_netmail( Area ) then
    out_msg := New( PNetmailMessage, Init )
  else
  begin
    out_pkt := New( PPacket, Init );
    out_msg := New( PPackedMessage, Init );
    with out_pkt^ do
    begin
      SetOrig( Cfg.RobotAddr );
      SetDest( Cfg.PrimaryAddr );
      SetDate( SysUtils.Now );
      Insert( out_msg );
    end;
  end;

  out_msg^.SetArea( Area );

  A := M^.GetOrig;
  pat_orig.Apply( A );
  out_msg^.SetOrig( A );

  A := M^.GetDest;
  pat_dest.Apply( A );
  out_msg^.SetDest( A );

  out_msg^.SetFrom( ReplaceSubstr(pat_from, '*', M^.GetFrom) );
  out_msg^.SetTo  ( ReplaceSubstr(pat_to,   '*', M^.GetTo) );
  out_msg^.SetSubj( ReplaceSubstr(pat_subj, '*', M^.GetSubj) );
  out_msg^.SetDate( M^.GetDate );

  out_msg^.SetAttr( pat_attr.Apply(M^.GetAttr) );

  if CompAddr( out_msg^.GetOrig, Cfg.PrimaryAddr ) = 0 then
    out_msg^.SetAttr( out_msg^.GetAttr or ATTR_LOCAL )
  else
    out_msg^.SetAttr( out_msg^.GetAttr or ATTR_TRANSIT );

  if M^.FirstThat( @has_fwd ) = nil then
  begin
    if M^.GetFrom <> '' then
      out_msg^.Insert( AllocStr( ^A'FWDFROM ' + M^.GetFrom ) );
    if M^.GetTo <> '' then
      out_msg^.Insert( AllocStr( ^A'FWDTO ' + M^.GetTo ) );
    if M^.GetOrig.Valid then
      out_msg^.Insert( AllocStr( ^A'FWDORIG ' + M^.GetOrig.MakeStr ) );
    if M^.GetDest.Valid then
      out_msg^.Insert( AllocStr( ^A'FWDDEST ' + M^.GetDest.MakeStr ) );
    if M^.GetSubj <> '' then
      out_msg^.Insert( AllocStr( ^A'FWDSUBJ ' + M^.GetSubj ) );
    if M^.Echomail then
      out_msg^.Insert( AllocStr( ^A'FWDAREA ' + M^.GetArea ) );
    if M^.GetMsgID <> '' then
      out_msg^.Insert( AllocStr( ^A'FWDMSGID ' + M^.GetMsgID ) );
  end;

  CopyMsgBody( M, out_msg, False );

  out_msg^.SetTearline( 'Forwarded by ' + PID );

  if CompAddr( out_msg^.GetOrig, M^.GetOrig ) = 0 then
    out_msg^.SetOrigin( M^.GetOrigin )
  else
    out_msg^.SetOrigin( Cfg.GetOrigin );

  if out_pkt <> nil then
  begin
    out_pkt^.SaveAs( NewPktFileName( Cfg.Inbound ) );
    Destroy( out_pkt );
  end
  else
  begin
    PNetmailMessage( out_msg )^.SaveAs( NewMsgFileName( Cfg.Netmail ));
    Destroy( out_msg );
  end;
end; { Apply }

////////////////////////////////////////////////////////////////
//
// TRouteAction
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TRouteAction.Eat( const S: String );
var
  t: String;
begin
  flavor := WhatFlavor( ExtractWord( 1, S, _BLANK ) );

  t := ExtractWord( 2, S, _BLANK );
  if cmpi( t, 'Hub' ) = 0 then
    modifier := route_hub
  else if cmpi( t, 'Host' ) = 0 then
    modifier := route_host
  else if cmpi( t, 'Boss' ) = 0 then
    modifier := route_boss
  else
  begin
    modifier := route_raw;
    route_to.Assign( t );
    if not route_to.Valid then
      raise Exception.Create( 'Bad route modifier or addr: "' + t + '"' );
  end;
end; { Eat }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TRouteAction.Apply( M: PFtnMessage ) : Boolean;
var
  pkt   : PPacket;
  out   : PPackedMessage;
  bin   : TBufStream;
  dest  : TAddress;
  target: TAddress;
  utname: String;
begin
  Result := False;

  if M^.Echomail then
    raise Exception.Create( 'Route: (!) can''t route an echomail message' );

  CheckDir( Cfg.Outbound, 'Outbound' );

  dest := M^.GetDest;

  case modifier of
    route_boss:
      begin
        target := dest;
        target.Point := 0;
      end;

    route_host:
      begin
        target := dest;
        target.Node  := 0;
        target.Point := 0;
      end;

    route_hub:
      begin
        if not Cfg.OpenNodelist or not Cfg.Nodelist^.FindHub( dest, target ) then
          raise Exception.Create( 'Route: (!) no hub-routing rule for ' + dest.MakeStr );
      end;

    route_raw:
      begin
        target := dest;
        route_to.Apply( target );
      end;
  end;

  if CompAddr( target, Cfg.PrimaryAddr ) = 0 then Exit;

  utname := make_ut_name( target, flavor );

  if not LockAddr( target ) then
    raise EIgnoreAction.Create( 'Message skipped.....' )
  else
  try
    out := New( PPackedMessage, Init );
    out^.Clone( M );
    out^.AddVia( Cfg.MakeOurViaStamp );
    if FileExists( utname ) then
    begin
      bin.Init( utname, stOpenWrite, STREAM_BUFFER_SIZE );
      try
        if bin.Status <> stOk then
          raise Exception.Create( 'Route: (!) could not open the packet "' + utname + '"' );
        bin.Seek( bin.GetSize - 2 );
        out^.BinSave( bin );
        bin.Write( _ZERO_WORD, 2 );
      finally
        bin.Done;
        Destroy( out );
      end;
    end
    else
    begin
      pkt := New( PPacket, Init );
      try
        with pkt^ do
        begin
          SetOrig( Cfg.PrimaryAddr );
          SetDest( target );
          SetDate( Now );
          Insert( out );
          SaveAs( utname );
        end;
      finally
        Destroy( pkt );
      end;
    end;
    Log( ll_Action, 'Route: ' + FlavorStr(Flavor) + ' => ' + target.MakeStr );

  finally
    UnlockAddr( target );
  end;

end; { Apply }

//////////////////////////////////////////////////////////////////////////
//
// TKillSentAction
//
//////////////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TKillSentAction.Apply( M: PFtnMessage ) : Boolean;
begin
  Result := True;
  if TestBit( M^.GetAttr, ATTR_KILLSENT ) then
  begin
    M^.Kill;
    Log( ll_Action, 'Kill-Sent fired' );
  end
  else
  begin
    M^.SetAttr( M^.GetAttr or ATTR_SENT );
    Log( ll_Action, 'Message marked as sent' );
  end;
end; { Apply }

//////////////////////////////////////////////////////////////////////////
//
// TIgnoreAction
//
//////////////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TIgnoreAction.Apply( M: PFtnMessage ) : Boolean;
begin
  raise EIgnoreAction.Create( 'Message ignored.....' );
end; { Apply }

{ --------------------------------------------------------- }
{ ConstructAction                                           }
{ --------------------------------------------------------- }

function ConstructAction( const S: String ) : PAction;
var
  Key: String;
  Val: String;
begin
  SplitPair( S, Key, Val );
  if cmpi( Key, 'Rewrite' ) = 0 then
    Result := New( PRewriteAction, Init )
  else if cmpi( Key, 'Delete' ) = 0 then
    Result := New( PDeleteAction, Init )
  else if cmpi( Key, 'Save' ) = 0 then
    Result := New( PSaveAction, Init )
  else if cmpi( Key, 'Copy' ) = 0 then
    Result := New( PCopyAction, Init )
  else if cmpi( Key, 'Bounce' ) = 0 then
    Result := New( PBounceAction, Init )
  else if cmpi( Key, 'Forward' ) = 0 then
    Result := New( PForwardAction, Init )
  else if cmpi( Key, 'Route' ) = 0 then
    Result := New( PRouteAction, Init )
  else if cmpi( Key, 'Kill-Sent' ) = 0 then
    Result := New( PKillSentAction, Init )
  else if cmpi( Key, 'Ignore' ) = 0 then
    Result := New( PIgnoreAction, Init )
  else
    raise Exception.Create( 'Unknown action: "' + Key + '"' );

  Result^.Eat( Val );

end; { ConstructAction }


end.

