unit _Catcher;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses Objects, MyLib, MsgApi, _Hook;

////////////////////////////////////////////////////////////////
//
// Catcher
//
////////////////////////////////////////////////////////////////

type
  PCatcher = ^TCatcher;
  TCatcher = object (TObject)
    constructor Init;
    destructor Done; virtual;
    function Eat( const Key, Val: String ) : Boolean;
    function Pickup( M: PFtnMessage ) : Boolean;
  private
    xmask  : PExtMask;
    Include: PStrings;
    Exclude: PStrings;
    Hooks  : PCollection;
    Actions: PCollection;
  end; { TCatcher }

  PEasyRouter = ^TEasyRouter;
  TEasyRouter = object (TCatcher)
    procedure Simulate( const S: String );
  end; { TEasyRouter }

{ =================================================================== }

implementation

uses _Action, _Config, SysUtils, UnixMask;

////////////////////////////////////////////////////////////////
//
// Catcher
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TCatcher.Init;
begin
  inherited Init;
  Include := New( PStrings, Init(10, 10) );
  Exclude := New( PStrings, Init(10, 10) );
  Hooks   := New( PCollection, Init(10, 10) );
  Actions := New( PCollection, Init(10, 10) );
  xmask   := nil;
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TCatcher.Done;
begin
  Destroy( Include );
  Destroy( Exclude );
  Destroy( Hooks );
  Destroy( Actions );
end; { Done }

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TCatcher.Eat( const Key, Val: String ) : Boolean;
var
  Hook  : PHook;
  Action: PAction;
begin
  Result := False;
  if xmask <> nil then
  begin
    if xmask^.Eat( Key + ' ' + Val ) then
      xmask := nil;
  end
  else if cmpi( Key, 'Include' ) = 0 then
    Include^.Insert( AllocStr(Val) )
  else if cmpi( Key, 'Exclude' ) = 0 then
    Exclude^.Insert( AllocStr(Val) )
  else if cmpi( Key, 'Mask' ) = 0 then
  begin
    Hook := New( PMask, Init );
    Hooks^.Insert( Hook );
    Hook^.Eat( Val );
  end
  else if cmpi( Key, 'Name' ) = 0 then
  begin
    Hook := New( PNameHook, Init );
    Hooks^.Insert( Hook );
    Hook^.Eat( Val );
  end
  else if cmpi( Key, 'XMask' ) = 0 then
  begin
    xmask := New( PExtMask, Init );
    Hooks^.Insert( xmask );
  end
  else if cmpi( Key, 'Action' ) = 0 then
  begin
    Action := ConstructAction( Val );
    if Action <> nil then
      Actions^.Insert( Action );
  end
  else if cmpi( Key, 'End' ) = 0 then
    Result := True
  else
    raise Exception.Create( 'Bad token in Catch block' );
end; { Eat }

{ --------------------------------------------------------- }
{ Pickup                                                    }
{ --------------------------------------------------------- }

function TCatcher.Pickup( M: PFtnMessage ) : Boolean;
var
  Changed: Boolean;

  function AreaMatch( const Pattern: VString ) : Boolean; far;
  var
    State: UnixMask.TState;
  begin
    UnixMask.MakeAutomatOn( Pattern^, State );
    Result := UnixMask.IsAcceptable( M^.GetArea, State );
  end; { AreaMatch }

  function HookMatch( Hook: PHook ) : Boolean; far;
  begin
    Result := Hook^.Hit( M );
  end; { HookMatch }

  procedure DoApply( Action: PAction ); far;
  begin
    if not M^.Deleted then
    begin
      if Action^.Apply( M ) then
        Changed := True;
    end;
  end; { DoApply }

begin
  Result := False;

  if M^.Deleted then
    Exit;

  if (Include^.FirstThat( @AreaMatch ) = nil) or
     (Exclude^.FirstThat( @AreaMatch ) <> nil) or
     (Hooks^.FirstThat( @HookMatch ) = nil)
  then
    Exit;

  Log( ll_Hook, 'Matched message found:' );
  Log( ll_Header, 'Area : ' + M^.GetArea );
  Log( ll_Header, 'From : ' + M^.GetFrom + ', ' + M.GetOrig.MakeStr );
  if M^.Echomail then
    Log( ll_Header, 'To   : ' + M^.GetTo )
  else
    Log( ll_Header, 'To   : ' + M^.GetTo + ', ' + M.GetDest.MakeStr );
  Log( ll_Header, 'Subj : ' + M^.GetSubj );
  Log( ll_Header, 'Date : ' + MsgDateStr(M^.GetDate) );
  if not M^.Echomail then
    Log( ll_Header, 'Attr : ' + AttrStr(M^.GetAttr) );
  Log( ll_Header, 'MsgID: ' + M^.GetMsgID );

  Changed := False;

  try
    Actions^.ForEach( @DoApply );
  except
    on I: EIgnoreAction do
      raise;
    on E: Exception do
      Log( ll_Error, E.Message + ' (rest of actions skipped)' );
  end;

  Result := Changed;

end; { Pickup }

//////////////////////////////////////////////////////////////////////////
//
// EasyRouter
//
//////////////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Simulate                                                  }
{ --------------------------------------------------------- }

procedure TEasyRouter.Simulate( const S: String );
var
  j: Integer;
  n: Integer;
  t: String;
  flavor   : String;
  route_to : String;
begin
  n := WordCount( S, _BLANK );

  flavor   := ExtractWord( 1, S, _BLANK );
  route_to := ExtractWord( 2, S, _BLANK );

  Eat( 'include', 'netmail' );
  Eat( 'xmask',   '' );
  Eat( 'attr',    '-snt' );
  Eat( 'dest',    '!' + Cfg.PrimaryAddr.makestr );
  Eat( 'union',   '' );

  for j := 3 to n do
  begin
    t := ExtractWord( j, S, _BLANK );
    if Pos( '=', t ) > 0 then
      Eat( ExtractWord( 1, t, _EQ ), ExtractWord( 2, t, _EQ ) )
    else
      Eat( 'dest', t );
  end;

  Eat( 'end', '' );
  Eat( 'end', '' );

  Eat( 'action', 'route ' + flavor + ' ' + route_to );
  Eat( 'action', 'kill-sent' );
  Eat( 'end',    '' );
end; { Simulate }

end.

