unit _BSO;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  FtnAddr;

type
  TBsoFlavor = (bso_norm, bso_hold, bso_dir);

function make_bso_name( Addr: TAddress ) : String;
function make_ut_name( Addr: TAddress; Flavor: TBsoFlavor ) : String;

function LockAddr( Addr: TAddress ) : Boolean;
procedure UnlockAddr( Addr: TAddress );

function FlavorStr( Flavor: TBsoFlavor ) : String;
function WhatFlavor( const S: String ) : TBsoFlavor;

{ =================================================================== }

implementation

uses
{$IFDEF VIRTUALPASCAL}
  VpUtils,
{$ENDIF}
  SysUtils, MyLib, _Config, _Busy;

type
  TBsoFlavorExt  = array [TBsoFlavor] of String[4];
  TBsoFlavorName = array [TBsoFlavor] of String [10];

const
  ut_ext: TBsoFlavorExt = ( '.out', '.hut', '.dut' );
  lo_ext: TBsoFlavorExt = ( '.flo', '.hlo', '.dlo' );

const
  FlavorName: TBsoFlavorName = ( 'Normal', 'Hold', 'Direct' );

{ --------------------------------------------------------- }
{ make_bso_name                                             }
{ --------------------------------------------------------- }

function make_bso_name( Addr: TAddress ) : String;
begin
  Result := AtPath( IntToHex(Addr.Net, 4) + IntToHex(Addr.Node, 4), Cfg.Outbound );
  if Addr.Point <> 0 then
  begin
    Result := Result + '.PNT';
    if not DirExists( Result ) then
    begin
      if not CreateDir( Result ) then
        raise Exception.Create( 'Could not create BSO-path: "' + Result + '"' );
    end;
    Result := Result + '\' + IntToHex(Addr.Point, 8);
  end;
end; { make_bso_name }

{ --------------------------------------------------------- }
{ make_ut_name                                              }
{ --------------------------------------------------------- }

function make_ut_name( Addr: TAddress; Flavor: TBsoFlavor ) : String;
begin
  Result := make_bso_name( Addr ) + ut_ext[Flavor];
end; { make_ut_name }

{ --------------------------------------------------------- }
{ FlavorStr                                                 }
{ --------------------------------------------------------- }

function FlavorStr( Flavor: TBsoFlavor ) : String;
begin
  Result := FlavorName[Flavor];
end; { FlavorStr }

{ --------------------------------------------------------- }
{ WhatFlavor                                                }
{ --------------------------------------------------------- }

function WhatFlavor( const S: String ) : TBsoFlavor;
begin
  for Result := Low(TBsoFlavor) to High(TBsoFlavor) do
    if cmpi( FlavorName[Result], S ) = 0 then
      Exit;
  raise EConvertError.Create( '"' + S + '" is not valid BSO flavor' );
end; { WhatFlavor }

{ --------------------------------------------------------- }
{ LockAddr                                                  }
{ --------------------------------------------------------- }

function LockAddr( Addr: TAddress ) : Boolean;
begin
  Result := SetBusyFlag( make_bso_name(Addr) + '.bsy' );
end; { LockAddr }

{ --------------------------------------------------------- }
{ UnlockAddr                                                }
{ --------------------------------------------------------- }

procedure UnlockAddr( Addr: TAddress );
begin
  DropBusyFlag( make_bso_name(Addr) + '.bsy' );
end; { UnlockAddr }

end.

