unit FtnAddr;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses MyLib;

type
  PAddress = ^TAddress;
  TAddress = object
    Zone, Net, Node, Point: SWORD;

    function Assign( const S: String ) : Boolean;
    procedure Merge( _Zone, _Net, _Node, _Point: SWORD );
    function Valid: Boolean;
    function MakeStr: String;
    function Match( const Source: TAddress ) : Boolean;
    procedure Apply( var Source: TAddress );
  end; { TAddress }

function CompAddr( const A1, A2: TAddress ) : Integer;

{ =================================================================== }

implementation

uses SysUtils;

////////////////////////////////////////////////////////////////
//
// TAddress
//
////////////////////////////////////////////////////////////////

const
  wildcard_all = $FFFE;

{ --------------------------------------------------------- }
{ Assign                                                    }
{                                                           }
{ This code based on GoldEd+ sources                        }
{ --------------------------------------------------------- }

function TAddress.Assign( const S: String ) : Boolean;

function atow( p: Integer ) : SWORD;
var
  E: Integer;
  j: Integer;
begin
  j := 0;
  while (p + j <= Length(S)) and isdigit(S[p+j]) do Inc(j);
  Val( Copy(S, p, j), Result, E );
  if E <> 0 then
    raise EConvertError.Create( 'Bad part in FTN-address' );
end; { atow }

function GetAddrPart( var Part: SWORD; p: Integer ) : Boolean;
begin
  Result := True;
  if p > Length(S) then
    Result := False
  else if S[p] = '*' then
    Part := wildcard_all
  else if isdigit( S[p] ) then
    try
      Part := atow( p )
    except
      Result := False;
    end
  else
    Result := False;
end; { GetAddrPart }

var
  str: Integer;
  ptr: Integer;
  gotzone : Boolean;
  gotnet  : Boolean;
  gotnode : Boolean;
  gotpoint: Boolean;

begin
  Zone  := 0;
  Net   := 0;
  Node  := 0;
  Point := 0;

  gotzone  := False;
  gotnet   := False;
  gotnode  := False;
  gotpoint := False;

  str := 1;
  ptr := 1;

  if (S <> '') and (S[1] = '*') then
  begin
    Zone  := wildcard_all;
    Net   := wildcard_all;
    Node  := wildcard_all;
    Point := wildcard_all;
  end;

  while (ptr <= Length(S)) and (S[ptr] > ' ') do
  begin
    case S[ptr] of
      ':':
        begin
          gotzone := GetAddrPart( Zone, str );
          Inc( ptr );
          str := ptr;
          if str <= Length(S) then
            gotnet := GetAddrPart( Net, str );
        end;

      '/':
        begin
          if not gotnet then
            gotnet := GetAddrPart( Net, str );
          Inc( ptr );
          str := ptr;
          if str <= Length(S) then
            gotnode := GetAddrPart( Node, str );
        end;

      '.':
        begin
          if not gotnode and (S[str] <> '.') then
            gotnode := GetAddrPart( Node, str );
          Inc( ptr );
          str := ptr;
          if str <= Length(S) then
            GetAddrPart( Point, str );
          gotpoint := True;
        end;

      '@':
        begin
          if not (gotzone or gotnet or gotnode or gotpoint) then
            gotnode := GetAddrPart( Node, str );
          while (ptr <= Length(S)) and (S[ptr] > ' ') do
            Inc(ptr);
        end;

    else
      Inc(ptr);
    end;
  end;

  if not (gotzone or gotnet or gotnode or gotpoint) then
    gotnode := GetAddrPart( Node, str );

  if gotzone and not gotnet then
    Net := 0;

  if gotzone and not gotnode then
  begin
    if Net = wildcard_all then
      Node := wildcard_all
    else
      Node := 0;
    gotnode := True;
  end;

  if gotnode and not gotpoint then
  begin
    if Node = wildcard_all then
      Point := wildcard_all
    else
      Point := 0;
  end;

  Result := Net <> 0;

end; { Assign }

{ --------------------------------------------------------- }
{ Merge                                                     }
{ --------------------------------------------------------- }

procedure TAddress.Merge( _Zone, _Net, _Node, _Point: SWORD );
begin
  Zone  := _Zone;
  Net   := _Net;
  Node  := _Node;
  Point := _Point;
end; { Merge }

{ --------------------------------------------------------- }
{ Valid                                                     }
{ --------------------------------------------------------- }

function TAddress.Valid: Boolean;
begin
  Result := Net <> 0;
end; { Valid }


{ --------------------------------------------------------- }
{ MakeStr                                                   }
{ --------------------------------------------------------- }

function TAddress.MakeStr: String;
begin
  Result := IntToStr(Zone) + ':' + IntToStr(Net) + '/' + IntToStr(Node);
  if Point <> 0 then
    Result := Result + '.' + IntToStr(Point);
end; { MakeStr }


{ --------------------------------------------------------- }
{ Match                                                     }
{ --------------------------------------------------------- }

function TAddress.Match( const Source: TAddress ) : Boolean;
begin
  Result := ((Zone  = wildcard_all) or (Source.Zone  = Zone))
        and ((Net   = wildcard_all) or (Source.Net   = Net))
        and ((Node  = wildcard_all) or (Source.Node  = Node))
        and ((Point = wildcard_all) or (Source.Point = Point ));
end; { Match }


{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

procedure TAddress.Apply( var Source: TAddress );
begin
  if Zone <> wildcard_all then
    Source.Zone := Zone;
  if Net <> wildcard_all then
    Source.Net := Net;
  if Node <> wildcard_all then
    Source.Node := Node;
  if Point <> wildcard_all then
    Source.Point := Point;
end; { Apply }

{ --------------------------------------------------------- }
{ CompAddr                                                  }
{ --------------------------------------------------------- }

function CompAddr( const A1, A2: TAddress ) : Integer;
begin
  if A1.Zone < A2.Zone then
    Result := -1
  else if A1.Zone > A2.Zone then
    Result := 1
  else if A1.Net < A2.Net then
    Result := -1
  else if A1.Net > A2.Net then
    Result := 1
  else if A1.Node < A2.Node then
    Result := -1
  else if A1.Node > A2.Node then
    Result := 1
  else if A1.Point < A2.Point then
    Result := -1
  else if A1.Point > A2.Point then
    Result := 1
  else
    Result := 0;
end; { CompAddr }

end.

