unit MyLib;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, SysUtils;

{$IFDEF VIRTUALPASCAL}
type
  DWORD = Longint;
  SWORD = SmallWord;
{$ENDIF}

{$IFDEF FPC}
type
  SWORD = Word;
{$ENDIF}

type
  VString = ^AnsiString;

  Charset  = set of Char;
  UnixTime = DWORD;

const
  _CR = #13;
  _LF = #10;

const
  _ZERO_BYTE : Byte = 0;
  _ZERO_WORD : SWORD = 0;
  _BLANK     : Charset = [' '];
  _COMMA     : Charset = [','];
  _EQ        : Charset = ['='];

const
  _MONTH_NAME: array [1..12] of String[3] = (
    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' );

const
  _DAY_NAME: array [1..7] of String[3] = (
    'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat' );

function AllocStr( const S: String ) : VString;
procedure FreeStr( var V: VString );
procedure ReplaceStr( var V: VString; const S: String );

procedure Replace( var S: String; pos, len: Integer; const D: String );
function  ReplaceSubstr( const S, Pattern, Target: String ) : String;
procedure SplitPair( const S: String; var Key, Val: String );

function ScanR( var P; Offset, Size: Integer; C: Char ): Integer;
function SkipR( var P; Offset, Size: Integer; C: Char ): Integer;

function cmpi( const s1, s2: String ) : Integer;
function isdigit( c: Char ) : Boolean;
function Trim( const S: String ) : String;
function CharStr( Ch: Char; n: Integer ) : String;
function StrSet( Ch: PChar; const S: String; Size: Integer ) : PChar;

function AddBackSlash(const DirName : string) : string;
function AtPath( const FileName, Path: String ) : String;

function WordCount(const S : string; WordDelims : CharSet) : Integer;
function WordPosition(N : Integer; const S : string; WordDelims : CharSet) : Integer;
function ExtractWord(N : Integer; const S : string; WordDelims : CharSet) : String;
function ExtractQuoted( const S: String; n: Integer ) : String;
function HexToInt( S: String ) : Integer;

procedure SetBit( var Where: Longint; Mask: Longint; Raised: Boolean );
function TestBit( Where, Mask: Longint ) : Boolean;

function DateTimeToUnix( DT: TDateTime ) : UnixTime;
Function UnixToDateTime( UT: UnixTime ) : TDateTime;
function CurrentUnixTime: UnixTime;

procedure SetupDateTimeFormat;

{$IFNDEF VIRTUALPASCAL}
function DirExists( _Path: String ) : Boolean;
function StrPNew( Str: String ) : PChar;
{$ENDIF}

procedure CheckDir( const Dir, Name: String );
procedure CheckFile( const FileName: String );

function CreateSemaphore( const FileName: String ) : Boolean;

procedure Destroy( O: PObject );

type
  PStrings = ^TStrings;
  TStrings = object (TCollection)
    procedure FreeItem( Item: Pointer ); virtual;
  end; { TStrings }

{ =================================================================== }

implementation

{$IFDEF VIRTUALPASCAL}
uses VpUtils;
{$ENDIF}

type
  TXlateArray = array [Char] of Char;

const
  Upper_Xlate: TXlateArray = (

#0, #1, #2, #3, #4, #5, #6, #7, #8, #9, #10, #11, #12, #13, #14, #15,
#16, #17, #18, #19, #20, #21, #22, #23, #24, #25, #26, #27, #28, #29, #30, #31,
#32, #33, #34, #35, #36, #37, #38, #39, #40, #41, #42, #43, #44, #45, #46, #47,
#48, #49, #50, #51, #52, #53, #54, #55, #56, #57, #58, #59, #60, #61, #62, #63,
#64, #65, #66, #67, #68, #69, #70, #71, #72, #73, #74, #75, #76, #77, #78, #79,
#80, #81, #82, #83, #84, #85, #86, #87, #88, #89, #90, #91, #92, #93, #94, #95,
#96, #65, #66, #67, #68, #69, #70, #71, #72, #73, #74, #75, #76, #77, #78, #79,
#80, #81, #82, #83, #84, #85, #86, #87, #88, #89, #90, #123, #124, #125, #126, #127,
#128, #129, #130, #131, #132, #133, #134, #135, #136, #137, #138, #139, #140, #141, #142, #143,
#144, #145, #146, #147, #148, #149, #150, #151, #152, #153, #154, #155, #156, #157, #158, #159,
#128, #129, #130, #131, #132, #133, #134, #135, #136, #169, #138, #139, #140, #141, #142, #143,
#176, #177, #178, #179, #180, #181, #182, #183, #184, #185, #186, #187, #188, #189, #190, #191,
#192, #193, #194, #195, #196, #197, #198, #199, #200, #201, #202, #203, #204, #205, #206, #207,
#208, #209, #210, #211, #212, #213, #214, #215, #216, #217, #218, #219, #220, #221, #222, #223,
#144, #145, #146, #147, #148, #149, #150, #151, #152, #153, #154, #155, #156, #157, #158, #159,
#240, #240, #242, #243, #244, #245, #246, #247, #248, #249, #250, #251, #252, #253, #254, #255

  );

{ --------------------------------------------------------- }
{ UpperCase                                                 }
{ --------------------------------------------------------- }

function UpperCase( S: String ) : String;
var
  j: Integer;
begin
  for j := 1 to Length(S) do
    S[j] := Upper_Xlate[S[j]];
  Result := S;
end; { UpperCase }

{ --------------------------------------------------------- }
{ Destroy                                                   }
{ --------------------------------------------------------- }

procedure Destroy( O: PObject );
begin
  if O <> nil then
    Dispose( O, Done );
end; { Destroy }

{ --------------------------------------------------------- }
{ ScanR                                                     }
{ --------------------------------------------------------- }

{$IFDEF VIRTUALPASCAL}

function ScanR(var P; Offset, Size: Integer; C: Char): Integer; assembler;
        {&USES esi,edi} {&FRAME-}
asm
        cld
        mov     edi,P
        mov     edx, edi
        add     edi,&Offset
        mov     ecx,Size
        sub     ecx,&Offset
        mov     al,C
        repne   scasb
        je      @@found
        mov     eax, Size
        jmp     @@exit
@@found:
        sub     edi, edx
        mov     eax, edi
        dec     eax
@@exit:
end;

{$ELSE}

function ScanR( var P; Offset, Size: Integer; C: Char ) : Integer;
var
  B: TByteArray absolute P;
begin
  while (Offset < Size) and (B[Offset] <> Byte(C)) do Inc(Offset);
  Result := Offset;
end; { ScanR }

{$ENDIF}

{ --------------------------------------------------------- }
{ SkipR                                                     }
{ --------------------------------------------------------- }

{$IFDEF VIRTUALPASCAL}

function SkipR( var P; Offset, Size: Integer; C: Char ): Integer; assembler;
        {&USES esi,edi} {&FRAME-}
asm
        cld
        mov     edi,P
        mov     edx, edi
        add     edi,&Offset
        mov     ecx,Size
        sub     ecx,&Offset
        mov     al,C
        repe    scasb
        jne     @@found
        mov     eax, Size
        jmp     @@exit
@@found:
        sub     edi, edx
        mov     eax, edi
        dec     eax
@@exit:
end;

{$ELSE}

function SkipR( var P; Offset, Size: Integer; C: Char ) : Integer;
var
  B: TByteArray absolute P;
begin
  while (Offset < Size) and (B[Offset] = Byte(C)) do Inc(Offset);
  Result := Offset;
end; { ScanR }

{$ENDIF}

{ --------------------------------------------------------- }
{ AllocStr                                                  }
{ --------------------------------------------------------- }

const
  EmptyStr: AnsiString = '';
  NullStr: VString = @EmptyStr;

function AllocStr( const S: String ) : VString;
begin
  if S = '' then
    Result := NullStr
  else
  begin
{$IFDEF VIRTUALPASCAL}
    New( Result );
{$ELSE}
  {$IFDEF FPC}
    GetMem( Result, Length(S) + 1 );
  {$ELSE}
    !!! Enter compiler-specific code here !!!
  {$ENDIF}
{$ENDIF}
    Result^ := S;
  end;
end; { AllocStr }

{ --------------------------------------------------------- }
{ FreeStr                                                   }
{ --------------------------------------------------------- }

procedure FreeStr( var V: VString );
begin
  if (V <> nil) and ( V <> NullStr) then
{$IFDEF VIRTUALPASCAL}
    Dispose( V );
{$ELSE}
  {$IFDEF FPC}
    FreeMem( V, Length(V^) + 1 );
  {$ELSE}
    !!! Enter compiler-specific code here !!!
  {$ENDIF}
{$ENDIF}
  V := nil;
end; { FreeStr }

{ --------------------------------------------------------- }
{ ReplaceStr                                                }
{ --------------------------------------------------------- }

procedure ReplaceStr( var V: VString; const S: String );
begin
  FreeStr( V );
  V := AllocStr( S );
end; { ReplaceStr }

{ --------------------------------------------------------- }
{ Replace                                                   }
{ --------------------------------------------------------- }

procedure Replace( var S: String; pos, len: Integer; const D: String );
begin
  System.Delete( S, pos, len );
  System.Insert( D, S, pos );
end; { Replace }

{ --------------------------------------------------------- }
{ ReplaceSubstr                                             }
{ --------------------------------------------------------- }

function ReplaceSubstr( const S, Pattern, Target: String ) : String;
var
  p: Integer;
begin
  Result := S;
  p := Pos( Pattern, Result );
  while p > 0 do
  begin
    Replace( Result, p, Length(Pattern), Target );
    p := Pos( Pattern, Result );
  end;
end; { ReplaceSubstr }

{ --------------------------------------------------------- }
{ cmpi                                                      }
{ --------------------------------------------------------- }

function cmpi( const s1, s2: String ) : Integer;
begin
  Result := AnsiCompareText( s1, s2 );
end; { cmpi }

{ --------------------------------------------------------- }
{ SplitPair                                                 }
{ --------------------------------------------------------- }

procedure SplitPair( const S: String; var Key, Val: String );
var
  j: Integer;
  x: Integer;
begin
  Key := '';
  Val := '';
  j   := 0;
  repeat
    Inc(j);
    if j > Length(S) then Exit;
  until S[j] <> ' ';
  x := j;
  repeat
    Inc(j);
  until (j > Length(S)) or (S[j] = ' ');
  Key := System.Copy( S, x, j - x );
  if j > Length(S) then Exit;
  repeat
    Inc(j);
    if j > Length(S) then Exit;
  until S[j] <> ' ';
  Val := System.Copy( S, j, Length(S) );
end; { SplitPair }

{ --------------------------------------------------------- }
{ isdigit                                                   }
{ --------------------------------------------------------- }

function isdigit( c: Char ) : Boolean;
begin
  Result := (c >= '0') and (c <= '9');
end; { isdigit }

{ --------------------------------------------------------- }
{ Trim                                                      }
{ --------------------------------------------------------- }

function Trim( const S: String ) : String;
begin
  Result := TrimLeft( TrimRight(S) );
end; { Trim }

{ --------------------------------------------------------- }
{ CharStr                                                   }
{ --------------------------------------------------------- }

function CharStr( Ch: Char; n: Integer ) : String;
begin
  SetLength( Result, n );
  FillChar( Result[1], n, Ch );
end; { CharStr }

{ --------------------------------------------------------- }
{ StrSet                                                    }
{ --------------------------------------------------------- }

function StrSet( Ch: PChar; const S: String; Size: Integer ) : PChar;
begin
  FillChar( Ch[0], Size, 0 );
  if Size > Length(S) then
    Size := Length(S);
  if Size > 0 then
    Move( S[1], Ch[0], Size );
end; { StrSet }

{ --------------------------------------------------------- }
{ WordCount                                                 }
{ --------------------------------------------------------- }

function WordCount(const S : string; WordDelims : CharSet) : Integer;
var
  I : Integer;
begin
  Result := 0;
  I := 1;
  while I <= Length(S) do
  begin
    {skip over delimiters}
    while (I <= Length(S)) and (S[I] in WordDelims) do
      Inc(I);

    {if we're not beyond end of S, we're at the start of a word}
    if I <= Length(S) then
      Inc(Result);

    {find the end of the current word}
    while (I <= Length(S)) and not(S[I] in WordDelims) do
      Inc(I);
  end;
end; { WordCount }

{ --------------------------------------------------------- }
{ WordPosition                                              }
{ --------------------------------------------------------- }

function WordPosition(N : Integer; const S : string; WordDelims : CharSet) : Integer;
var
  Count : Integer;
  I : Integer;
begin
  Count := 0;
  I := 1;
  Result := 0;

  while (I <= Length(S)) and (Count <> N) do
  begin
    {skip over delimiters}
    while (I <= Length(S)) and (S[I] in WordDelims) do
      Inc(I);

    {if we're not beyond end of S, we're at the start of a word}
    if I <= Length(S) then
      Inc(Count);

    {if not finished, find the end of the current word}
    if Count <> N then
      while (I <= Length(S)) and not(S[I] in WordDelims) do
        Inc(I)
    else
        Result := I;
  end;
end; { WordPosition }


{ --------------------------------------------------------- }
{ ExtractWord                                               }
{ --------------------------------------------------------- }

function ExtractWord(N : Integer; const S : string; WordDelims : CharSet) : String;
var
  x, y : Integer;
begin
  x := WordPosition(N, S, WordDelims);
  y := x;
  if y <> 0 then
    while (y <= Length(S)) and not(S[y] in WordDelims) do
      Inc(y);
  Result := Copy( S, x, y - x );
end; { ExtractWord }

{ --------------------------------------------------------- }
{ ExtractQuoted                                             }
{ --------------------------------------------------------- }

function ExtractQuoted( const S: String; n: Integer ) : String;
var
  Q: Char;
begin
  Q := S[n];
  Inc(n);
  Result := '';

  while n <= Length(S) do
  begin
    if S[n] = Q then
    begin
      if (n + 1 <= Length(S)) and (S[n+1] = Q) then
        Inc(n)
      else
        Exit;
    end;
    Result := Result + S[n];
    Inc(n);
  end;

end; { ExtractQuoted }

{ --------------------------------------------------------- }
{ HexToInt                                                  }
{ --------------------------------------------------------- }

function HexToInt( S: String ) : Integer;
var
  j: Integer;
begin
  S := UpperCase( S );
  Result := 0;
  for j := Length(S) downto 1 do
  begin
    if isdigit( S[j] ) then
      Result := Result or (( Ord(S[j]) - Ord('0') ) shl (4 * (Length(S) - j) ))
    else if S[j] in ['A'..'F'] then
      Result := Result or (( 10 + Ord(S[j]) - Ord('A') ) shl (4 * (Length(S) - j) ))
    else
      raise EConvertError.Create( '"' + S + '" is invalid hex number' );
  end;
end; { HexToInt }

{ --------------------------------------------------------- }
{ AddBackSlash                                              }
{ --------------------------------------------------------- }

const
  DosDelimSet = ['\', ':', #0];

function AddBackSlash(const DirName : string) : string;
begin
  if DirName[Length(DirName)] in DosDelimSet then
    Result := DirName
  else
    Result := DirName + '\';
end; { AddBackSlash }

{ --------------------------------------------------------- }
{ AtPath                                                    }
{ --------------------------------------------------------- }

function AtPath( const FileName, Path: String ) : String;
begin
  Result := AddBackSlash(Path) + ExtractFileName(FileName);
end; { AtPath }

{ --------------------------------------------------------- }
{ SetBit                                                    }
{ --------------------------------------------------------- }

procedure SetBit( var Where: Longint; Mask: Longint; Raised: Boolean );
begin
  if Raised then
    Where := Where or Mask
  else
    Where := Where and not Mask;
end; { SetBit }

{ --------------------------------------------------------- }
{ TestBit                                                   }
{ --------------------------------------------------------- }

function TestBit( Where, Mask: Longint ) : Boolean;
begin
  Result := (Where and Mask) <> 0;
end; { TestBit }

{$IFNDEF VIRTUALPASCAL}

{ --------------------------------------------------------- }
{ DirExists                                                 }
{ --------------------------------------------------------- }

function DirExists( _Path: String ): Boolean;
var
  DirInfo: TSearchRec;
begin
  // Remove trailing backslash, if any
  if _Path[Length(_Path)] = '\' then
    Delete( _Path, Length(_Path), 1 );
  // FindFirst: Allow any attribute; require Directory attribute
  Result := FindFirst( _Path, $3F, DirInfo ) = 0;
  if Result then
    FindClose(DirInfo);
end;

{ --------------------------------------------------------- }
{ StrPNew                                                   }
{ --------------------------------------------------------- }

function StrPNew(Str: String): PChar;
var
  Size: Cardinal;
begin
  if Str = '' then Result := nil else
  begin
    Size := Length(Str) + 1;
    Result := StrAlloc(Size);
    if Size > 1 then
      Move(Str[1], Result^, Size-1);
    Result[Size-1] := #0;
  end;
end;

{$ENDIF}

{ --------------------------------------------------------- }
{ CheckDir                                                  }
{ --------------------------------------------------------- }

procedure CheckDir( const Dir, Name: String );
begin
  if (Dir = '') or not DirExists( Dir ) then
    raise Exception.Create( 'Bad ' + Name + ' dir: "' + Dir + '"' );
end; { CheckDir }

{ --------------------------------------------------------- }
{ CheckFile                                                 }
{ --------------------------------------------------------- }

procedure CheckFile( const FileName: String );
begin
  if not SysUtils.FileExists( FileName ) then
    raise Exception.Create( 'File not found: "' + FileName + '"' );
end; { CheckFile }

{ --------------------------------------------------------- }
{ CreateSemaphore                                           }
{ --------------------------------------------------------- }

function CreateSemaphore( const FileName: String ) : Boolean;
var
  F: File;
begin
  try
    Assign( F, FileName );
    Rewrite( F );
  except
    Result := False;
    Exit;
  end;
  Close( F );
  Result := True;
end; { CreateSemaphore }

{ --------------------------------------------------------- }
{ SetupDateTimeFormat                                       }
{ --------------------------------------------------------- }

procedure SetupDateTimeFormat;
var
  j: Integer;
begin
  for j := 1 to 12 do
    ShortMonthNames[j] := _MONTH_NAME[j];
  for j := 1 to 7 do
    ShortDayNames[j] := _DAY_NAME[j];
  TimeSeparator   := ':';
end; { SetupDateTimeFormat }

////////////////////////////////////////////////////////////////
//
// Unix-Time routines
//
////////////////////////////////////////////////////////////////

const
  C1970 = 2440588;
  D0    =    1461;
  D1    =  146097;
  D2    = 1721119;


Procedure GregorianToJulianDN( Year, Month, Day : SWord;
                               var JulianDN : Longint );
var
  Century,
  XYear    : longint;
begin
  If Month <= 2 then
    begin
      Year  := Pred(Year);
      Month := Month + 12;
    end;
  Month    := Month - 3;
  Century  := Year div 100;
  XYear    := Year mod 100;
  Century  := (Century * D1) shr 2;
  XYear    := (XYear * D0) shr 2;
  JulianDN := ((((Month * 153) + 2) div 5) + Day) + D2 + XYear + Century;
end;


Procedure JulianDNToGregorian( JulianDN : longint;
                               var Year, Month, Day : SWord );
var
  Temp,
  XYear   : longint;
  YYear,
  YMonth,
  YDay    : integer;
begin
  Temp     := (((JulianDN - D2) shl 2) - 1);
  XYear    := (Temp mod D1) or 3;
  JulianDN := Temp div D1;
  YYear    := (XYear div D0);
  Temp     := ((((XYear mod D0) + 4) shr 2) * 5) - 3;
  YMonth   := Temp div 153;
  If YMonth >= 10 then
    begin
      YYear  := YYear + 1;
      YMonth := YMonth - 12;
    end;
  YMonth := YMonth + 3;
  YDay   := Temp mod 153;
  YDay   := (YDay + 5) div 5;
  Year   := YYear + (JulianDN * 100);
  Month  := YMonth;
  Day    := YDay;
end;

{ --------------------------------------------------------- }
{ DateTimeToUnix                                            }
{ --------------------------------------------------------- }

function DateTimeToUnix( DT: TDateTime ) : UnixTime;
var
  DateNum  : Longint;
  SecsPast : Longint;
  DaysPast : Longint;
  Day, Month, Year, Hour, Min, Sec, MSec: SWORD;
begin
  DecodeDate( DT, Year, Month, Day );
  DecodeTime( DT, Hour, Min, Sec, MSec );
  GregorianToJulianDN( Year, Month, Day, DateNum );
  DaysPast := DateNum - c1970;
  SecsPast := DaysPast * 86400;
  Result   := SecsPast + DWORD(Hour) * 3600 + Min * 60 + Sec;
end; { DateTimeToUnix }

{ --------------------------------------------------------- }
{ UnixToDateTime                                            }
{ --------------------------------------------------------- }

Function UnixToDateTime( UT: UnixTime ) : TDateTime;
var
  DateNum  : Longint;
  Day, Month, Year, Hour, Min, Sec, MSec: SWORD;
begin
  DateNum := (UT div 86400) + c1970;
  JulianDNToGregorian( DateNum, Year, Month, Day );
  ut := ut mod 86400;
  Hour := ut div 3600;
  ut := ut mod 3600;
  Min := ut div 60;
  Sec := ut mod 60;
  Result := EncodeDate( Year, Month, Day ) + EncodeTime( Hour, Min, Sec, MSec );
end; { UnixToDateTime }

{ --------------------------------------------------------- }
{ CurrentUnixTime                                           }
{ --------------------------------------------------------- }

function CurrentUnixTime: UnixTime;
begin
  Result := DateTimeToUnix( Now );
end; { CurrentUnixTime }


////////////////////////////////////////////////////////////////
//
// TStrings
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ FreeItem                                                  }
{ --------------------------------------------------------- }

procedure TStrings.FreeItem( Item: Pointer );
begin
  FreeStr( VString(Item) );
end; { FreeItem }

end.

